Package: tidyfst
Title: Tidy Verbs for Fast Data Manipulation
Version: 1.8.4
Depends: R (>= 4.1.0)
Authors@R: 
    person(given = "Tian-Yuan",
           family = "Huang",
           role = c("aut", "cre"),
           email = "huang.tian-yuan@qq.com",
           comment = c(ORCID = "0000-0002-4151-3764"))
Description: A toolkit of tidy data manipulation verbs with 'data.table' as the backend.
  Combining the merits of syntax elegance from 'dplyr' and computing performance from 'data.table', 
  'tidyfst' intends to provide users with state-of-the-art data manipulation tools with least pain.
  This package is an extension of 'data.table'. While enjoying a tidy syntax, 
  it also wraps combinations of efficient functions to facilitate frequently-used data operations.  
URL: https://github.com/hope-data-science/tidyfst,
        https://hope-data-science.github.io/tidyfst/
BugReports: https://github.com/hope-data-science/tidyfst/issues
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: data.table (>= 1.15.0), fst (>= 0.9.0), stringr (>= 1.5.0),
        pak
Suggests: knitr, rmarkdown, nycflights13, tidyr, ggplot2, dplyr, bench,
        testthat, tidytext
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-27 06:39:15 UTC; Admin
Author: Tian-Yuan Huang [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4151-3764>)
Maintainer: Tian-Yuan Huang <huang.tian-yuan@qq.com>
Repository: CRAN
Date/Publication: 2026-04-27 07:00:02 UTC

# Generated by roxygen2: do not edit by hand

export("%>%")
export("%chin%")
export("%like%")
export(CJ)
export(add_count_dt)
export(add_prop)
export(anti_join_dt)
export(arrange_dt)
export(as.data.table)
export(as_dt)
export(as_fst)
export(between)
export(bind_rows_dt)
export(bind_tf_idf_dt)
export(chop_dt)
export(col_max)
export(col_min)
export(col_rn)
export(complete_dt)
export(copy)
export(count_dt)
export(cummean)
export(data.table)
export(delete_na_cols)
export(delete_na_rows)
export(df_mat)
export(distinct_dt)
export(drop_na_dt)
export(dummy_dt)
export(export_fst)
export(fcase)
export(fcoalesce)
export(fill_na_dt)
export(filter_dt)
export(filter_fst)
export(fintersect)
export(fread)
export(frollapply)
export(fsetdiff)
export(fsetequal)
export(full_join_dt)
export(funion)
export(fwrite)
export(get_fst_chunk_size)
export(group_by_dt)
export(group_dt)
export(group_exe_dt)
export(import_fst)
export(import_fst_chunked)
export(impute_dt)
export(in_dt)
export(inner_join_dt)
export(intersect_dt)
export(key)
export(lag_dt)
export(lead_dt)
export(left_join_dt)
export(like)
export(longer_dt)
export(mat_df)
export(maxth)
export(minth)
export(mutate_dt)
export(mutate_vars)
export(mutate_when)
export(nest_dt)
export(nth)
export(object_size)
export(pairwise_count_dt)
export(parse_fst)
export(percent)
export(pkg_load)
export(pkg_unload)
export(print_options)
export(pst)
export(pull_dt)
export(rbindlist)
export(rec_char)
export(rec_num)
export(relocate_dt)
export(rename_dt)
export(rename_with_dt)
export(replace_dt)
export(replace_na_dt)
export(right_join_dt)
export(rleid)
export(rleidv)
export(rn_col)
export(round0)
export(rowwise_dt)
export(sample_dt)
export(sample_frac_dt)
export(sample_n_dt)
export(select_dt)
export(select_fst)
export(select_mix)
export(semi_join_dt)
export(separate_dt)
export(set)
export(setDF)
export(setDT)
export(setdiff_dt)
export(setequal_dt)
export(setnames)
export(shift_fill)
export(slice_dt)
export(slice_fst)
export(slice_head_dt)
export(slice_max_dt)
export(slice_min_dt)
export(slice_sample_dt)
export(slice_tail_dt)
export(sql_join_dt)
export(squeeze_dt)
export(summarise_dt)
export(summarise_vars)
export(summarise_when)
export(summarize_dt)
export(summarize_vars)
export(summarize_when)
export(summary_fst)
export(sys_time_print)
export(t_dt)
export(tables)
export(transmute_dt)
export(unchop_dt)
export(uncount_dt)
export(union_dt)
export(uniqueN)
export(unite_dt)
export(unnest_dt)
export(utf8_encoding)
export(wider_dt)
import(data.table)
import(fst)
import(stringr)
importFrom(data.table,"%chin%")
importFrom(data.table,"%like%")
importFrom(data.table,CJ)
importFrom(data.table,as.data.table)
importFrom(data.table,between)
importFrom(data.table,copy)
importFrom(data.table,data.table)
importFrom(data.table,fcase)
importFrom(data.table,fcoalesce)
importFrom(data.table,fintersect)
importFrom(data.table,fread)
importFrom(data.table,frollapply)
importFrom(data.table,fsetdiff)
importFrom(data.table,fsetequal)
importFrom(data.table,funion)
importFrom(data.table,fwrite)
importFrom(data.table,key)
importFrom(data.table,like)
importFrom(data.table,rbindlist)
importFrom(data.table,rleid)
importFrom(data.table,rleidv)
importFrom(data.table,set)
importFrom(data.table,setDF)
importFrom(data.table,setDT)
importFrom(data.table,setnames)
importFrom(data.table,tables)
importFrom(data.table,uniqueN)
importFrom(pak,pak)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(utils,head)
importFrom(utils,install.packages)
importFrom(utils,installed.packages)
importFrom(utils,object.size)
importFrom(utils,tail)

# tidyfst: Tidy Verbs for Fast Data Manipulation<img src="man/figures/hex-tidyfst_url.png" align="right" alt="" width="120" />

 [![](https://www.r-pkg.org/badges/version/tidyfst?color=orange)](https://cran.r-project.org/package=tidyfst) [![](https://img.shields.io/badge/devel%20version-2.0.0-purple.svg)](https://github.com/hope-data-science/tidyfst) ![](https://img.shields.io/badge/lifecycle-stable-green.svg)  [![downloads](http://cranlogs.r-pkg.org/badges/grand-total/tidyfst?color=yellow)](https://r-pkg.org/pkg/tidyfst)

 [![download](https://cranlogs.r-pkg.org/badges/tidyfst?color=red)](https://rdrr.io/cran/tidyfst/) [![downloads](https://cranlogs.r-pkg.org/badges/last-week/tidyfst?color=ff69b4)](https://cran.r-project.org/package=tidyfst) [![downloads](https://cranlogs.r-pkg.org/badges/last-day/tidyfst?color=9cf)](https://cran.r-project.org/package=tidyfst)

 [![ZENODO DOI](https://zenodo.org/badge/240626994.svg)](https://zenodo.org/badge/latestdoi/240626994) [![JOSS DOI](http://joss.theoj.org/papers/10.21105/joss.02388/status.svg)](https://joss.theoj.org/papers/10.21105/joss.02388)



## Overview

*tidyfst* is a toolkit of tidy data manipulation verbs with *data.table* as the backend . Combining the merits of syntax elegance from *dplyr* and computing performance from *data.table*,  *tidyfst* intends to provide users with state-of-the-art data manipulation tools with least pain. This package is an extension of *data.table*, while enjoying a tidy syntax, it also wraps combinations of efficient functions to facilitate frequently-used data operations.  Also, *tidyfst* would introduce more tidy data verbs from other packages, including but not limited to *tidyverse* and *data.table*. If you are a *dplyr* user but have to use *data.table* for speedy computation,  or *data.table* user looking for readable coding syntax, *tidyfst* is designed for you (and me of course). For further details and tutorials, see [vignettes](https://hope-data-science.github.io/tidyfst/). Both [Chinese](https://hope-data-science.github.io/tidyfst/articles/chinese_tutorial.html) and [English](https://hope-data-science.github.io/tidyfst/articles/english_tutorial.html) tutorials could be found there.

Till now, *tidyfst* has an API that might even transcend its predecessors (e.g. [`select_dt`](https://hope-data-science.github.io/tidyfst/reference/select.html) could accept nearly anything for super column selection). Enjoy the efficient data operations in *tidyfst* !

PS: For extreme performance in tidy syntax, try *tidyfst*'s mirror package [tidyft](https://github.com/hope-data-science/tidyft). 



## Features

- Receives any data.frame (tibble/data.table/data.frame) and returns a data.table.
- Show the variable class of data.table as default.
- Never use in place replacement (also known as modification by reference, which means the original variable would not be modified without notification).
- Use suffix ("_dt") rather than prefix to increase the efficiency (especially when you have IDE with automatic code completion).
- More flexible verbs (e.g. [pairwise_count_dt](https://hope-data-science.github.io/tidyfst/reference/pairwise.html)) for big data manipulation.
- Supporting data importing and parsing with *fst*, which saves both time and memory. Details see [parse_fst/select_fst/filter_fst](https://hope-data-science.github.io/tidyfst/reference/fst.html) and [import_fst/export_fst](https://hope-data-science.github.io/tidyfst/reference/fst_io.html).
- Low and stable dependency on mature packages (data.table, fst, stringr)



## Installation

```R
install.packages("tidyfst")
```



## Example

```R
library(tidyfst)

iris %>%
  mutate_dt(group = Species,sl = Sepal.Length,sw = Sepal.Width) %>%
  select_dt(group,sl,sw) %>%
  filter_dt(sl > 5) %>%
  arrange_dt(group,sl) %>%
  distinct_dt(sl,.keep_all = T) %>%
  summarise_dt(sw = max(sw),by = group)
#>         group  sw
#>        <fctr> <num>
#> 1:     setosa 4.4
#> 2: versicolor 3.4
#> 3:  virginica 3.8

iris %>%
  count_dt(Species) %>%
  add_prop()
#>       Species     n      prop prop_label
#>        <fctr> <int>     <num>     <char>
#> 1:     setosa    50 0.3333333      33.3%
#> 2: versicolor    50 0.3333333      33.3%
#> 3:  virginica    50 0.3333333      33.3%

iris[3:8,] %>%
  mutate_when(Petal.Width == .2,
              one = 1,Sepal.Length=2)
#>    Sepal.Length Sepal.Width Petal.Length Petal.Width Species one
#>          <num>       <num>        <num>       <num>  <fctr> <num>
#> 1:          2.0         3.2          1.3         0.2  setosa   1
#> 2:          2.0         3.1          1.5         0.2  setosa   1
#> 3:          2.0         3.6          1.4         0.2  setosa   1
#> 4:          5.4         3.9          1.7         0.4  setosa  NA
#> 5:          4.6         3.4          1.4         0.3  setosa  NA
#> 6:          2.0         3.4          1.5         0.2  setosa   1


```



## Future plans

*tidyfst* will keep up with the [updates](https://github.com/Rdatatable/data.table/blob/master/NEWS.md) of *data.table* , in the next step would introduce more new features to improve the performance and flexibility to facilitate fast data manipulation in tidy syntax. 



## Vignettes
- [Example 1: Basic usage](https://hope-data-science.github.io/tidyfst/articles/example1_intro.html)
- [Example 2: Join tables](https://hope-data-science.github.io/tidyfst/articles/example2_join.html)
- [Example 3: Reshape](https://hope-data-science.github.io/tidyfst/articles/example3_reshape.html)
- [Example 4: Nest](https://hope-data-science.github.io/tidyfst/articles/example4_nest.html)
- [Example 5: Fst](https://hope-data-science.github.io/tidyfst/articles/example5_fst.html) 
- [Example 6: Dt](https://hope-data-science.github.io/tidyfst/articles/example6_dt.html) 

## Cheat sheet

<a href="https://github.com/hope-data-science/tidyfst/blob/master/docs/tidyfst_cheatsheet.pdf"><img src="tidyfst_cheatsheet.png"/></a>

## Suggested citation

Huang et al., (2020). tidyfst: Tidy Verbs for Fast Data Manipulation. Journal of Open Source Software, 5(52), 2388, https://doi.org/10.21105/joss.02388



## Related work

- [data.table](https://github.com/Rdatatable/data.table)
- [fst](https://github.com/fstpackage/fst)
- [tidyr](https://github.com/tidyverse/tidyr)
- [dplyr](https://github.com/tidyverse/dplyr)
- [dtplyr](https://github.com/tidyverse/dtplyr)



## Acknowledgement

The author of [maditr](https://github.com/gdemin/maditr), [Gregory Demin](https://github.com/gdemin) and the author of [fst](https://github.com/fstpackage/fst), [Marcus Klik](https://github.com/MarcusKlik) have helped me a lot in the development of this work. It is so lucky to have them (and many other selfless contributors) in the same open source community of R.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  warning = FALSE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
# load packages
library(tidyfst)
library(data.table)
library(dplyr)
library(bench)

# generate the data
# if you have a HPC and want to try larger data sets, increase N
N = 1e4 
K = 1e2

set.seed(2020)

cat(sprintf("Producing data of %s rows and %s K groups factors\n", N, K))

DT = data.table(

  id1 = sample(sprintf("id%03d",1:K), N, TRUE),      # large groups (char)

  id2 = sample(sprintf("id%03d",1:K), N, TRUE),      # large groups (char)

  id3 = sample(sprintf("id%010d",1:(N/K)), N, TRUE), # small groups (char)

  id4 = sample(K, N, TRUE),                          # large groups (int)

  id5 = sample(K, N, TRUE),                          # large groups (int)

  id6 = sample(N/K, N, TRUE),                        # small groups (int)

  v1 =  sample(5, N, TRUE),                          # int in range [1,5]

  v2 =  sample(5, N, TRUE),                          # int in range [1,5]

  v3 =  round(runif(N,max=100),4)                    # numeric e.g. 23.5749

)

object_size(DT)

## -----------------------------------------------------------------------------
bench::mark(
  data.table = DT[,.(median_v3 = median(v3),
                     sd_v3 = sd(v3)),
                  by = .(id4,id5)],
  tidyfst = DT %>%
    summarise_dt(
      by = c("id4", "id5"),
      median_v3 = median(v3),
      sd_v3 = sd(v3)
    ),
  dplyr = DT %>%
    group_by(id4,id5,.drop = TRUE) %>%
    summarise(median_v3 = median(v3),sd_v3 = sd(v3)),
  check = FALSE,iterations = 10
) -> q1

q1

## -----------------------------------------------------------------------------
bench::mark(
  data.table =DT[,.(range_v1_v2 = max(v1) - min(v2)),by = id3],
  tidyfst = DT %>% summarise_dt(
    by = id3,
    range_v1_v2 = max(v1) - min(v2)
  ),
  dplyr = DT %>%
    group_by(id3,.drop = TRUE) %>%
    summarise(range_v1_v2 = max(v1) - min(v2)),
  check = FALSE,iterations = 10
) -> q2

q2

## -----------------------------------------------------------------------------
bench::mark(
  data.table =DT[order(-v3),.(largest2_v3 = head(v3,2L)),by = id6],
  tidyfst = DT %>%
    in_dt(order(-v3),.(largest2_v3 = head(v3,2L)),by = id6),
  dplyr = DT %>%
    select(id6,largest2_v3 = v3) %>%
    group_by(id6) %>%
    slice_max(largest2_v3,n = 2,with_ties = FALSE),
  check = FALSE,iterations = 10
) -> q3

q3

## -----------------------------------------------------------------------------
bench::mark(
  data.table =DT[,lapply(.SD,mean),by = id4,.SDcols = v1:v3],
  tidyfst = DT %>%
    summarise_vars(
      v1:v3,
      mean,
      by = id4
    ),
  dplyr = DT %>%
    group_by(id4) %>%
    summarise(across(v1:v3,mean)),
  check = FALSE,iterations = 10
) -> q4

q4

## -----------------------------------------------------------------------------
bench::mark(
  data.table =DT[,.(v3 = sum(v3),count = .N),by = id1:id6],
  tidyfst = DT %>%
    summarise_dt(
      by = id1:id6,
      v3 = sum(v3),
      count = .N
    ),
  dplyr = DT %>%
    group_by(id1,id2,id3,id4,id5,id6) %>%
    summarise(v3 = sum(v3),count = n()),
  check = FALSE,iterations = 10
) -> q5

q5

## -----------------------------------------------------------------------------
sessionInfo()

#' @title Bind multiple data frames by row
#' @description
#' Bind any number of data frames by row, making a longer result.
#' Similar to `dplyr::bind_rows`, however, columns with same names but different data types
#' would be coerced to a single proper data type.
#' @param ... Data frames to combine. Each argument can either be a data frame,
#' a list that could be a data frame, or a list of data frames.
#' Columns are matched by name, and any missing columns will be filled with `NA`.
#' @return data.table
#' @seealso \code{\link[dplyr]{bind_rows}},\code{\link[data.table]{rbindlist}}
#'
#' @examples
#'
#' bind_rows_dt(iris[1:3,],iris[6:8,])
#'
#' # data frames with same name but different type
#' # numeric data would be coerced to character data in this case
#' df1 <- data.frame(x = 1:2, y = letters[1:2])
#' df2 <- data.frame(x = 4:5, y = 1:2)
#' bind_rows_dt(df1, df2)
#'

#' @export

bind_rows_dt = function(...){
  rbindlist(lapply(list(...),as.data.table))
}

#' @title Select distinct/unique rows in data.frame
#' @description Select only unique/distinct rows from a data frame.
#' @param .data data.frame
#' @param ... Optional variables to use when determining uniqueness.
#'  If there are multiple rows for a given combination of inputs,
#'  only the first row will be preserved.
#'  If omitted, will use all variables.
#' @param .keep_all If \code{TRUE}, keep all variables in data.frame. If a combination of ... is not distinct,
#' this keeps the first row of values.
#' @param fromLast Logical indicating if duplication should be
#' considered from the reverse side. Defaults to \code{FALSE}.
#' @return data.table
#' @seealso \code{\link[dplyr]{distinct}}
#' @examples
#' iris %>% distinct_dt()
#' iris %>% distinct_dt(Species)
#' iris %>% distinct_dt(Species,.keep_all = TRUE)
#' mtcars %>% distinct_dt(cyl,vs)
#' mtcars %>% distinct_dt(cyl,vs,.keep_all = TRUE)
#' mtcars %>% distinct_dt(cyl,vs,.keep_all = TRUE,fromLast = TRUE)
#'
#'
#' @export

distinct_dt = function(.data,...,.keep_all = FALSE,fromLast = FALSE){
  dt = as_dt(.data)
  sel_name = select_dt(dt[0],...) %>% names()
  if(.keep_all) unique(dt,by = sel_name,fromLast = fromLast)
  else unique(dt[,.SD,.SDcols = sel_name],fromLast = fromLast)
}

#' @title Group by variable(s) and implement operations
#' @description Carry out data manipulation within specified groups. Different from \code{group_dt},
#' the implementation is split into two operations, namely grouping and implementation.
#' @description Using \code{setkey} and \code{setkeyv} in \pkg{data.table}
#' to carry out \code{group_by}-like functionalities in \pkg{dplyr}. This is
#' not only convenient but also efficient in computation.
#' @param .data A data frame
#' @param ... Variables to group by for \code{group_by_dt},
#'  namely the columns to sort by. Do not quote the column names.
#'  Any data manipulation arguments that could be
#'  implemented on a data.frame for \code{group_exe_dt}.
#'  It can receive what \code{select_dt} receives.
#' @param cols A character vector of column names to group by.
#' @return A data.table with keys
#' @details \code{group_by_dt} and \code{group_exe_dt} are a pair of functions
#' to be used in combination. It utilizes the feature of key setting in data.table,
#' which provides high performance for group operations, especially when you have
#' to operate by specific groups frequently.
#' @examples
#'
#' # aggregation after grouping using group_exe_dt
#' as.data.table(iris) -> a
#' a %>%
#'   group_by_dt(Species) %>%
#'   group_exe_dt(head(1))
#'
#' a %>%
#'   group_by_dt(Species) %>%
#'   group_exe_dt(
#'     head(3) %>%
#'       summarise_dt(sum = sum(Sepal.Length))
#'   )
#'
#' mtcars %>%
#'   group_by_dt("cyl|am") %>%
#'   group_exe_dt(
#'     summarise_dt(mpg_sum = sum(mpg))
#'   )
#' # equals to
#' mtcars %>%
#'   group_by_dt(cols = c("cyl","am")) %>%
#'   group_exe_dt(
#'     summarise_dt(mpg_sum = sum(mpg))
#'   )

#' @rdname group_by
#' @export

group_by_dt = function(.data,...,cols = NULL){
  #dt = as_dt(.data)
  .data = as.data.table(.data)

  if(!is.null(cols)) setkeyv(.data,cols)
  else {
    .data[0] %>% select_dt(...) %>% names() -> cols
    setkeyv(.data,cols)
  }

  .data
}

#' @rdname group_by
#' @export
group_exe_dt = function(.data,...){
  dt_keys = key(.data)
  if(is.null(dt_keys))
    stop("`group_exe_dt` must receive a data.table with key(s).")
  else
    dt_keys = str_c(dt_keys,collapse = ",") %>%
      str_c(".(",.,")")

  dt = as_dt(.data)
  eval(parse(text = str_glue("group_dt(.data = dt,by = {dt_keys},...)")))
}

#' @title Set operations for data frames
#' @description Wrappers of set operations in \pkg{data.table}.
#' Only difference is it could be applied to non-data.table data frames by
#' recognizing and coercing them to data.table automatically.
#' @param x A data.frame
#' @param y A data.frame
#' @param all Logical. When \code{FALSE} (default),
#'  removes duplicate rows on the result.
#' @return A data.table
#' @seealso \code{\link[data.table]{setops}}
#' @examples
#'
#' x = iris[c(2,3,3,4),]
#' x2 = iris[2:4,]
#' y = iris[c(3:5),]
#'
#' intersect_dt(x, y)            # intersect
#' intersect_dt(x, y, all=TRUE)  # intersect all
#' setdiff_dt(x, y)              # except
#' setdiff_dt(x, y, all=TRUE)    # except all
#' union_dt(x, y)                # union
#' union_dt(x, y, all=TRUE)      # union all
#' setequal_dt(x, x2, all=FALSE) # setequal
#' setequal_dt(x, x2)            # setequal all
#'

#' @rdname setops
#' @export

intersect_dt = function(x,y,all = FALSE){
  x = as_dt(x)
  y = as_dt(y)
  fintersect(x,y,all)
}

#' @rdname setops
#' @export

union_dt = function(x,y,all = FALSE){
  x = as_dt(x)
  y = as_dt(y)
  funion(x,y,all)
}

#' @rdname setops
#' @export

setdiff_dt = function(x,y,all = FALSE){
  x = as_dt(x)
  y = as_dt(y)
  fsetdiff(x,y,all)
}

#' @rdname setops
#' @export

setequal_dt = function(x,y,all = TRUE){
  x = as_dt(x)
  y = as_dt(y)
  fsetequal(x,y,all)
}

#' @title Read and write fst files
#' @description Wrapper for \code{\link[fst]{read_fst}} and \code{\link[fst]{write_fst}}
#' from \pkg{fst}, but use a different default. For data import, always return a data.table.
#' For data export, always compress the data to the smallest size.
#' @param x a data frame to write to disk
#' @param path path to fst file
#' @param compress value in the range 0 to 100, indicating the amount of compression to use.
#' Lower values mean larger file sizes. The default compression is set to 50.
#' @param uniform_encoding If `TRUE`, all character vectors will be assumed to have elements with equal encoding.
#' The encoding (latin1, UTF8 or native) of the first non-NA element will used as encoding for the whole column.
#' This will be a correct assumption for most use cases.
#' If `uniform.encoding` is set to `FALSE`, no such assumption will be made and all elements will be converted
#' to the same encoding. The latter is a relatively expensive operation and will reduce write performance for
#' character columns.
#' @param columns Column names to read. The default is to read all columns.
#' @param from Read data starting from this row number.
#' @param to Read data up until this row number. The default is to read to the last row of the stored dataset.
#' @param as.data.table If TRUE, the result will be returned as a \code{data.table} object. Any keys set on
#' dataset \code{x} before writing will be retained. This allows for storage of sorted datasets. This option
#' requires \code{data.table} package to be installed.
#' @param old_format must be FALSE, the old fst file format is deprecated and can only be read and
#' converted with fst package versions 0.8.0 to 0.8.10.
#' @return `import_fst` returns a data.table with the selected columns and rows. `export_fst`
#' writes `x` to a `fst` file and invisibly returns `x` (so you can use this function in a pipeline).
#' @seealso \code{\link[fst]{read_fst}}
#' @examples
#' \dontrun{
#' export_fst(iris,"iris_fst_test.fst")
#' iris_dt = import_fst("iris_fst_test.fst")
#' iris_dt
#' unlink("iris_fst_test.fst")
#' }

#' @rdname fst_io
#' @export
export_fst = function(x, path, compress = 100, uniform_encoding = TRUE){
  write_fst(x, path, compress, uniform_encoding)
}

#' @rdname fst_io
#' @export
import_fst = function(path, columns = NULL, from = 1, to = NULL,
                      as.data.table = TRUE, old_format = FALSE){
  read_fst(path, columns, from, to,
           as.data.table, old_format)
}

#' @title Fast creation of dummy variables
#' @description Quickly create dummy (binary) columns from character and factor type columns in the inputted data (and numeric columns if specified.)
#' This function is useful for statistical analysis when you want binary columns rather than character columns.
#' @param .data data.frame
#' @param ... Columns you want to create dummy variables from.
#' Very flexible, find in the examples.
#' @param longname logical. Should the output column labeled with the
#' original column name? Default uses \code{TRUE}.
#' @return data.table
#' @details If no columns provided, will return the original data frame.
#'  When NA exist in the input column, they would also be considered. If
#'  the input character column contains both NA and string "NA", they would be
#'  merged.
#' @details This function is inspired by \pkg{fastDummies} package, but provides
#' simple and precise usage, whereas \code{fastDummies::dummy_cols} provides more
#' features for statistical usage.
#' @references https://stackoverflow.com/questions/18881073/creating-dummy-variables-in-r-data-table
#' @seealso \code{\link[fastDummies]{dummy_cols}}
#' @examples
#' iris %>% dummy_dt(Species)
#' iris %>% dummy_dt(Species,longname = FALSE)
#'
#' mtcars %>% head() %>% dummy_dt(vs,am)
#' mtcars %>% head() %>% dummy_dt("cyl|gear")
#'
#' # when there are NAs in the column
#' df <- data.table(x = c("a", "b", NA, NA),y = 1:4)
#' df %>%
#'   dummy_dt(x)
#'
#' # when NA  and "NA" both exist, they would be merged
#' df <- data.table(x = c("a", "b", NA, "NA"),y = 1:4)
#' df %>%
#'   dummy_dt(x)
#'

#' @export
dummy_dt = function(.data,...,longname = TRUE){
  # dt = as_dt(.data)
  dt = as.data.table(.data)
  if((substitute(list(...)) %>% deparse())=="list()")
    warning("No columns provided, return the orginal data.")
  else{
    dt[0] %>%
      select_dt(...) %>%
      names() -> col_names
    for(i in col_names){
      dummy_col(dt,i,longname = longname) -> dt
    }
  }
  dt[]
}

dummy_col = function(dt,col_name,longname){

  dt[is.na(dt[[col_name]]),(col_name) := "NA"]
  dt[[col_name]] %>% unique() %>% as.character()-> old_names

  if(longname) new_names = str_c(col_name,old_names,sep="_")
  else new_names = old_names

  dt[,(old_names):=0]

  for (current_value in old_names) {
    set(dt,i = which(chmatch(as.character(dt[[col_name]]),current_value,nomatch = 0) == 1L),j = current_value,value = 1L)
  }

  setnames(dt,old = old_names,new = new_names)[]
}

# # this version is slow
# dummy_col = function(dt,col_name,longname){
#
#   dt[[col_name]] %>% unique() -> old_values
#   if(is.character(old_values) & anyNA(old_values) & "NA" %in% old_values){
#     dt[is.na(dt[[col_name]]),(col_name):="NA"]
#     dt[[col_name]] %>% unique() -> old_values
#   }
#
#   dt[[col_name]] %>% unique() %>% as.character()-> old_names
#   old_names[is.na(old_values)] = "NA"
#   if(!longname){
#     dt[,(old_names):=lapply(old_values,function(x) {
#       sapply(as.list(dt[[col_name]]),FUN = identical,x)
#     })][
#       ,(old_names):=lapply(.SD,as.numeric),.SDcols = old_names
#     ][,(col_name):=NULL][]
#   }else{
#     str_c(col_name,old_names,sep="_") -> new_names
#     dt[,(old_names):=lapply(old_values,function(x) {
#       sapply(as.list(dt[[col_name]]),FUN = identical,x)
#     })]
#     setnames(dt,old = old_names,new = new_names)[
#       ,(new_names):=lapply(.SD,as.numeric),.SDcols = new_names
#     ][,(col_name):=NULL][]
#   }
# }

# # for this version, when there are NAs, yields an error
# dummy_col = function(dt,col_name,longname){
#
#   dt[[col_name]] %>% unique() %>% as.character()-> old_names
#   if(!longname){
#     dt[,(old_names):=lapply(old_names,function(x) x == dt[[col_name]])][
#       ,(old_names):=lapply(.SD,as.numeric),.SDcols = old_names
#     ][,(col_name):=NULL][]
#   }else{
#     str_c(col_name,old_names,sep="_") -> new_names
#     dt[,(old_names):=lapply(old_names,function(x) x == dt[[col_name]])]
#     setnames(dt,old = old_names,new = new_names)[
#       ,(new_names):=lapply(.SD,as.numeric),.SDcols = new_names
#     ][,(col_name):=NULL][]
#   }
# }

#' @title Separate a character column into two columns using
#' a regular expression separator
#' @description Given either regular expression,
#' \code{separate_dt()} turns a single character column into two columns.
#' @param .data A data frame.
#' @param separated_colname Column to be separated, can be a character or alias.
#' @param into Character vector of length 2.
#' @param sep Separator between columns.
#' @param remove If \code{TRUE}, remove input column from output data frame.
#' @seealso \code{\link[tidyr]{separate}}, \code{\link[tidyfst]{unite_dt}}
#' @examples
#' df <- data.frame(x = c(NA, "a.b", "a.d", "b.c"))
#' df %>% separate_dt(x, c("A", "B"))
#' # equals to
#' df %>% separate_dt("x", c("A", "B"))
#'
#' # If you just want the second variable:
#' df %>% separate_dt(x,into = c(NA,"B"))

#' @export
separate_dt = function(.data,separated_colname,into,
                       sep = "[^[:alnum:]]+",
                       remove = TRUE){
  dt = as.data.table(.data)
  substitute(separated_colname) %>% deparse() -> parse_name
  if(!str_detect(parse_name,"^\"")) separated_colname = parse_name

  if(anyNA(into)) into[is.na(into)] = "NA_COL_"

  dt[[separated_colname]] %>%
    tstrsplit(split = sep) %>%
    setDT() %>%
    setnames(names(.),into) %>%
    select_dt(-"NA_COL_")-> split_columns
  if(remove)
    dt[,(separated_colname):=NULL][,names(split_columns):=split_columns][]
  else dt[,names(split_columns):=split_columns][]

}

#' @title Save a data.frame as a fst table
#' @description This function first export the data.frame to a temporal file,
#' and then parse it back as a fst table (class name is "fst_table").
#' @param .data A data.frame
#' @return An object of class \code{fst_table}
#' @examples
#'
#' \dontrun{
#'   iris %>%
#'     as_fst() -> iris_fst
#'   iris_fst
#' }

#' @export

as_fst = function(.data){
  if(!is.data.frame(.data))
    stop("Only a data.frame could be stored as fst table.")

  substitute(.data) %>%
    deparse() %>%
    tempfile(fileext = ".fst") -> fn
  export_fst(.data,path = fn)
  parse_fst(fn)
}

#' @title Read a fst file by chunks
#' @description For `import_fst_chunked`, if a large fst file which could not be imported into the memory all at once,
#'  this function could read the fst file by chunks and preprocessed the chunk to ensure the
#'  results yielded by the chunks are small enough to be summarised in the end.
#'  For `get_fst_chunk_size`, this function can measure the memory used by a specified row number.
#' @param path Path to fst file
#' @param chunk_size Integer. The number of rows to include in each chunk
#' @param chunk_f A function implemented on every chunk.
#' @param combine_f A function to aggregate all the elements from the list of results from chunks.
#' @param nrows Number of rows to test.
#'
#' @return For `import_fst_chunked`, default to the whole data.frame in data.table.
#'  Could be adjusted to any type. For `get_fst_chunk_size`, return the file size.
#' @seealso \code{\link[readr]{read_csv_chunked}}
#' @examples
#'
#' \dontrun{
#'   # Generate some random data frame with 10 million rows and various column types
#'   nr_of_rows <- 1e7
#'   df <- data.frame(
#'     Logical = sample(c(TRUE, FALSE, NA), prob = c(0.85, 0.1, 0.05), nr_of_rows, replace = TRUE),
#'     Integer = sample(1L:100L, nr_of_rows, replace = TRUE),
#'     Real = sample(sample(1:10000, 20) / 100, nr_of_rows, replace = TRUE),
#'     Factor = as.factor(sample(labels(UScitiesD), nr_of_rows, replace = TRUE))
#'   )
#'
#'   # Write the file to disk
#'   fst_file <- tempfile(fileext = ".fst")
#'   write_fst(df, fst_file)
#'
#'   # Get the size of 10000 rows
#'   get_fst_chunk_size(fst_file,1e4)
#'
#'   # File all rows that Integer == 7 by chunks
#'   import_fst_chunked(fst_file,chunk_f = \(x) x[Integer==7])
#'
#' }
#'

#' @rdname import_fst_chunked
#' @export

import_fst_chunked = \(path,chunk_size = 10000L,chunk_f = identity,combine_f = rbindlist){
  parse_fst(path) %>% nrow() -> ft_nrow
  seq(from = 1,to = ft_nrow,by = chunk_size)-> start_index
  c(start_index[-1] - 1,ft_nrow) -> end_index
  Map(f = \(s,e){
    read_fst(path,from = s,to = e,as.data.table = TRUE) %>%
      chunk_f
  },start_index,end_index) %>% combine_f
}

#' @rdname import_fst_chunked
#' @export
get_fst_chunk_size = \(path,nrows){
  object_size(read_fst(path,from = 1,to = nrows))
}

#' @title Filter entries in data.frame
#' @description Choose rows where conditions are true.
#' @param .data data.frame
#' @param ... List of variables or name-value pairs of summary/modifications
#'   functions.
#' @return data.table
#' @seealso \code{\link[dplyr]{filter}}
#' @examples
#' iris %>% filter_dt(Sepal.Length > 7)
#' iris %>% filter_dt(Sepal.Length == max(Sepal.Length))
#'
#' # comma is not supported in tidyfst after v0.9.8
#' # which means you can't use:
#' \dontrun{
#'  iris %>% filter_dt(Sepal.Length > 7, Sepal.Width > 3)
#' }
#' # use following code instead
#' iris %>% filter_dt(Sepal.Length > 7 & Sepal.Width > 3)
#'

#' @export
filter_dt = function(.data,...){
  dt = as_dt(.data)
  eval.parent(dt[...])
}

#' @title "Uncount" a data frame
#' @description  Duplicating rows according to a weighting variable.
#'  This is the opposite operation of `count_dt`.
#'  Analogous to `tidyr::uncount`.
#' @param .data A data.frame
#' @param wt A vector of weights.
#' @param .remove Should the column for \code{weights} be removed?
#' Default uses \code{TRUE}.
#' @seealso \code{\link[dplyr]{count}}, \code{\link[tidyr]{uncount}}
#' @examples
#'
#' df <- data.table(x = c("a", "b"), n = c(1, 2))
#' uncount_dt(df, n)
#' uncount_dt(df,n,FALSE)
#' @export

uncount_dt = function(.data,wt,.remove = TRUE){
  dt = as_dt(.data)
  if(.remove) eval.parent(substitute(dt[rep(1:.N,wt)][,wt:=NULL][]))
  else eval.parent(substitute(dt[rep(1:.N,wt)]))
}

#' @title Sample rows randomly from a table
#' @description Select a number or proportion of rows randomly from the data frame
#' @param .data A data.frame
#' @param n Number of rows to select
#' @param prop Fraction of rows to select
#' @param size 	For \code{sample_n_dt}, the number of rows to select.
#' For \code{sample_frac_dt}, the fraction of rows to select.
#' @param replace Sample with or without replacement? Default uses \code{FALSE}.
#' @param by (Optional) Character. Specify if you want to sample by group.
#' @description \code{sample_dt} is a merged version of \code{sample_n_dt} and
#' \code{sample_frac_dt}, this could be convenient.
#' @return data.table
#' @seealso \code{\link[dplyr]{sample_n}},\code{\link[dplyr]{sample_frac}}
#' @examples
#' sample_n_dt(mtcars, 10)
#' sample_n_dt(mtcars, 50, replace = TRUE)
#' sample_frac_dt(mtcars, 0.1)
#' sample_frac_dt(mtcars, 1.5, replace = TRUE)
#'
#'
#' sample_dt(mtcars,n=10)
#' sample_dt(mtcars,prop = 0.1)
#'
#'
#' # sample by group(s)
#' iris %>% sample_n_dt(2,by = "Species")
#' iris %>% sample_frac_dt(.1,by = "Species")
#'
#' mtcars %>% sample_n_dt(1,by = c("cyl","vs"))


#' @rdname sample
#' @export
sample_dt = function(.data,n = NULL,prop = NULL,replace = FALSE,by = NULL){
  dt = as_dt(.data)
  if(is.null(n) & !is.null(prop))
    sample_frac_dt(dt,size = prop,replace = replace,by = by)
  else if(!is.null(n) & is.null(prop))
    sample_n_dt(dt,size = n,replace = replace,by = by)
  else stop("Both or none of `n` and `prop` are provided!")
}

#' @rdname sample
#' @export
#'
sample_n_dt = function(.data,size,replace = FALSE,by = NULL){
  dt = as_dt(.data)
  by_char = substitute(by) %>% deparse()
  if(by_char == "NULL"){
    if(size > nrow(.data) & replace == FALSE) stop("Sample size is too large.")
    index = sample(nrow(dt),size = size,replace = replace)
    dt[index]
  }else{
    dt[,.SD[sample(.N,size = size,replace = replace)],by]
  }
}



#' @rdname sample
#' @export
sample_frac_dt = function(.data,size,replace = FALSE,by = NULL){
  dt = as_dt(.data)
  if(!between(size,0,1) & replace == FALSE) stop("The fraction is invalid.")
  size = trunc(nrow(dt) * size)
  sample_n_dt(.data = .data,size = size,replace = replace,by = by)
}

#' @title Impute missing values with mean, median or mode
#' @description Impute the columns of data.frame with its mean, median or mode.
#' @param .data A data.frame
#' @param ... Columns to select
#' @param .func Character, "mode" (default), "mean" or "median".
#' Could also define it by oneself.
#' @return A data.table
#' @examples
#'
#' Pclass <- c(3, 1, 3, 1, 3, 2, 2, 3, NA, NA)
#' Sex <- c('male', 'male', 'female', 'female', 'female',
#'          'female', NA, 'male', 'female', NA)
#' Age <- c(22, 38, 26, 35, NA,
#'          45, 25, 39, 28, 40)
#' SibSp <- c(0, 1, 3, 1, 2, 3, 2, 2, NA, 0)
#' Fare <- c(7.25, 71.3, 7.92, NA, 8.05, 8.46, 51.9, 60, 32, 15)
#' Embarked <- c('S', NA, 'S', 'Q', 'Q', 'S', 'C', 'S', 'C', 'S')
#' data <- data.frame('Pclass' = Pclass,
#'  'Sex' = Sex, 'Age' = Age, 'SibSp' = SibSp,
#'  'Fare' = Fare, 'Embarked' = Embarked)
#'
#' data
#' data %>% impute_dt() # defalut uses "mode" as `.func`
#' data %>% impute_dt(is.numeric,.func = "mean")
#' data %>% impute_dt(is.numeric,.func = "median")
#'
#' my_fun = function(x){
#'   x[is.na(x)] = (max(x,na.rm = TRUE) - min(x,na.rm = TRUE))/2
#'   x
#' }
#' data %>% impute_dt(is.numeric,.func = my_fun)
#'

#' @export

impute_dt = function(.data,...,.func = "mode"){

  dt = as.data.table(.data)
  if (substitute(list(...)) %>% deparse() == "list()")
    sel_cols = names(dt)
  else
    dt[0] %>% select_dt(...) %>% names() -> sel_cols

  if(!is.function(.func)){
    if(.func == "mode")
      dt[,(sel_cols):=lapply(.SD,
                             function(x){
                               uniqv = unique(x)
                               x_i = uniqv[which.max(tabulate(match(x, uniqv)))]
                               x[which(is.na(x))] = x_i
                               x
                             }),
         .SDcols = sel_cols][]
    else if(.func == "mean"){
      dt[,(sel_cols):=lapply(.SD,
                             function(x){
                               x_i = mean(x,na.rm = TRUE)
                               x[which(is.na(x))] = x_i
                               x
                             }),
         .SDcols = sel_cols][]
    }else if(.func == "median"){
      dt[,(sel_cols):=lapply(.SD,
                             function(x){
                               x_i = median(x,na.rm = TRUE)
                               x[which(is.na(x))] = x_i
                               x
                             }),
         .SDcols = sel_cols][]
    }
  }else dt[,(sel_cols):=lapply(.SD,.func),.SDcols = sel_cols][]
}

#' @title Set global printing method for data.table
#' @description This function allow user to define how data.table is printed.
#' @param topn The number of rows to be printed from the beginning and
#'  end of tables with more than \code{nrow} rows.
#' @param nrows The number of rows which will be printed before truncation is enforced.
#' @param class If \code{TRUE}, the resulting output will include above each column its storage class (or a self-evident abbreviation thereof).
#' @param row.names If \code{TRUE}, row indices will be printed.
#' @param col.names One of three flavours for controlling the display of column names in output. \code{"auto"} includes column names above the data, as well as below the table if \code{nrow(x) > 20}. \code{"top"} excludes this lower register when applicable, and \code{"none"} suppresses column names altogether (as well as column classes if \code{class = TRUE}.
#' @param print.keys If \code{TRUE}, any \code{\link{key}} currently assigned to \code{x} will be printed prior to the preview of the data.
#' @param trunc.cols  If \code{TRUE}, only the columns that can be printed in the console without wrapping the columns to new lines will be printed (similar to \code{tibbles}).
#' @seealso \code{\link[data.table]{print.data.table}}
#' @details Notice that \pkg{tidyfst} has a slightly different printing default for data.table,
#'  which is it always prints the keys and variable class (not like \pkg{data.table}).
#' @return None. This function is used for its side effect of changing options.
#' @examples
#'
#' iris %>% as.data.table()
#' print_options(topn = 3,trunc.cols = TRUE)
#' iris %>% as.data.table()
#'
#' # set all settings to default in tidyfst
#' print_options()
#' iris %>% as.data.table()
#'

#' @export
print_options = function(
  topn=5,
  nrows=100,
  class=TRUE,
  row.names=TRUE,
  col.names="auto",
  print.keys=TRUE,
  trunc.cols=FALSE
){
  options(
    "datatable.print.topn" = topn,
    "datatable.print.nrows" = nrows,
    "datatable.print.class" = class,
    "datatable.print.rownames" = row.names,
    "datatable.print.colnames" = col.names,
    "datatable.print.keys" = print.keys,
    "datatable.print.trunc.cols" = trunc.cols
  )
}

# default in data.able
# print_options = function(
#   topn=getOption("datatable.print.topn"),             # default: 5
#   nrows=getOption("datatable.print.nrows"),           # default: 100
#   class=getOption("datatable.print.class"),           # default: FALSE
#   row.names=getOption("datatable.print.rownames"),    # default: TRUE
#   col.names=getOption("datatable.print.colnames"),    # default: "auto"
#   print.keys=getOption("datatable.print.keys"),       # default: FALSE
#   trunc.cols=getOption("datatable.print.trunc.cols") # default: FALSE
# ){
#   options(
#     "datatable.print.topn" = topn,
#     "datatable.print.nrows" = nrows,
#     "datatable.print.class" = class,
#     "datatable.print.rownames" = row.names,
#     "datatable.print.colnames" = col.names,
#     "datatable.print.keys" = print.keys,
#     "datatable.print.trunc.cols" = trunc.cols
#   )
# }

#' @title Data manipulation within groups
#' @description Carry out data manipulation within specified groups.
#' @param .data A data.frame
#' @param by Variables to group by,unquoted name of grouping variable of list of unquoted names of grouping variables.
#' @param ... Any data manipulation arguments that could be implemented on a data.frame.
#' @return data.table
#' @details If you want to use \code{summarise_dt} and \code{mutate_dt} in
#' \code{group_dt}, it is better to use the "by" parameter in those functions,
#' that would be much faster because you don't have to use \code{.SD} (which takes
#' extra time to copy).
#' @references https://stackoverflow.com/questions/36802385/use-by-each-row-for-data-table
#' @examples
#' iris %>% group_dt(by = Species,slice_dt(1:2))
#' iris %>% group_dt(Species,filter_dt(Sepal.Length == max(Sepal.Length)))
#' iris %>% group_dt(Species,summarise_dt(new = max(Sepal.Length)))
#'
#' # you can pipe in the `group_dt`
#' iris %>% group_dt(Species,
#'                   mutate_dt(max= max(Sepal.Length)) %>%
#'                     summarise_dt(sum=sum(Sepal.Length)))
#'
#' # for users familiar with data.table, you can work on .SD directly
#' # following codes get the first and last row from each group
#' iris %>%
#'   group_dt(
#'     by = Species,
#'     rbind(.SD[1],.SD[.N])
#'   )
#'
#' #' # for summarise_dt, you can use "by" to calculate within the group
#' mtcars %>%
#'   summarise_dt(
#'    disp = mean(disp),
#'    hp = mean(hp),
#'    by = cyl
#' )
#'
#'   # but you could also, of course, use group_dt
#'  mtcars %>%
#'    group_dt(by =.(vs,am),
#'      summarise_dt(avg = mean(mpg)))
#'
#'   # and list of variables could also be used
#'  mtcars %>%
#'    group_dt(by =list(vs,am),
#'             summarise_dt(avg = mean(mpg)))
#'
#' # examples for `rowwise_dt`
#' df <- data.table(x = 1:2, y = 3:4, z = 4:5)
#'
#' df %>% mutate_dt(m = mean(c(x, y, z)))
#'
#' df %>% rowwise_dt(
#'   mutate_dt(m = mean(c(x, y, z)))
#' )


#' @rdname group_dt
#' @export

group_dt = function(.data,by = NULL,...){
  #dt = as_dt(.data)
  dt = as.data.table(.data)

  by = substitute(by)
  deparse(by) -> by_deparse
  if(by_deparse == "NULL") stop("Please provide the group(s).")
  else if(!str_detect(by_deparse,"^\\.|^list\\("))
    by_deparse %>%
      str_c(".(",.,")") -> by_deparse

  substitute(list(...)) %>%
    deparse() %>%
    str_c(collapse = "") %>%
    str_squish() %>%
    str_extract("\\(.+\\)") %>%
    str_sub(2,-2) -> dot_string

  if(str_detect(dot_string,"\\.SD"))
    to_eval = "dt[,{dot_string},by = {by_deparse}]"
  else
    dot_string %>%
    strsplit("%>%") %>%
    unlist() %>%
    str_squish() %>%
    lapply(dot_convert) %>%
    str_c("[,",.,",","by = {by_deparse}]") %>%
    str_c(collapse = "") %>%
    str_c("dt",.) -> to_eval

  eval(parse(text = str_glue(to_eval)))
}


#' @rdname group_dt
#' @export
rowwise_dt = function(.data,...){
  dt = as_dt(.data)
  dt %>%
    group_dt(
      by = row.names(dt),
      ...
    ) %>%
    select_dt(-row.names)
}



dot_convert = function(string){
  if(str_detect(string,",\\s*\\.\\s*,"))
    str_replace(string,",\\s*\\.\\s*,",",.SD,") -> string
  else if(str_detect(string,",s*\\.s*\\)"))
    str_replace(string,",s*\\.s*\\)",",.SD\\)") -> string
  else str_replace(string,"\\(","\\(.SD,") -> string
  string
}

#' @title Unite multiple columns into one by pasting strings together
#' @description Convenience function to paste together multiple columns into one.
#' @param .data A data frame.
#' @param united_colname The name of the new column, string only.
#' @param ... A selection of columns. If want to select all columns,
#' pass "" to the parameter. See example.
#' @param sep Separator to use between values.
#' @param remove If \code{TRUE}, remove input columns from output data frame.
#' @param na2char If \code{FALSE}, missing values would be merged into \code{NA},
#' otherwise \code{NA} is treated as character "NA". This is different from
#' \pkg{tidyr}.
#' @seealso \code{\link[tidyr]{unite}},\code{\link[tidyfst]{separate_dt}}
#' @examples
#' df <- expand.grid(x = c("a", NA), y = c("b", NA))
#' df
#'
#' # Treat missing value as NA, default
#' df %>% unite_dt("z", x:y, remove = FALSE)
#' # Treat missing value as character "NA"
#' df %>% unite_dt("z", x:y, na2char = TRUE, remove = FALSE)
#' df %>%
#'   unite_dt("xy", x:y)
#'
#' # Select all columns
#' iris %>% unite_dt("merged_name",".")

#' @export
unite_dt = function(.data,united_colname,...,
                    sep = "_",remove = FALSE,
                    na2char = FALSE){
  #dt = as_dt(.data)
  dt = as.data.table(.data)

  dt %>%
    select_dt(...) -> selected_cols
  paste_fun = ifelse(na2char,paste,str_c)
  Reduce(function(x,y) paste_fun(x,y,sep = sep),selected_cols) -> new_col

  if(!remove) dt[,(united_colname):=new_col][]
  else
    dt[,names(selected_cols):=NULL][,(united_colname):=new_col][]

}

#' @title Count observations by group
#' @description Count the unique values of one or more variables.
#' @param .data data.table/data.frame data.frame will be automatically converted
#'   to data.table.
#' @param ... Variables to group by, could receive what `select_dt` receives.
#' @param sort logical. If TRUE result will be sorted in desending order by resulting variable.
#' @param .name character. Name of resulting variable. Default uses "n".
#'
#' @return data.table
#' @seealso \code{\link[dplyr]{count}}
#' @examples
#' iris %>% count_dt(Species)
#' iris %>% count_dt(Species,.name = "count")
#' iris %>% add_count_dt(Species)
#' iris %>% add_count_dt(Species,.name = "N")
#'
#' mtcars %>% count_dt(cyl,vs)
#' mtcars %>% count_dt("cyl|vs")
#' mtcars %>% count_dt(cyl,vs,.name = "N",sort = FALSE)
#' mtcars %>% add_count_dt(cyl,vs)
#' mtcars %>% add_count_dt("cyl|vs")
#'
#' @rdname count
#' @export

count_dt = function(.data,...,sort = TRUE,.name = "n"){
  dt = as_dt(.data)
  dt[0] %>% select_dt(...) %>% names() -> dot_string
  if(sort)
    eval(parse(text =
                 str_glue("dt[,.({.name} = .N),by = dot_string][order(-{.name})]")))
  else
    eval(parse(text = str_glue("dt[,.({.name} = .N),by = dot_string]")))
}

#' @rdname count
#' @export
add_count_dt = function(.data,...,.name = "n"){
  dt = as.data.table(.data)
  dt[0] %>% select_dt(...) %>% names() -> dot_string
  dt[,(.name):=.N,by = dot_string][]
}

#' @title Case insensitive table joining like SQL
#' @name sql_join
#' @description Work like the `*_join_dt` series functions, joining
#' tables with common or customized keys  in various ways. The only
#' difference is the joining is case insensitive like SQL.
#' @param x A data.table
#' @param y A data.table
#' @param by (Optional) A character vector of variables to join by.
#'
#'   If `NULL`, the default, `*_join_dt()` will perform a natural join, using all
#'   variables in common across `x` and `y`. A message lists the variables so that you
#'   can check they're correct; suppress the message by supplying `by` explicitly.
#'
#'   To join by different variables on `x` and `y`, use a named vector.
#'   For example, `by = c("a" = "b")` will match `x$a` to `y$b`.
#'
#'   To join by multiple variables, use a vector with length > 1.
#'   For example, `by = c("a", "b")` will match `x$a` to `y$a` and `x$b` to
#'   `y$b`. Use a named vector to match different variables in `x` and `y`.
#'   For example, `by = c("a" = "b", "c" = "d")` will match `x$a` to `y$b` and
#'   `x$c` to `y$d`.
#'
#'   Notice that in `sql_join`, the joining variables would turn to upper case
#'   in the output table.
#' @param type Which type of join would you like to use?
#'  Default uses "inner", other options include
#'  "left", "right", "full", "anti", "semi".
#' @param suffix If there are non-joined duplicate variables in x and y, these
#'   suffixes will be added to the output to disambiguate them. Should be a
#'   character vector of length 2.
#' @return A data.table
#' @seealso \code{\link[tidyfst]{join}}
#' @examples
#' dt1 = data.table(x = c("A","b"),y = 1:2)
#' dt2 = data.table(x = c("a","B"),z = 4:5)
#' sql_join_dt(dt1,dt2)


#' @rdname sql_join
#' @export
sql_join_dt = function(x,y,by = NULL,type = "inner",suffix = c(".x",".y")){

  x = as.data.table(x)
  y = as.data.table(y)
  by_ = substitute(by) %>% deparse()
  if(by_ == "NULL"){
    by = intersect(names(x), names(y))
    x = x[,(by) := lapply(.SD,toupper),.SDcols = by]
    y = y[,(by) := lapply(.SD,toupper),.SDcols = by]
  }  else{
    x = x[,(names(by)) := lapply(.SD,toupper),.SDcols = names(by)]
    y = y[,(by) := lapply(.SD,toupper),.SDcols = by]
  }
  if(type %in% c("anti","semi"))
    get(str_glue("{type}_join_dt"))(x,y,by = by)
  else
    get(str_glue("{type}_join_dt"))(x,y,by = by,suffix = suffix)
}

#' @title Efficient transpose of data.frame
#' @description An efficient way to transpose data frames(data.frame/data.table/tibble).
#' @param .data A data.frame/data.table/tibble
#' @return A transposed data.frame
#' @details This function would return the original data.frame structure,
#' keeping all the row names and column names. If the row names are not
#' available or, "V1,V2..." will be provided.
#' @examples
#'
#' t_dt(iris)
#' t_dt(mtcars)

#' @export
t_dt = function(.data){
  dt = transpose(.data) %>% setDF()
  rownames(dt) = colnames(.data)
  if(setequal(rownames(.data),as.character(1:nrow(.data))))
    colnames(dt) = paste0("V",rownames(.data))
  else
    colnames(dt) = rownames(.data)
  dt
}

# t_dt = function(.data){
#   dt = as_dt(.data)
#   dt = transpose(dt) %>% as.data.frame()
#   rownames(dt) = colnames(.data)
#   if(setequal(rownames(.data),as.character(1:nrow(.data))))
#     colnames(dt) = paste0("V",rownames(.data))
#   else
#     colnames(dt) = rownames(.data)
#   dt
# }

#' @title Cumulative mean
#' @description Returns a vector whose elements are the cumulative mean of the elements of the argument.
#' @param x a numeric or complex object,
#'  or an object that can be coerced to one of these.
#' @examples
#' cummean(1:10)
#'
#' @export

cummean = function(x) cumsum(x)/seq_along(x)

#' @title Nice printing of report the Space Allocated for an Object
#' @description Provides an estimate of the memory that is being used to store an R object.
#' A wrapper of `object.size`, but use a nicer printing unit.
#' @param object an R object.
#' @return An object of class "object_size"
#' @examples
#'
#' iris %>% object_size()
#'

#' @export
object_size = function(object){
  object.size(object) %>% print(unit = "auto")
}

#' @title Tools for working with row names
#' @description The enhanced data.frame, including tibble and data.table, do not
#' support row names. To link to some base r facilities, there should be functions
#' to save information in row names. These functions are analogous to
#' \code{rownames_to_column} and \code{column_to_rownames} in \pkg{tibble}.
#' @param .data A data.frame.
#' @param var Name of column to use for rownames.
#' @return \code{rn_col} returns a data.table,
#' \code{col_rn} returns a data frame.
#' @examples
#'
#'  mtcars %>% rn_col()
#'  mtcars %>% rn_col("rn")
#'
#'  mtcars %>% rn_col() -> new_mtcars
#'
#'  new_mtcars %>% col_rn() -> old_mtcars
#'  old_mtcars
#'  setequal(mtcars,old_mtcars)

#' @rdname rownames
#' @export

rn_col = function(.data,var = "rowname"){
  dt = as_dt(.data)
  eval(parse(text = str_glue("dt[,{var}:=rownames(.data)] %>% relocate_dt({var})")))
}

#' @rdname rownames
#' @export

col_rn = function(.data,var = "rowname"){
  # dt = as_dt(.data)
  dt = as.data.table(.data)
  setDF(dt[,(var):=NULL])
  rownames(dt) = .data[[var]]
  dt
}

#' @title Dump, replace and fill missing values in data.frame
#' @description A set of tools to deal with missing values in data.frames. 
#' It can dump, replace, fill (with next or previous observation) or delete entries according to their missing values.
#'
#' @param .data data.frame
#' @param ... Colunms to be replaced or filled. If not specified, use all columns.
#' @param prop If proportion of NAs is larger than or equal to "prop", would be deleted.
#' @param n If number of NAs is larger than or equal to "n", would be deleted.
#' @param to What value should NA replace by?
#' @param direction Direction in which to fill missing values.
#' Currently either "down" (the default) or "up".
#' @param x A vector with missing values to be filled.
#' @return data.table
#' @details \code{drop_na_dt} drops the entries with NAs in specific columns.
#' \code{fill_na_dt} fill NAs with observations ahead ("down") or below ("up"),
#' which is also known as last observation carried forward (LOCF) and
#' next observation carried backward(NOCB).
#' @details \code{delete_na_cols} could drop the columns with NA proportion larger
#' than or equal to "prop" or NA number larger than or equal to "n",
#' \code{delete_na_rows} works alike but deals with rows.
#' @details \code{shift_fill} could fill a vector with missing values.
#' @references https://stackoverflow.com/questions/23597140/how-to-find-the-percentage-of-nas-in-a-data-frame
#' @references https://stackoverflow.com/questions/2643939/remove-columns-from-dataframe-where-all-values-are-na
#' @references https://stackoverflow.com/questions/7235657/fastest-way-to-replace-nas-in-a-large-data-table
#' @seealso \code{\link[tidyr]{drop_na}},\code{\link[tidyr]{replace_na}},
#' \code{\link[tidyr]{fill}}
#' @examples
#' df <- data.table(x = c(1, 2, NA), y = c("a", NA, "b"))
#'  df %>% drop_na_dt()
#'  df %>% drop_na_dt(x)
#'  df %>% drop_na_dt(y)
#'  df %>% drop_na_dt(x,y)
#'
#'  df %>% replace_na_dt(to = 0)
#'  df %>% replace_na_dt(x,to = 0)
#'  df %>% replace_na_dt(y,to = 0)
#'  df %>% replace_na_dt(x,y,to = 0)
#'
#'  df %>% fill_na_dt(x)
#'  df %>% fill_na_dt() # not specified, fill all columns
#'  df %>% fill_na_dt(y,direction = "up")
#'
#' x = data.frame(x = c(1, 2, NA, 3), y = c(NA, NA, 4, 5),z = rep(NA,4))
#' x
#' x %>% delete_na_cols()
#' x %>% delete_na_cols(prop = 0.75)
#' x %>% delete_na_cols(prop = 0.5)
#' x %>% delete_na_cols(prop = 0.24)
#' x %>% delete_na_cols(n = 2)
#'
#' x %>% delete_na_rows(prop = 0.6)
#' x %>% delete_na_rows(n = 2)
#'
#' # shift_fill
#' y = c("a",NA,"b",NA,"c")
#'
#' shift_fill(y) # equals to
#' shift_fill(y,"down")
#'
#' shift_fill(y,"up")

#' @rdname missing
#' @export
drop_na_dt = function(.data,...){
  dt = as_dt(.data)
  if(substitute(list(...)) %>% deparse() == "list()")
    dot_string <- NULL
  else
    dt[0] %>%
      select_dt(...) %>%
      names() -> dot_string
  if(is.null(dot_string)) na.omit(dt)
  else {
    index = str_glue("!is.na({dot_string})") %>% paste0(collapse = " & ")
    eval(parse(text = str_glue("dt[{index}]")))
  }
}

#' @rdname missing
#' @export

replace_na_dt = function(.data,...,to){
  if(is.na(to)) return(.data)
  dt = as.data.table(.data)
  # dt = as_dt(.data)
  if(substitute(list(...)) %>% deparse() == "list()")
    dot_string <- NULL
  else
    dt[0] %>%
      select_dt(...) %>%
      names() -> dot_string
  if(is.null(dot_string)) {
    for (j in seq_len(ncol(dt)))
      set(dt,which(is.na(dt[[j]])),j,to)
  } else{
    for (j in dot_string)
      set(dt,which(is.na(dt[[j]])),j,to)
  }
  dt
}


#' @rdname missing
#' @export

delete_na_cols = function(.data,prop = NULL, n = NULL){
  dt = as_dt(.data)

  if(is.numeric(prop) && prop > 1) {
    n = prop
    prop = NULL
  }
  if(is.null(prop) & !is.null(n)) prop = n/nrow(dt)
  if(is.null(prop)) prop = 1
  if(prop %between% c(0,1)){
    colMeans(is.na(dt)) %>%
      subset(subset = .< prop) %>%
      names() -> to_save

    select_dt(dt,cols = to_save)
  }
  else stop("Inputs are invalid.")
}


#' @rdname missing
#' @export
delete_na_rows = function(.data,prop = NULL, n = NULL){
  dt = as_dt(.data)

  if(is.numeric(prop) && prop > 1) {
    n = prop
    prop = NULL
  }
  if(is.null(prop) & !is.null(n)) prop = n/nrow(dt)
  if(is.null(prop)) prop = 1
  if(prop %between% c(0,1)){
    which(rowMeans(is.na(dt)) < prop) -> to_save

    dt[to_save]
  }
  else stop("Inputs are invalid.")
}


#' @rdname missing
#' @export

fill_na_dt = function(.data,...,direction = "down"){
  dt = as.data.table(.data)
  #dt = as_dt(.data)
  if(substitute(list(...)) %>% deparse() == "list()")
    update_cols <- names(dt)
  else
    dt[0] %>%
      select_dt(...) %>%
      names() -> update_cols

  dt[0] %>% select_dt(cols = update_cols) %>%
    select_dt(is.numeric) %>%
    names()-> num_cols
  setdiff(update_cols,num_cols) -> nnum_cols

  if(length(num_cols) > 0) {
    if(direction == "down") setnafill(dt,type = "locf",cols = num_cols)
    else if(direction == "up") setnafill(dt,type = "nocb",cols = num_cols)
  }
  if(length(nnum_cols) > 0){
    dt[,(nnum_cols) := lapply(.SD, shift_fill,direction = direction),
       .SDcols = nnum_cols]
  }

  dt[]
}

#' @rdname missing
#' @export
shift_fill = function(x,direction = "down"){
  if(direction == "down") type = "lag"
  else if(direction == "up") type = "lead"
  repeat{
    x = fcoalesce(x,shift(x,type = type))
    if(!anyNA(x)) break
  }
  x
}

#' @title Short cut to data.table
#' @description To use facilities provided by \pkg{data.table}, but do not have to
#' load \pkg{data.table} package.
#' @param .data A data.frame
#' @param ... Recieve \code{B} in data.table's \code{A[B]} syntax.
#' @details The \code{as_dt} could turn any data frame to data.table class. If the data is
#' not a data frame, return error.
#' @details  The \code{in_dt} function creates a virtual environment in data.table, it could be
#' piped well because it still follows the principals of \pkg{tidyfst}, which are: (1) Never
#' use in place replacement and (2) Always recieves a data frame (data.frame/tibble/data.table)
#' and returns a data.table. Therefore, the in place functions like \code{:=} will still
#' return the results.
#' @seealso \code{\link[data.table]{data.table}}
#' @examples
#' iris %>% as_dt()
#' iris %>% in_dt(order(-Sepal.Length),.SD[.N],by=Species)

#' @rdname dt
#' @export
in_dt = function(.data,...){
  #dt = as_dt(.data)
  dt = as.data.table(.data)
  dt[...][]
}

#' @rdname dt
#' @export


as_dt = function (.data) {
  if(is.data.table(.data)) .data
  else if (is.data.frame(.data) || ("fst_table" %chin% class(.data)))
    as.data.table(.data)
  else stop("Only a data.frame or fst_table could be received.")
}



# as_dt = function (.data) {
#   if (is.data.frame(.data) || ("fst_table" %chin% class(.data)))
#     as.data.table(.data)
#   else stop("Only a data.frame or fst_table could be received.")
# }

#' @title Conditional update of columns in data.table
#' @description Update or add columns when the given condition is met.
#' @description \code{mutate_when} integrates \code{mutate} and \code{case_when}
#' in \pkg{dplyr} and make a new tidy verb for data.table. \code{mutate_vars} is
#'  a super function to do updates in specific columns according to conditions.
#' @param .data data.frame
#' @param when An object which can be coerced to logical mode
#' @param ... Name-value pairs of expressions for \code{mutate_when}.
#' Additional parameters to be passed to parameter '.func' in \code{mutate_vars}.
#' @param .cols Any types that can be accepted by \code{\link[tidyfst]{select_dt}}.
#' @param .func Function to be run within each column, should return a value or
#' vectors with same length.
#' @param by (Optional) Mutate by what group?
#' @return data.table
#' @seealso \code{\link[tidyfst]{select_dt}}, \code{\link[dplyr]{case_when}}
#' @examples
#' iris[3:8,]
#' iris[3:8,] %>%
#'   mutate_when(Petal.Width == .2,
#'               one = 1,Sepal.Length=2)
#'
#' iris %>% mutate_vars("Pe",scale)
#' iris %>% mutate_vars(is.numeric,scale)
#' iris %>% mutate_vars(-is.factor,scale)
#' iris %>% mutate_vars(1:2,scale)
#' iris %>% mutate_vars(.func = as.character)

#' @rdname mutate_vars
#' @export

mutate_when = function(.data,when,...,by){
  dt = as.data.table(.data)
  eval.parent(substitute(dt[when,`:=`(...),by][]),1)
  # eval(substitute(dt[when,`:=`(...),by][]))
}

#' @rdname mutate_vars
#' @export
mutate_vars = function(.data,.cols = NULL,.func,...,by){
  #dt = as_dt(.data)
  dt = as.data.table(.data)
  deparse(substitute(.cols)) -> .cols
  deparse(substitute(by)) -> .by
  if (.cols == "NULL")
    sel_name = names(dt[0])
  else{
    eval(
      parse(
        text =
          str_glue("select_dt(dt[0],{.cols}) %>% names() -> sel_name")))
  }
  eval(parse(text = str_glue(
  "dt[,(sel_name) := lapply(.SD,.func,...), by = {.by},.SDcols = sel_name][]")))

}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  #eval = FALSE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(tidyfst)

set.seed(1L)

## Create a data table
DF <- data.table(V1 = rep(c(1L, 2L), 5)[-10],
                V2 = 1:9,
                V3 = c(0.5, 1.0, 1.5),
                V4 = rep(LETTERS[1:3], 3))
copy(DF) -> DT

class(DF)
DF

## -----------------------------------------------------------------------------
### Filter rows using indices
slice_dt(DF, 3:4)

### Discard rows using negative indices
slice_dt(DF, -(3:7))

### Filter rows using a logical expression
filter_dt(DF, V2 > 5)
filter_dt(DF, V4 %in% c("A", "C"))
filter_dt(DF, V4 %chin% c("A", "C")) # fast %in% for character

### Filter rows using multiple conditions
filter_dt(DF, V1 == 1 & V4 == "A")

### Filter unique rows
distinct_dt(DF) # unique(DF)
distinct_dt(DF, V1,V4)

### Discard rows with missing values
drop_na_dt(DF) # na.omit(DF)

### Other filters
sample_n_dt(DF, 3)      # n random rows
sample_frac_dt(DF, 0.5) # fraction of random rows
slice_max_dt(DF, V1,1)     # top n entries (includes equals)

filter_dt(DT,V4 %like% "^B")
filter_dt(DT,V2 %between% c(3, 5))
filter_dt(DT,between(V2, 3, 5, incbounds = FALSE))
filter_dt(DT,V2 %inrange% list(-1:1, 1:3))  # see also ?inrange

## -----------------------------------------------------------------------------
### Sort rows by column
arrange_dt(DF, V3)

### Sort rows in decreasing order
arrange_dt(DF, -V3)

### Sort rows based on several columns
arrange_dt(DF, V1, -V2)

## -----------------------------------------------------------------------------
### Select one column using an index (not recommended)
pull_dt(DT,3) # returns a vector
select_dt(DT,3) # returns a data.table

### Select one column using column name
select_dt(DF, V2) # returns a data.table
pull_dt(DF, V2)   # returns a vector

### Select several columns
select_dt(DF, V2, V3, V4)
select_dt(DF, V2:V4) # select columns between V2 and V4

### Exclude columns
select_dt(DF, -V2, -V3)

### Select/Exclude columns using a character vector
cols <- c("V2", "V3")
select_dt(DF,cols = cols)
select_dt(DF,cols = cols,negate = TRUE)

### Other selections
select_dt(DF, cols = paste0("V", 1:2))
relocate_dt(DF, V4) # reorder columns
select_dt(DF, "V")
select_dt(DF, "3$")
select_dt(DF, ".2")
select_dt(DF, "V1")
select_dt(DF, -"^V2")
# remove variables using "-" prior to function

## -----------------------------------------------------------------------------
### Summarise one column
summarise_dt(DF, sum(V1)) # returns a data.table
summarise_dt(DF, sumV1 = sum(V1)) # returns a data.table

### Summarise several columns
summarise_dt(DF, sum(V1), sd(V3))

### Summarise several columns and assign column names
DF %>%
  summarise_dt(sumv1 = sum(V1),
              sdv3  = sd(V3))

### Summarise a subset of rows
DT[1:4, sum(V1)]
DF %>%
  slice_dt(1:4) %>%
  summarise_dt(sum(V1))

### Misc
summarise_dt(DF, nth(V3,1))
summarise_dt(DF, nth(V3,-1))
summarise_dt(DF, nth(V3, 5))
summarise_dt(DF, uniqueN(V4))
uniqueN(DF)

## -----------------------------------------------------------------------------
### By group
# not recommended
DF %>%
  group_dt(
    by = V4,
    summarise_dt(sumV2 = sum(V2))
  )
# recommended
DF %>% 
  summarise_dt(sumV2 = sum(V2),by = V4)


### By several groups
DF %>% 
  summarise_dt(sumV2 = sum(V2),by = .(V1,V4))

### Calling function in by
DF %>% 
  summarise_dt(sumV2 = sum(V2),by = tolower(V4))


### Assigning column name in by
DF %>% 
  summarise_dt(sumV2 = sum(V2),by = .(abc = tolower(V4)))


### Using a condition in by
DF %>% 
  summarise_dt(sumV2 = sum(V2),by = V4 == "A")



### By on a subset of rows
DF %>% 
  slice_dt(1:5) %>% 
  summarise_dt(sumV1 = sum(V1),by = V4)

### Count number of observations for each group
count_dt(DF, V4)

### Add a column with number of observations for each group
add_count_dt(DF, V1)

### Retrieve the first/last/nth observation for each group
DF %>% summarise_dt(by = V4,nth(V2,1))
DF %>% summarise_dt(by = V4,nth(V2,-1))
DF %>% summarise_dt(by = V4,nth(V2,2))


## -----------------------------------------------------------------------------
### Summarise all the columns
summarise_vars(DT,.func = max)

### Summarise several columns
summarise_vars(DT,c("V1", "V2"),mean)

### Summarise several columns by group
DT %>% 
    summarise_vars(c("V1", "V2"),mean,by = V4)

## using patterns (regex)
DT %>% 
    summarise_vars("V1|V2",mean,by = V4)

## Summarise with more than one function by group
# when you can't find a way, you can always use `in_dt` to use data.table
DT %>% 
  in_dt(, by = V4, 
     c(lapply(.SD, sum),
       lapply(.SD, mean)))

### Summarise using a condition
summarise_vars(DF, is.numeric, mean)

### Modify all the columns
mutate_vars(DF, .func = rev)

### Modify several columns (dropping the others)
DF %>% 
  select_dt(cols = c("V1", "V2")) %>% 
  mutate_vars(.func = sqrt)
DF %>% 
  select_dt(-V4) %>% 
  mutate_vars(.func = exp)

### Modify several columns (keeping the others)
DF %>% 
  mutate_vars(c("V1", "V2"), sqrt)
DF %>% 
  mutate_vars(-"V4",  exp)

### Modify columns using a condition (dropping the others)
select_dt(DT,is.numeric)

### Modify columns using a condition (keeping the others)
mutate_vars(DT,is.numeric,as.integer)

### Use a complex expression
DF %>% 
  group_dt(
    by = V4,
    slice_dt(1:2) %>% 
      transmute_dt(V1 = V1,
            V2 = "X")
  )

### Use multiple expressions (with DT[,{j}])
DT %>% 
  in_dt(,{
      print(V1) #  comments here!
      print(summary(V1))
      x <- V1 + sum(V2)
     .(A = 1:.N, B = x) # last list returned as a data.table
     }
  )


## -----------------------------------------------------------------------------
### Select first/last/… row by group
DT %>% 
  group_dt(
    by = V4,
    head(1)
  )

DT %>% 
  group_dt(
    by = V4,
    tail(2)
  )

DT %>% 
  group_dt(
    by = V4,
    slice_dt(1,.N)
  )

### Select rows using a nested query
DF %>% 
  group_dt(
    by = V4,
    arrange_dt(V2) %>% 
      slice_dt(1)
  )

### Add a group counter column
DT %>% 
  mutate_dt(Grp = .GRP,by = .(V4, V1))

### Get row number of first (and last) observation by group
DT %>% summarise_dt(I = .I,by = V4)
DT %>% summarise_dt(I = .I[1],by = V4)
DT %>% summarise_dt(I = .I[c(1,.N)],by = V4)

### Handle list-columns by group

DT %>% 
  select_dt(V1,V4) %>% 
  chop_dt(V1) # return V1 as a list
DT %>% nest_dt(V4) # subsets of the data

### Grouping sets (multiple by at once)
# use data.table directly, tidyfst does not provide new methods for it yet
data.table::rollup(DT,
       .(SumV2 = sum(V2)),
       by = c("V1", "V4"))

data.table::rollup(DT,
       .(SumV2 = sum(V2), .N),
       by = c("V1", "V4"),
       id = TRUE)

data.table::cube(DT,
     .(SumV2 = sum(V2), .N),
     by = c("V1", "V4"),
     id = TRUE)

data.table::groupingsets(DT,
             .(SumV2 = sum(V2), .N),
             by   = c("V1", "V4"),
             sets = list("V1", c("V1", "V4")),
             id   = TRUE)


## ----eval=FALSE---------------------------------------------------------------
# ### Write data to a csv file
# fwrite(DT, "DT.csv")
# 
# ### Write data to a tab-delimited file
# fwrite(DT, "DT.txt", sep = "\t")
# 
# ### Write list-column data to a csv file
# fwrite(setDT(list(0, list(1:5))), "DT2.csv")
# #
# ### Read a csv / tab-delimited file
# fread("DT.csv")
# # fread("DT.csv", verbose = TRUE) # full details
# fread("DT.txt", sep = "\t")
# 
# ### Read a csv file selecting / droping columns
# fread("DT.csv", select = c("V1", "V4"))
# fread("DT.csv", drop = "V4")
# # NA
# ### Read and rbind several files
# rbindlist(lapply(c("DT.csv", "DT.csv"), fread))
# # c("DT.csv", "DT.csv") %>% lapply(fread) %>% rbindlist
# 

## -----------------------------------------------------------------------------
### Melt data (from wide to long)
fsetequal(DT,DF)

mDT = DT %>% longer_dt(V3,V4)
mDF = DF %>% longer_dt(-"V1|V2")

fsetequal(mDT,mDF)
mDT

### Cast data (from long to wide)
mDT %>% 
  wider_dt(V4,name = "name",value = "value")
# below is a special case and could only be done in tidyfst
mDT %>% 
  wider_dt(V4,name = "name",value = "value",fun = list)

mDT %>% 
  wider_dt(V4,name = "name",value = "value",fun = sum)

### Split
split(DT, by = "V4")


## -----------------------------------------------------------------------------
### Lead/Lag
lag_dt(1:10,n = 1)
lag_dt(1:10,n = 1:2)
lead_dt(1:10,n = 1)


## -----------------------------------------------------------------------------
x <- data.table(Id  = c("A", "B", "C", "C"),
                X1  = c(1L, 3L, 5L, 7L),
                XY  = c("x2", "x4", "x6", "x8"),
                key = "Id")

y <- data.table(Id  = c("A", "B", "B", "D"),
                Y1  = c(1L, 3L, 5L, 7L),
                XY  = c("y1", "y3", "y5", "y7"),
                key = "Id")

### left join
left_join_dt(x, y, by = "Id")

### right join
right_join_dt(x, y, by = "Id")

### inner join
inner_join_dt(x, y, by = "Id")

### full join
full_join_dt(x, y, by = "Id")

### semi join
semi_join_dt(x, y, by = "Id")

### anti join
anti_join_dt(x, y, by = "Id")


## -----------------------------------------------------------------------------
x <- data.table(1:3)
y <- data.table(4:6)
z <- data.table(7:9, 0L)
### Bind rows
rbind(x, y)
rbind(x, z, fill = TRUE)

### Bind rows using a list
rbindlist(list(x, y), idcol = TRUE)

### Bind columns
cbind(x, y)


## -----------------------------------------------------------------------------
x <- data.table(c(1, 2, 2, 3, 3))
y <- data.table(c(2, 2, 3, 4, 4))
### Intersection
fintersect(x, y)
fintersect(x, y, all = TRUE)

### Difference
fsetdiff(x, y)
fsetdiff(x, y, all = TRUE)

### Union
funion(x, y)
funion(x, y, all = TRUE)

### Equality
fsetequal(x, x[order(-V1),])
all.equal(x, x) # S3 method
setequal(x, x[order(-V1),])

#' @title Parse,inspect and extract data.table from fst file
#' @name fst
#' @description A tookit of APIs for reading fst file as data.table, could select by column, row and conditional filtering.
#' @param path path to fst file
#' @param ft An object of class fst_table, returned by \code{parse_fst}
#' @param row_no An integer vector (Positive)
#' @param ... The filter conditions
#' @return \code{parse_fst} returns a fst_table class.
#' @return \code{select_fst} and \code{filter_fst} returns a data.table.
#' @details \code{summary_fst} could provide some basic information about
#' the fst table.
#' @seealso \code{\link[fst]{fst}}, \code{\link[fst]{metadata_fst}}
#' @examples
#'
#' \dontrun{
#'   fst::write_fst(iris,"iris_test.fst")
#'   # parse the file but not reading it
#'   parse_fst("iris_test.fst") -> ft
#'   ft
#'
#'   class(ft)
#'   lapply(ft,class)
#'   names(ft)
#'   dim(ft)
#'   summary_fst(ft)
#'
#'   # get the data by query
#'   ft %>% slice_fst(1:3)
#'   ft %>% slice_fst(c(1,3))
#'
#'   ft %>% select_fst(Sepal.Length)
#'   ft %>% select_fst(Sepal.Length,Sepal.Width)
#'   ft %>% select_fst("Sepal.Length")
#'   ft %>% select_fst(1:3)
#'   ft %>% select_fst(1,3)
#'   ft %>% select_fst("Se")
#'   ft %>% select_fst("nothing")
#'   ft %>% select_fst("Se|Sp")
#'   ft %>% select_fst(cols = names(iris)[2:3])
#'
#'   ft %>% filter_fst(Sepal.Width > 3)
#'   ft %>% filter_fst(Sepal.Length > 6 , Species == "virginica")
#'   ft %>% filter_fst(Sepal.Length > 6 & Species == "virginica" & Sepal.Width < 3)
#'
#'   unlink("iris_test.fst")
#' }


globalVariables(c("."))

#' @rdname fst
#' @export
parse_fst = function(path){
  fst(path)
}


#' @rdname fst
#' @export

slice_fst = function(ft,row_no){
  setDT(ft[row_no,])[]
}

#' @rdname fst
#' @export
# select_fst = function(ft,...){
#
#   setDT(ft[1,])[0] %>% select_dt(...) %>% names() -> sel_names
#   setDT(ft[names(ft) %chin% sel_names])[]
#
# }

select_fst = function(ft,...){

  setDT(ft[1,])[0] %>% select_dt(...) %>% names() -> sel_names
  names(ft) %chin% sel_names -> logical_vec
  if(all(logical_vec == FALSE)) {
    warning("No matched columns,try other patterns. Names of the `fst_table` are listed.")
    names(ft)
  } else setDT(ft[logical_vec])[]

}

#' @rdname fst
#' @export

filter_fst = function(ft,...){
  substitute(list(...)) %>%
    lapply(deparse) %>%
    .[-1] %>%
    str_c(collapse = " & ")-> dot_string
  names(ft) -> ft_names
  ft_names[str_detect(dot_string,ft_names)] -> old
  paste0("ft$",old) -> new
  for(i in seq_along(old)) gsub(old[i],new[i],dot_string) -> dot_string
  eval(parse(text = str_glue("ft[{dot_string},] %>% as.data.table()")))
}

#' @rdname fst
#' @export
summary_fst = function(ft){
  metadata_fst(.subset2(ft, "meta")[[1]])
}

#' @title Complete a data frame with missing combinations of data
#' @description Turns implicit missing values into explicit missing values.
#'  All the combinations of column values (should be unique) will be constructed.
#'  Other columns will be filled with NAs or constant value.
#' @param .data data.frame
#' @param ... Specification of columns to expand.The selection of columns is
#' supported by the flexible \code{\link[tidyfst]{select_dt}}.
#' To find all unique combinations of provided columns, including those not found in the data,
#' supply each variable as a separate argument. But the two modes (select the
#' needed columns and fill outside values) could not be mixed,
#' find more details in examples.
#' @param fill Atomic value to fill into the missing cell, default uses \code{NA}.
#' @details When the provided columns with addtion data are of different length,
#' all the unique combinations would be returned. This operation should be used
#' only on unique entries, and it will always returned the unique entries.
#' @details If you supply fill parameter, these values will also replace existing explicit missing values in the data set.
#' @seealso \code{\link[tidyr]{complete}}
#' @return data.table
#' @examples
#' df <- data.table(
#'   group = c(1:2, 1),
#'   item_id = c(1:2, 2),
#'   item_name = c("a", "b", "b"),
#'   value1 = 1:3,
#'   value2 = 4:6
#' )
#'
#' df %>% complete_dt(item_id,item_name)
#' df %>% complete_dt(item_id,item_name,fill = 0)
#' df %>% complete_dt("item")
#' df %>% complete_dt(item_id=1:3)
#' df %>% complete_dt(item_id=1:3,group=1:2)
#' df %>% complete_dt(item_id=1:3,group=1:3,item_name=c("a","b","c"))
#'

#' @export

complete_dt = function(.data,...,fill = NA){
  dt = as_dt(.data)

  if(
    substitute(list(...)) %>%
    deparse() %>%
    .[1] %>%
    str_detect("=")
  ) {
    list(...) %>%
      lapply(unique) %>%
      deparse() %>%
      str_replace("list\\(","CJ\\(") %>%
      parse(text = .) %>%
      eval() %>%
      merge(dt,all = TRUE) %>%
      replace_na_dt(to=fill) %>%
      unique()
  }else
    dt %>%
    select_dt(...) %>%
    lapply(unique) %>%
    deparse() %>%
    str_replace("list\\(","CJ\\(") %>%
    parse(text = .) %>%
    eval() %>%
    merge(dt,all = TRUE) %>%
    replace_na_dt(to=fill) %>%
    unique()
}

#' @title Pull out a single variable
#' @description Extract vector from data.frame, works likt `[[`. Analogous function for \code{pull} in \pkg{dplyr}
#' @param .data data.frame
#' @param col A name of column or index (should be positive).
#' @return vector
#' @seealso \code{\link[dplyr]{pull}}
#' @examples
#' mtcars %>% pull_dt(2)
#' mtcars %>% pull_dt(cyl)
#' mtcars %>% pull_dt("cyl")
#' @export

pull_dt = function(.data,col) .data[[substitute(col)]]

# pull_dt = function(.data,col) eval.parent(substitute(.data[[col]]))

#' @title Count pairs of items within a group
#' @description Count the number of times each pair
#' of items appear together within a group.
#' For example, this could count the number of times two words appear within documents.
#' This function has referred to \code{pairwise_count} in \strong{widyr} package,
#' but with very different defaults on several parameters.
#' @param .data A data.frame.
#' @param .group Column name of counting group.
#' @param .value Item to count pairs, will end up in \code{V1} and \code{V2} columns.
#' @param upper When \code{FALSE}(Default), duplicated combinations would be removed.
#' @param diag Whether to include diagonal (V1==V2) in output. Default uses \code{FALSE}.
#' @param sort Whether to sort rows by counts. Default uses \code{TRUE}.
#' @return A data.table with 3 columns (named as "item1","item2" and "n"), containing combinations
#'  in "item1" and "item2", and counts in "n".
#' @seealso \code{\link[widyr]{pairwise_count}}
#' @examples
#'
#' dat <- data.table(group = rep(1:5, each = 2),
#'               letter = c("a", "b",
#'                          "a", "c",
#'                          "a", "c",
#'                          "b", "e",
#'                          "b", "f"))
#' pairwise_count_dt(dat,group,letter)
#' pairwise_count_dt(dat,group,letter,sort = FALSE)
#' pairwise_count_dt(dat,group,letter,upper = TRUE)
#' pairwise_count_dt(dat,group,letter,diag = TRUE)
#' pairwise_count_dt(dat,group,letter,diag = TRUE,upper = TRUE)
#'
#' # The column name could be specified using character.
#' pairwise_count_dt(dat,"group","letter")

#' @rdname pairwise
#' @export
pairwise_count_dt = function(.data,.group,.value,upper = FALSE,diag = FALSE,sort = TRUE){
  dt = as_dt(.data)
  parse_name <- substitute(.group) %>% deparse()
  if (!str_detect(parse_name, "^\""))
    .group = parse_name
  parse_name <- substitute(.value) %>% deparse()
  if (!str_detect(parse_name, "^\""))
    .value = parse_name

  str_glue("dt[,CJ(V1={.value},V2={.value},unique = TRUE),by = .group][,(.group):=NULL]") %>%
    parse(text = .) %>% eval() -> res

  if(!diag) res = res[V1!=V2]
  if(!upper) res = res[V1 <= V2]

  res %>%
    count_dt(V1,V2,sort = sort) %>%
    rename_dt(item1 = V1,item2 = V2)

}

globalVariables(c("V1","V2","item1","item2"))

#' @title Summarise columns to single values
#' @name summarise_dt
#' @description Summarise group of values into one value for each group. If there is only one group, then only one value would be returned.
#'  The summarise function should always return a single value.
#' @param .data data.frame
#' @param ... List of variables or name-value pairs of summary/modifications
#'   functions for \code{summarise_dt}.Additional parameters to be passed to
#'    parameter '.func' in \code{summarise_vars}.
#' @param by unquoted name of grouping variable of list of unquoted names of
#'   grouping variables. For details see \link[data.table]{data.table}
#' @param when An object which can be coerced to logical mode
#' @param .cols Columns to be summarised.
#' @param .func Function to be run within each column, should return a value or vectors with same length.
#' @details \code{summarise_vars} could complete summarise on specific columns.
#' @return data.table
#' @seealso \code{\link[dplyr]{summarise}}
#' @examples
#' iris %>% summarise_dt(avg = mean(Sepal.Length))
#' iris %>% summarise_dt(avg = mean(Sepal.Length),by = Species)
#' mtcars %>% summarise_dt(avg = mean(hp),by = .(cyl,vs))
#'
#' # the data.table way
#' mtcars %>% summarise_dt(cyl_n = .N, by = .(cyl, vs)) # `.` is short for list
#'
#' iris %>% summarise_vars(is.numeric,min)
#' iris %>% summarise_vars(-is.factor,min)
#' iris %>% summarise_vars(1:4,min)
#'
#' iris %>% summarise_vars(is.numeric,min,by ="Species")
#' mtcars %>% summarise_vars(is.numeric,mean,by = c("vs", "am"))
#'
#' # use multiple functions on multiple columns
#' iris %>%
#'   summarise_vars(is.numeric,.func = list(mean,sd,median))
#' iris %>%
#'   summarise_vars(is.numeric,.func = list(mean,sd,median),by = Species)
#'

globalVariables("res")

#' @rdname summarise_dt
#' @export

summarise_dt = function(.data,...,by = NULL){
  dt = as_dt(.data)
  eval.parent(substitute(dt[,.(...),by = by]))
}

#' @rdname summarise_dt
#' @export

summarize_dt = summarise_dt

#' @rdname summarise_dt
#' @export

summarise_when = function(.data,when,...,by = NULL){
  dt = as_dt(.data)
  eval.parent(substitute(dt[when,.(...),by = by]))
}

#' @rdname summarise_dt
#' @export

summarize_when = summarise_when

#' @rdname summarise_dt
#' @export

summarise_vars = function (.data, .cols = NULL, .func, ...,by) {
  dt = as_dt(.data)
  deparse(substitute(.cols)) -> .cols
  deparse(substitute(by)) -> .by
  if (.cols == "NULL")
    sel_name = names(dt[0])
  else{
    eval(
      parse(
        text =
          str_glue("select_dt(dt[0],{.cols}) %>% names() -> sel_name")))
  }

  if(!is.list(.func)){
    eval(parse(text = str_glue(
      "res = dt[,lapply(.SD, .func, ...), by = {.by},.SDcols = sel_name]")))
    res[, unique(names(res)), with = FALSE]
  }else{
    func_names = sapply(substitute(.func),deparse) %>% .[-1]
    lapply(func_names,function(func){
      eval(parse(text = str_glue(
        "res = dt[,lapply(.SD, get(func), ...),
        by = {.by},.SDcols = sel_name][,fun_name:=func]")))
    }) %>% rbindlist()
  }

}

#' @rdname summarise_dt
#' @export

summarize_vars = summarise_vars

#' @import data.table
#' @import fst
#' @import stringr
#' @importFrom stats na.omit setNames median
#' @importFrom utils object.size head tail install.packages installed.packages
#' @importFrom pak pak

#' @export
stringr::`%>%`

#' @export
data.table::`%like%`

#' @export
data.table::`%chin%`

#' @export
data.table::data.table

#' @export
data.table::as.data.table

#' @export
data.table::fread

#' @export
data.table::fwrite

#' @export
data.table::fintersect

#' @export
data.table::fsetdiff

#' @export
data.table::funion

#' @export
data.table::fsetequal

#' @export
data.table::frollapply

#' @export
data.table::fcoalesce

#' @export
data.table::uniqueN

#' @export
data.table::rbindlist

#' @export
data.table::tables

#' @export
data.table::like

#' @export
data.table::copy

#' @export
data.table::key

#' @export
data.table::CJ

#' @export
data.table::rleid

#' @export
data.table::rleidv

#' @export
data.table::fcase

#' @export
data.table::between

#' @export
data.table::set

#' @export
data.table::setDT

#' @export
data.table::setDF

#' @export
data.table::setnames

#' @title Pivot data from wide to long
#' @description Turning a wide table to its longer form. It takes multiple columns and collapses into key-value pairs.
#' @param .data A data.frame
#' @param ... Pattern for unchanged group or unquoted names. Pattern can accept
#' regular expression to match column names. It can recieve what \code{select_dt}
#' recieves.
#' @param name Name for the measured variable names column.
#' The default name is 'name'.
#' @param value Name for the molten data values column(s).
#' The default name is 'value'.
#' @param na.rm If \code{TRUE}, \code{NA} values will be removed from the molten data.
#' @return A data.table
#' @seealso \code{\link[tidyfst]{wider_dt}},
#'   \code{\link[data.table]{melt}},
#'   \code{\link[tidyr]{pivot_longer}}
#' @examples
#'
#' ## Example 1:
#' stocks = data.frame(
#'   time = as.Date('2009-01-01') + 0:9,
#'   X = rnorm(10, 0, 1),
#'   Y = rnorm(10, 0, 2),
#'   Z = rnorm(10, 0, 4)
#' )
#'
#' stocks
#'
#' stocks %>%
#'   longer_dt(time)
#'
#' stocks %>%
#'   longer_dt("ti")
#'
#' # Example 2:
#'
#' \donttest{
#'   library(tidyr)
#'
#'   billboard %>%
#'     longer_dt(
#'       -"wk",
#'       name = "week",
#'       value = "rank",
#'       na.rm = TRUE
#'     )
#'
#'   # or use:
#'   billboard %>%
#'     longer_dt(
#'       artist,track,date.entered,
#'       name = "week",
#'       value = "rank",
#'       na.rm = TRUE
#'     )
#'
#'   # or use:
#'   billboard %>%
#'     longer_dt(
#'       1:3,
#'       name = "week",
#'       value = "rank",
#'       na.rm = TRUE
#'     )
#' }


#' @rdname longer
#' @export
longer_dt = function(.data,...,
                     name = "name",
                     value = "value",
                     na.rm = FALSE){
  dt = as_dt(.data)
  group = dt[0] %>% select_mix(...) %>% names()
  melt(data = dt,
       id = group,
       variable.name = name,
       value.name = value,
       na.rm = na.rm)
}

#' @title Arrange entries in data.frame
#' @description Order the rows of a data frame rows by the values of selected columns.
#' @param .data data.frame
#' @param ... Arrange by what group? Minus symbol means arrange by
#' descending order.
#' @return data.table
#' @seealso \code{\link[dplyr]{arrange}}
#' @examples
#'
#' iris %>% arrange_dt(Sepal.Length)
#'
#' # minus for decreasing order
#' iris %>% arrange_dt(-Sepal.Length)
#'
#' # arrange by multiple variables
#' iris %>% arrange_dt(Sepal.Length,Petal.Length)
#'

#' @export

arrange_dt = function(.data,...){
  dt = as_dt(.data)
  dt[order(...)]
}

#' @title Fast value replacement in data frame
#' @description While \code{replace_na_dt} could replace all NAs to another
#' value, \code{replace_dt} could replace any value(s) to another specific
#' value.
#' @param .data A data.frame
#' @param ... Colunms to be replaced. If not specified, use all columns.
#' @param from A value, a vector of values or a function returns a logical value.
#' Defaults to \code{is.nan}.
#' @param to A value. Defaults to \code{NA}.
#' @return A data.table.
#' @seealso \code{\link[tidyfst]{replace_na_dt}}
#' @examples
#' iris %>% mutate_vars(is.factor,as.character) -> new_iris
#'
#' new_iris %>%
#'   replace_dt(Species, from = "setosa",to = "SS")
#' new_iris %>%
#'   replace_dt(Species,from = c("setosa","virginica"),to = "sv")
#' new_iris %>%
#'   replace_dt(Petal.Width, from = .2,to = 2)
#' new_iris %>%
#'   replace_dt(from = .2,to = NA)
#' new_iris %>%
#'   replace_dt(is.numeric, from = function(x) x > 3, to = 9999 )


#' @export
replace_dt = function (.data, ..., from = is.nan,to = NA) {
  dt = as.data.table(.data)
  #dt = as_dt(.data)
  if(!is.function(from)) {
    if (setequal(from,to)) return(.data)
    if(length(from) == 1) .func = function(x) x == from
    else if(is.character(from)) .func = function(x) x %chin% from
    else .func = function(x) x %in% from
  } else .func = from


  if (substitute(list(...)) %>% deparse() == "list()")
    dot_string <- NULL
  else dot_string <- dt[0] %>% select_dt(...) %>% names()
  if (is.null(dot_string)) {
    for (j in seq_len(ncol(dt))) set(dt, which(.func(dt[[j]])),
                                     j, to)
  }
  else {
    for (j in dot_string) set(dt, which(.func(dt[[j]])),
                              j, to)
  }
  dt
}

#' @title Fast lead/lag for vectors
#' @description Find the "next" or "previous" values in a vector.
#' It has wrapped \pkg{data.table}'s \code{shift} function.
#' @param x A vector
#' @param n a positive integer of length 1,
#' giving the number of positions to lead or lag by. Default uses 1
#' @param fill Value to use for padding when the window goes beyond the input length.
#' Default uses \code{NA}
#' @return A vector
#' @seealso \code{\link[dplyr]{lead}},\code{\link[data.table]{shift}}
#' @examples
#' lead_dt(1:5)
#' lag_dt(1:5)
#' lead_dt(1:5,2)
#' lead_dt(1:5,n = 2,fill = 0)

#' @rdname lag_lead
#' @export
lead_dt = function(x,n = 1L,fill = NA){
  shift(x,n,fill,type = "lead")
}

#' @rdname lag_lead
#' @export
lag_dt = function(x,n = 1L,fill = NA){
  shift(x,n,fill,type = "lag")
}

#' @title Recode number or strings
#' @name rec
#' @description Recode discrete variables, including numerice and character
#' variable.
#' @param x A numeric or character vector.
#' @param rec String with recode pairs of old and new values.
#'  Find the usage in examples.
#' @param keep Logical. Decide whether to keep the original values if not recoded.
#'  Defaults to \code{TRUE}.
#' @return A vector.
#' @seealso \code{\link[sjmisc]{rec}}
#' @examples
#'
#' x = 1:10
#' x
#' rec_num(x, rec = "1=10; 4=2")
#' rec_num(x, rec = "1:3=1; 4:6=2")
#' rec_num(x, rec = "1:3=1; 4:6=2",keep = FALSE)
#'
#' y = letters[1:5]
#' y
#' rec_char(y,rec = "a=A;b=B")
#' rec_char(y,rec = "a,b=A;c,d=B")
#' rec_char(y,rec = "a,b=A;c,d=B",keep = FALSE)
#'

#' @rdname rec
#' @export
rec_num = function(x,rec,keep = TRUE){
  str_split(rec,";",simplify = TRUE) %>%
    str_squish() %>%
    strsplit("=") -> rec_list

  lapply(rec_list,function(x) paste0("x %in% c(",x[1],"),",x[2])) %>%
    str_c(collapse = ",") -> in_fcase

  res = eval(parse(text = str_glue("fcase({in_fcase})")))
  if(keep) fcoalesce(res,as.numeric(x))
  else res

}

#' @rdname rec
#' @export
rec_char = function(x, rec, keep = TRUE){
  str_split(rec,";",simplify = TRUE) %>%
    str_squish() %>%
    strsplit("=") -> rec_list

  lapply(rec_list,function(x) {
    x[1]%>% str_split(",",simplify = T) %>%
      lapply(function(x) paste0("'",x,"'")) %>%
      str_c(collapse = ",") %>%
      str_c("c(",.,")") %>%
      str_c("x %chin% ",.,",'",x[2],"'")
  }) %>% str_c(collapse = ",") -> in_fcase

  res = eval(parse(text = str_glue("fcase({in_fcase})")))
  if(keep) fcoalesce(res,x)
  else res

}

#' Load or unload R package(s)
#'
#' This function is a wrapper for \code{\link[base]{require}} and
#' \code{\link[base]{detach}}.  \code{pkg_load} checks to see if a
#' package is installed, if not it attempts to install the package
#' from CRAN. \code{pkg_unload} can detach one or more loaded packages.
#'
#' @param ... Name(s) of package(s).
#' @param pkg_names (Optional)Character vector containing packages to load
#'  or unload. Default uses \code{NULL}.
#' @seealso
#' \code{\link[base]{require}},
#' \code{\link[base]{detach}},
#' \code{\link[pacman]{p_load}},
#' \code{\link[pacman]{p_unload}}
#' @examples
#' \dontrun{
#' pkg_load(data.table)
#' pkg_unload(data.table)
#'
#' pkg_load(stringr,fst)
#' pkg_unload(stringr,fst)
#'
#' pkg_load(pkg_names = c("data.table","fst"))
#' p_unload(pkg_names = c("data.table","fst"))
#'
#' pkg_load(data.table,stringr,fst)
#' pkg_unload("all") # shortcut to unload all loaded packages
#' }

#' @rdname pkg_load
#' @export
pkg_load = function(...,pkg_names = NULL){
  if(is.null(pkg_names)){
    (match.call(expand.dots = FALSE)$...) %>%
      as.character() %>%
      setdiff(.packages())-> list.of.packages
  }else list.of.packages = pkg_names

  new.packages <- list.of.packages[!(list.of.packages %in% installed.packages()[,"Package"])]
  if(length(new.packages)) pak::pak(new.packages)
  sapply(list.of.packages,\(x){
    suppressMessages(require(x,character.only = TRUE))
  }) %>% all() %>% invisible()
}

#' @rdname pkg_load
#' @export
pkg_unload = function(...,pkg_names = NULL){
  if(is.null(pkg_names)){
    (match.call(expand.dots = FALSE)$...) %>%
      as.character() -> list.of.packages
  }else list.of.packages = pkg_names

  if("all" %in% list.of.packages)
    list.of.packages = names(utils::sessionInfo()[["otherPkgs"]])
  else
    list.of.packages = list.of.packages %>%
      intersect(names(utils::sessionInfo()[["otherPkgs"]]))

  sapply(list.of.packages,\(x){
    x <- paste0("package:", x)
    suppressWarnings(detach(x, unload = TRUE, character.only = TRUE,
                            force = TRUE))
  })
  if(length(list.of.packages)>0){
    message("The following packages have been unloaded:\n",
            paste0("\b", paste(list.of.packages, collapse = ", ")))
  }

}

#' @title Conversion between tidy table and named matrix
#' @description Convenient fucntions to implement conversion between
#'  tidy table and named matrix.
#' @param m A matrix
#' @param df A data.frame with at least 3 columns, one for row name,
#'  one for column name, and one for values. The names for column and
#'  row should be unique.
#' @param row Unquoted expression of column name for row
#' @param col Unquoted expression of column name for column
#' @param value Unquoted expression of column name for values
#' @return For \code{mat_df}, a data.frame.
#' For \code{df_mat}, a named matrix.
#' @examples
#'
#' mm = matrix(c(1:8,NA),ncol = 3,dimnames = list(letters[1:3],LETTERS[1:3]))
#' mm
#' tdf = mat_df(mm)
#' tdf
#' mat = df_mat(tdf,row,col,value)
#' setequal(mm,mat)
#'
#' tdf %>%
#'   setNames(c("A","B","C")) %>%
#'   df_mat(A,B,C)
#'

#' @rdname tidymat
#' @export
mat_df = function(m){
  stopifnot(is.matrix(m))
  res = as.data.frame.table(m)
  setnames(res,old = names(res),new = c("row","col","value"))
  res
}

#' @rdname tidymat
#' @export
df_mat = function(df,row,col,value){
  stopifnot(is.data.frame(df))
  eval.parent(substitute(with(df, tapply(value, list(row, col), identity))),1)
  # df = df
  # eval(substitute(with(df, tapply(value, list(row, col), identity))))
}

#' @title Mutate columns in data.frame
#' @description Adds or updates columns in data.frame.
#' @param .data data.frame
#' @param ... List of variables or name-value pairs of summary/modifications
#'   functions.
#' @param by (Optional) Mutate by what group?
#' @return data.table
#' @seealso \code{\link[dplyr]{mutate}}
#' @examples
#'
#' iris %>% mutate_dt(one = 1,Sepal.Length = Sepal.Length + 1)
#' iris %>% transmute_dt(one = 1,Sepal.Length = Sepal.Length + 1)
#' # add group number with symbol `.GRP`
#' iris %>% mutate_dt(id = 1:.N,grp = .GRP,by = Species)
#'
#' @rdname mutate
#' @export

mutate_dt = function(.data,...,by){
  dt = as.data.table(.data)
  eval.parent(substitute(dt[,`:=`(...),by][]),1)
}

# mutate_dt = function(.data,...,by){
#   dt = as.data.table(.data)
#   substitute(dt[,`:=`(...),by][]) %>% eval()
# }

#' @rdname mutate
#' @export

transmute_dt = function(.data,...,by){
  dt = as_dt(.data)
  eval.parent(substitute(dt[,.(id_ = 1:.N,...),by]),1)[,id_:=NULL][]
}

# transmute_dt = function(.data,...,by){
#   dt = as_dt(.data)
#   substitute(dt[,.(id_ = 1:.N,...),by][]) %>% eval() %>%
#     .[,id_:=NULL] %>% .[]
# }

globalVariables("id_")

# transmute_dt = function(.data,...,by){
#   dt = as_dt(.data)
#   substitute(dt[,.(...),by][]) %>% eval()
# }

#' @title Compute TF–IDF Using data.table with Optional Counting and Grouping
#' @description
#' This function computes term frequency–inverse document frequency (tf–idf)
#' on a dataset with one row per term occurrence (or pre-counted).
#' It preserves original column names and returns new columns:
#'   - `n`: raw count (computed or user-supplied)
#'   - `tf`: term frequency per document
#'   - `idf`: inverse document frequency per group (or corpus)
#'   - `tf_idf`: tf × idf
#' If `group_col` is `NULL`, all documents are treated as a single group.
#'
#' @param .data       A data.frame or data.table of text data.
#' @param group_col   Character name of grouping column, or `NULL` for no grouping.
#' @param doc_col     Character name of document identifier column.
#' @param term_col    Character name of term/word column.
#' @param n_col       (Optional) Character name of pre-counted term-frequency column.
#'                    If `NULL` (default), counts are computed via `.N`.
#'
#' @return A data.table containing:
#'   - Original grouping, document, and term columns
#'   - `n`, `tf`, `idf`, and `tf_idf`
#'
#' @seealso \code{\link[tidytext]{bind_tf_idf}}
#' @examples
#'
#' # With groups
#' df <- data.frame(
#'   category = rep(c("A","B"), each = 6),
#'   doc_id   = rep(c("d1","d2","d3"), times = 4),
#'   word     = c("apple","banana","apple","banana","cherry","apple",
#'                "dog","cat","dog","mouse","cat","dog"),
#'   stringsAsFactors = FALSE
#' )
#' result <- bind_tf_idf_dt(df, "category", "doc_id", "word")
#' result
#'
#' # Without groups
#' df %>%
#'   filter_dt(category == "A") %>%
#'   bind_tf_idf_dt(doc_col = "doc_id",term_col = "word")
#'
#' # With counts provided
#' df %>%
#'   filter_dt(category == "A") %>%
#'   count_dt() %>%
#'   bind_tf_idf_dt(doc_col = "doc_id",term_col = "word",n_col = "n")
#' df %>%
#'   count_dt() %>%
#'   bind_tf_idf_dt(group_col = "category",
#'                  doc_col = "doc_id",
#'                  term_col = "word",n_col = "n")
#'


#' @export

bind_tf_idf_dt <- function(.data,
                           group_col = NULL,
                           doc_col,
                           term_col,
                           n_col = NULL) {
  # Convert to data.table
  dt <- as.data.table(.data)

  # Handle optional grouping
  if (is.null(group_col)) {
    dt[, .tmp_group := "__ALL__"]
    grp <- ".tmp_group"
    drop_tmp <- TRUE
  } else {
    grp <- group_col
    drop_tmp <- FALSE
  }

  # Compute or use existing n
  if (!is.null(n_col) && n_col %in% names(dt)) {
    # Copy pre-counted column
    dt_counts <- dt[, .(n = get(n_col)),
                    by = c(grp, doc_col, term_col)]
  } else {
    # Count occurrences per (group, document, term)
    dt_counts <- dt[, .(n = .N),
                    by = c(grp, doc_col, term_col)]
  }

  # Compute term frequency: tf = n / sum(n) per (group, document)
  dt_counts[, tf := n / sum(n), by = c(grp, doc_col)]

  # Compute inverse document frequency:
  # docs_in_group = number of unique docs in each group
  dt_counts[, docs_in_group := uniqueN(get(doc_col)), by = grp]
  # docs_with_term = number of docs in group containing the term
  dt_counts[, docs_with_term := uniqueN(get(doc_col)),
            by = c(grp, term_col)]
  # idf = log(docs_in_group / docs_with_term)
  dt_counts[, idf := log(docs_in_group / docs_with_term)]

  # Compute tf–idf
  dt_counts[, tf_idf := tf * idf]

  # Clean up helper columns
  dt_counts[, c("docs_in_group", "docs_with_term") := NULL]
  if (drop_tmp) dt_counts[, .tmp_group := NULL]

  # Return result
  return(dt_counts[])
}

globalVariables(c(".tmp_group","docs_in_group","docs_with_term","idf","n","tf","tf_idf"))

#' @title Select column from data.frame
#' @description Select specific column(s) via various ways. One can select columns by their column names, indexes or regular expression recognizing the column name(s).
#' @param .data data.frame
#' @param ... List of variables or name-value pairs of summary/modifications
#'   functions. It can also recieve conditional function to select columns.
#'   When starts with `-`(minus symbol) or `!`, return the negative columns.
#' @param cols (Optional)A numeric or character vector.
#' @param negate Applicable when regular expression and "cols" is used.
#' If \code{TRUE}, return the non-matched pattern. Default uses \code{FALSE}.
#' @param rm.dup Should duplicated columns be removed? Defaults to \code{TRUE}.
#' @return data.table
#' @seealso \code{\link[dplyr]{select}}, \code{\link[dplyr]{select_if}}
#' @examples
#' iris %>% select_dt(Species)
#' iris %>% select_dt(Sepal.Length,Sepal.Width)
#' iris %>% select_dt(Sepal.Length:Petal.Length)
#' iris %>% select_dt(-Sepal.Length)
#' iris %>% select_dt(-Sepal.Length,-Petal.Length)
#' iris %>% select_dt(-(Sepal.Length:Petal.Length))
#' iris %>% select_dt(c("Sepal.Length","Sepal.Width"))
#' iris %>% select_dt(-c("Sepal.Length","Sepal.Width"))
#' iris %>% select_dt(1)
#' iris %>% select_dt(-1)
#' iris %>% select_dt(1:3)
#' iris %>% select_dt(-(1:3))
#' iris %>% select_dt(1,3)
#' iris %>% select_dt("Pe")
#' iris %>% select_dt(-"Se")
#' iris %>% select_dt(!"Se")
#' iris %>% select_dt("Pe",negate = TRUE)
#' iris %>% select_dt("Pe|Sp")
#' iris %>% select_dt(cols = 2:3)
#' iris %>% select_dt(cols = 2:3,negate = TRUE)
#' iris %>% select_dt(cols = c("Sepal.Length","Sepal.Width"))
#' iris %>% select_dt(cols = names(iris)[2:3])
#'
#' iris %>% select_dt(is.factor)
#' iris %>% select_dt(-is.factor)
#' iris %>% select_dt(!is.factor)
#'
#' # select_mix could provide flexible mix selection
#' select_mix(iris, Species,"Sepal.Length")
#' select_mix(iris,1:2,is.factor)
#'
#' select_mix(iris,Sepal.Length,is.numeric)
#' # set rm.dup to FALSE could save the duplicated column names
#' select_mix(iris,Sepal.Length,is.numeric,rm.dup = FALSE)
#'


#' @rdname select
#' @export

select_dt = function(.data,...,cols = NULL,negate =FALSE){
  dt = as_dt(.data)
  if(is.null(cols)){
    substitute(list(...)) %>%
      deparse() %>%
      paste0(collapse = "") %>%
      str_extract("\\(.+\\)") %>%
      str_sub(2,-2)-> dot_string
    if(is.na(dot_string)) dt
    else if(str_detect(dot_string,"^[-,0-9 ]+$"))
      eval(parse(text = str_glue("dt[,c({dot_string})]")))
    else if(str_detect(dot_string,"^[-!]?\"")){
      if(dot_string %like% "^[-!]") {
        dot_string = str_remove(dot_string,"^-|^!")
        negate = TRUE
      }
      str_remove_all(dot_string,"\"") %>%
        str_subset(names(dt),.,negate = negate) %>%
        str_c(collapse = ",") -> dot_string
      eval(parse(text = str_glue("dt[,.({dot_string})]")))
    }
    else if(str_detect(dot_string,"^[-!]?c\\(")|str_count(dot_string,":") == 1)
      eval(parse(text = str_glue("dt[,{dot_string}]")))
    else if(dot_string %like% "^[-!]?is\\.")
      eval(parse(text = str_glue("select_if_dt(dt,{dot_string})")))
    else if(str_detect(dot_string,"^-")){
      dot_string = str_remove_all(dot_string,"-") %>% str_squish()
      if(!str_detect(dot_string,","))
        dt[,.SD,.SDcols = - dot_string]
      else{
        str_split(dot_string,",",simplify = TRUE) %>%
          str_squish()-> delete_names
        dt[,.SD,.SDcols = -delete_names]
      }
    }
    else eval(parse(text = str_glue("dt[,.({dot_string})]")))
  }
  else {
    if(!negate) dt[,.SD,.SDcols = cols]
    else eval(dt[, .SD,.SDcols = !cols])
  }
}


select_if_dt = function(dt,.if){
  if_name = substitute(.if) %>% deparse()
  if(str_detect(if_name,"^[-!]")){
    .if = str_remove(if_name,"[-!]")
    eval(parse(text = str_glue("sel_name = subset(sapply(dt,{.if}),
                        sapply(dt,{.if}) == FALSE) %>% names()")))
  } else
    sel_name = subset(sapply(dt,.if),sapply(dt,.if) == TRUE) %>% names()

  dt[,.SD, .SDcols = sel_name]
}

#' @rdname select
#' @export
select_mix = function(.data,...,rm.dup = TRUE){
  dt = as_dt(.data)

  substitute(list(...)) %>%
    lapply(deparse) %>%
    .[-1] %>%
    lapply(function(col) eval(parse(text = str_glue("select_dt(dt,{col})")))) %>%
    Reduce(f = cbind,x = .) -> res

  if(rm.dup) res = res[,unique(names(res)),with=FALSE]

  res

}

#' @title Use UTF-8 for character encoding in a data frame
#' @description \code{fread} from \pkg{data.table} could not recognize the encoding
#' and return the correct form, this could be unconvenient for text mining tasks. The
#' \code{utf8-encoding} could use "UTF-8" as the encoding to override the current
#' encoding of characters in a data frame.
#' @param .data A data.frame.
#' @return A data.table with characters in UTF-8 encoding

#' @export


utf8_encoding = function(.data){
  dt = as_dt(.data)
  mutate_vars(dt,is.character,str_conv,encoding = "UTF-8")
}

#' @title Convenient print of time taken
#' @name sys_time_print
#' @description Convenient printing of time elapsed. A wrapper of
#' \code{data.table::timetaken}, but showing the results more directly.
#' @param expr Valid R expression to be timed.
#' @return  A character vector of the form HH:MM:SS,
#' or SS.MMMsec if under 60 seconds (invisibly for \code{show_time}). See examples.
#' @seealso \code{\link[data.table]{timetaken}}, \code{\link[base]{system.time}}
#' @examples
#'
#' sys_time_print(Sys.sleep(1))
#'
#' a = iris
#' sys_time_print({
#'   res = iris %>%
#'     mutate_dt(one = 1)
#' })
#' res

#' @rdname sys_time_print
#' @export
sys_time_print = function (expr) {
  started.at = proc.time()
  eval(substitute(expr), envir = parent.frame())
  paste("# Finished in", timetaken(started.at))
}

#' @rdname sys_time_print
#' @export
pst = sys_time_print

.onAttach = function(...) {
    options(
        "datatable.print.class" = TRUE, # print class in data.table
        "datatable.print.trunc.cols" = TRUE,
        "datatable.print.keys" = TRUE
    )
    # hints = c(
    #     "Life's short, use R."
    # )
    # curr_hint = sample(hints, 1)
    # packageStartupMessage(paste0("\n", curr_hint, "\n"))

  packageStartupMessage("Thank you for using tidyfst!")
  packageStartupMessage("To acknowledge our work, please cite the package:")
  packageStartupMessage("Huang et al., (2020). tidyfst: Tidy Verbs for Fast Data Manipulation. Journal of Open Source Software, 5(52), 2388, https://doi.org/10.21105/joss.02388")
}

#' @title Nest and unnest
#' @description Create or melt list columns in data.frame.
#' @description Analogous function for \code{nest} and \code{unnest} in \pkg{tidyr}.
#' \code{unnest_dt} will automatically remove other list-columns except for the
#' target list-columns (which would be unnested later). Also, \code{squeeze_dt} is
#' designed to merge multiple columns into list column.
#' @param .data data.table, nested or unnested
#' @param ... The variables for nest group(for \code{nest_dt}),
#' columns to be nested(for \code{squeeze_dt} and \code{chop_dt}),
#' or column(s) to be unnested(for \code{unnest_dt}).
#' Could recieve anything that \code{\link[tidyfst]{select_dt}} could receive.
#' @param mcols Name-variable pairs in the list, form like
#' @param .name Character. The nested column name. Defaults to "ndt".
#' \code{list(petal="^Pe",sepal="^Se")}, see example.
#' @return data.table, nested or unnested
#' @details In the \code{nest_dt}, the data would be nested to a column named `ndt`,
#'  which is short for nested data.table.
#' @details The \code{squeeze_dt} would not remove the originial columns.
#' @details The \code{unchop_dt} is the reverse operation of \code{chop_dt}.
#' @details These functions are experiencing the experimental stage, especially
#' the \code{unnest_dt}. If they don't work on some circumtances, try \pkg{tidyr}
#' package.
#' @seealso \code{\link[tidyr]{nest}}, \code{\link[tidyr]{chop}}
#' @references https://www.r-bloggers.com/much-faster-unnesting-with-data-table/
#' @references https://stackoverflow.com/questions/25430986/create-nested-data-tables-by-collapsing-rows-into-new-data-tables
#' @examples
#'
#' # examples for nest_dt
#' # nest by which columns?
#'  mtcars %>% nest_dt(cyl)
#'  mtcars %>% nest_dt("cyl")
#'  mtcars %>% nest_dt(cyl,vs)
#'  mtcars %>% nest_dt(vs:am)
#'  mtcars %>% nest_dt("cyl|vs")
#'  mtcars %>% nest_dt(c("cyl","vs"))
#'
#'  # change the nested column name
#'  mtcars %>% nest_dt(cyl,.name = "data")
#'
#' # nest two columns directly
#' iris %>% nest_dt(mcols = list(petal="^Pe",sepal="^Se"))
#'
#' # nest more flexibly
#' iris %>% nest_dt(mcols = list(ndt1 = 1:3,
#'   ndt2 = "Pe",
#'   ndt3 = Sepal.Length:Sepal.Width))
#'
#' # examples for unnest_dt
#' # unnest which column?
#'  mtcars %>% nest_dt("cyl|vs") %>%
#'    unnest_dt(ndt)
#'  mtcars %>% nest_dt("cyl|vs") %>%
#'    unnest_dt("ndt")
#'
#' df <- data.table(
#'   a = list(c("a", "b"), "c"),
#'   b = list(c(TRUE,TRUE),FALSE),
#'   c = list(3,c(1,2)),
#'   d = c(11, 22)
#' )
#'
#' df
#' df %>% unnest_dt(a)
#' df %>% unnest_dt(2)
#' df %>% unnest_dt("c")
#' df %>% unnest_dt(cols = names(df)[3])
#'
#' # You can unnest multiple columns simultaneously
#' df %>% unnest_dt(1:3)
#' df %>% unnest_dt(a,b,c)
#' df %>% unnest_dt("a|b|c")
#'
#' # examples for squeeze_dt
#' # nest which columns?
#' iris %>% squeeze_dt(1:2)
#' iris %>% squeeze_dt("Se")
#' iris %>% squeeze_dt(Sepal.Length:Petal.Width)
#' iris %>% squeeze_dt(1:2,.name = "data")
#'
#' # examples for chop_dt
#' df <- data.table(x = c(1, 1, 1, 2, 2, 3), y = 1:6, z = 6:1)
#' df %>% chop_dt(y,z)
#' df %>% chop_dt(y,z) %>% unchop_dt(y,z)

#' @rdname nest
#' @export

# nest by which columns?

nest_dt = function(.data,...,mcols = NULL,.name = "ndt"){

  dt = as_dt(.data)

  if(substitute(mcols) %>% deparse() == "NULL")
    setnames(nest_by(dt,...),old = "ndt",new = .name)[]
  else{
    name_list = substitute(mcols)%>%
      lapply(deparse) %>%
      .[-1] %>%
      lapply(function(x)
        eval(parse(text = str_glue("names(select_dt(dt[0],{x}))"))))
    group_names = setdiff(names(dt),unique(unlist(name_list)))
    lapply(name_list,function(x) c(group_names,x)) %>%
      lapply(function(x) nest_by(dt,cols = group_names)) -> list_table
    for(i in seq_along(list_table)){
      list_table[[i]] = setnames(list_table[[i]],
                                 old = "ndt",new = names(list_table[i]))
    }

    Reduce(f = merge, x = list_table)
  }

}


nest_by = function(.data,...){
  dt = as_dt(.data)
  dt[0] %>% select_dt(...) %>% names() %>%
    str_c(collapse = ",")-> group
  eval(parse(text = str_glue("dt[,.(ndt = list(.SD)),by = .({group})]")))
}

#' @rdname nest
#' @export

# unnest which column(s)?

unnest_dt = function(.data,...){
  dt = as_dt(.data)
  col_names = dt[0] %>% select_dt(...) %>% names()
  if(length(col_names) == 1) unnest_col(dt,...)
  else
    lapply(col_names,function(x) unnest_col(dt,cols = x)) %>%
    Reduce(x = ., f = function(x,y) cbind(x,y)) %>%
    .[, which(duplicated(names(.))) := NULL] %>% .[]
}

# unnest_dt = function(.data,...){
#   dt = as_dt(.data)
#   col_names = dt[0] %>% select_dt(...) %>% names()
#   if(length(col_names) == 1) unnest_col(dt,...)
#   else
#     lapply(col_names,function(x) unnest_col(dt,cols = x)) %>%
#     Reduce(x = ., f = function(x,y) merge(x,y,all = TRUE))
# }

unnest_col = function(.data,...){
  dt = as_dt(.data)
  col_name = dt[0] %>% select_dt(...) %>% names()
  lapply(dt,class) -> dt_class
  names(subset(dt_class,dt_class != "list")) -> valid_col_names
  if(!col_name %chin% names(dt)) stop("The column does not exist.")
  valid_col_names %>%
    str_c(collapse = ",") %>%
    str_c("list(",.,")") -> group_name
  dt[[col_name]][[1]] -> first_element
  if(is.vector(first_element))
    eval(parse(text = str_glue("dt[,.({col_name} = unlist({col_name},recursive = FALSE)),by = {group_name}]")))
  else
    eval(parse(text = str_glue("dt[,unlist({col_name},recursive = FALSE),by = {group_name}]")))
}


#' @rdname nest
#' @export

# nest which columns?

squeeze_dt = function(.data,...,.name = "ndt"){
  dt = as.data.table(.data)
  dt %>% select_dt(...) %>%
    setNames(NULL) %>%
    apply(1,list) %>%
    lapply(unlist)-> ndt
  dt[,(.name) := ndt][]
}

#' @rdname nest
#' @export
chop_dt = function(.data,...){
  dt = as_dt(.data)
  dt[0] %>% select_dt(...) %>% names() -> data_cols
  setdiff(names(dt),data_cols) -> group_cols
  group_cols %>%
    str_c(collapse = ",") %>%
    str_c("list(",.,")") -> group_names
  eval(parse(text = str_glue("dt[,lapply(.SD,list),
                             by = {group_names}]")))
}

#' @rdname nest
#' @export
unchop_dt = function(.data,...){
  dt = as_dt(.data)
  col_names = dt[0] %>% select_dt(...) %>% names()
  group_names = setdiff(names(dt),col_names)
  if(length(col_names) == 1) unnest_col(dt,...)
  else
    lapply(col_names,function(x) unnest_col(dt,cols = x)) %>%
    Reduce(x = ., f = function(x,y) cbind(x,y)) %>%
    .[,unique(names(.)),with=FALSE]
}

#' @title Get the column name of the max/min number each row
#' @description For a data.frame with numeric values, add a new column
#' specifying the column name of the first max/min value each row.
#' @param .data A data.frame with numeric column(s)
#' @param ... Variables for screening, could receive what `select_dt` receives.
#' When starts with '-'(minus symbol) or '!', return the negative columns.
#' @param .name The column name of the new added column
#' @return A data.table
#' @references https://stackoverflow.com/questions/17735859/for-each-row-return-the-column-name-of-the-largest-value
#'
#' @examples
#' set.seed(199057)
#' DT <- data.table(matrix(sample(10, 100, TRUE), ncol=10))
#' DT
#' col_max(DT)
#' col_max(DT,V1:V3)
#' col_max(DT,.name = "max_col_name")
#' col_min(DT)
#' col_min(DT,2:4)
#'
#' col_max(iris)
#'


#' @rdname col_max
#' @export
col_max = function(.data,...,.name = "max_col"){
  dt = as.data.table(.data)
  dt[0] %>% select_dt(...) %>% names() -> dot_string
  if(setequal(dot_string,names(dt))){
    dt[, (.name) := colnames(.SD)[max.col(.SD, ties.method = "first")],
       .SDcols = is.numeric][]
  }else{
    dt[, (.name) := colnames(.SD)[max.col(.SD, ties.method = "first")],
       .SDcols = dot_string][]
  }
}

#' @rdname col_max
#' @export
col_min = function(.data,...,.name = "min_col"){
  dt = as.data.table(.data)
  dt[0] %>% select_dt(...) %>% names() -> dot_string
  if(setequal(dot_string,names(dt))){
    dt[, (.name) := colnames(.SD)[max.col(-.SD, ties.method = "first")],
       .SDcols = is.numeric][]
  }else{
    dt[, (.name) := colnames(.SD)[max.col(-.SD, ties.method = "first")],
       .SDcols = dot_string][]
  }
}



# past code
# col_max = function(.data,.name = "max_col"){
#   DT = as.data.table(.data)
#   DT[, (.name) := colnames(.SD)[max.col(.SD, ties.method = "first")],
#      .SDcols = is.numeric][]
# }
# col_min = function(.data,.name = "min_col"){
#   DT = as.data.table(.data)
#   DT[, (.name) := colnames(.SD)[max.col(-.SD, ties.method = "first")],
#      .SDcols = is.numeric][]
# }

#' @title Add percentage to counts in data.frame
#' @description Add percentage for counts in the data.frame, both numeric and
#'  character with `%` would be provided.
#' @param x A number (numeric).
#' @param digits How many digits to keep in the percentage. Default uses 1.
#' @param .data A data frame.
#' @param count_name Column name of counts (Character).
#'  Default uses the last column of data.frame.
#' @references https://stackoverflow.com/questions/7145826/how-to-format-a-number-as-percentage-in-r
#' @examples
#'
#'  percent(0.9057)
#'  percent(0.9057,3)
#'
#'  iris %>%
#'    count_dt(Species) %>%
#'    add_prop()
#'
#'  iris %>%
#'    count_dt(Species) %>%
#'    add_prop(count_name = "n",digits = 2)
#'

#' @rdname add_prop
#' @export
percent = function(x,digits = 1){
  sprintf(paste0("%.",digits,"f%%"),x * 100)
}

#' @rdname add_prop
#' @export
add_prop = function(.data,count_name = last(names(.data)),digits = 1){
  dt = as.data.table(.data)
  dt[,prop:=dt[[count_name]]/sum(dt[[count_name]])][,
    prop_label:= percent(prop,digits = digits)
  ][]
}

globalVariables(c("prop","prop_label"))

#' @title Change column order
#' @description Change the position of columns,
#'  using the same syntax as `select_dt()`. Check similar function
#'  as `relocate` in \pkg{dplyr}.
#' @param .data A data.frame
#' @param ... Columns to move
#' @param how The mode of movement, including "first","last","after","before".
#' Default uses "first".
#' @param where Destination of columns selected by \code{...}.
#' Applicable for "after" and "before" mode.
#' @return A data.table with rearranged columns.
#' @seealso \code{\link[dplyr]{relocate}}

#' @examples
#' df <- data.table(a = 1, b = 1, c = 1, d = "a", e = "a", f = "a")
#' df
#' df %>% relocate_dt(f)
#' df %>% relocate_dt(a,how = "last")
#'
#' df %>% relocate_dt(is.character)
#' df %>% relocate_dt(is.numeric, how = "last")
#' df %>% relocate_dt("[aeiou]")
#'
#' df %>% relocate_dt(a, how = "after",where = f)
#' df %>% relocate_dt(f, how = "before",where = a)
#' df %>% relocate_dt(f, how = "before",where = c)
#' df %>% relocate_dt(f, how = "after",where = c)
#'
#' df2 <- data.table(a = 1, b = "a", c = 1, d = "a")
#' df2 %>% relocate_dt(is.numeric,
#'                     how = "after",
#'                     where = is.character)
#' df2 %>% relocate_dt(is.numeric,
#'                     how="before",
#'                     where = is.character)

#' @export

relocate_dt = function(.data,...,
                       how= "first",
                       where = NULL){
  #dt = as_dt(.data)
  dt = as.data.table(.data)

  dt[0] %>% select_dt(...) %>% names() -> sel_names
  names(dt[0]) %>% setdiff(sel_names) -> rest_names
  dt[0] %>% select_dt(cols = rest_names) -> rest_dt
  substitute(where) %>% deparse() -> where_n
  if(str_detect(where_n,"^\"|^'")) where_n = where

  if(how == "first") c(sel_names,rest_names) -> name_order
  else if(how == "last") c(rest_names,sel_names) -> name_order
  else if(where_n == "NULL") stop("The `where` parameter should be provided.")
  else
  {
    if(where_n %like% "^is\\."){
      eval(parse(text = str_glue("dt1 = select_dt(rest_dt,{where_n}) %>% names")))
      eval(parse(text = str_glue("dt2 = select_dt(rest_dt,-{where_n}) %>% names")))
      if(how == "after") c(dt1,sel_names,dt2) -> name_order
      else c(sel_names,dt1,dt2) -> name_order
    }
    else{
      chmatch(where_n,rest_names) -> position
      if(how == "after"){
        if(position < length(rest_names))
          c(rest_names[1:position], sel_names,
            rest_names[(position+1):length(rest_names)]) -> name_order
        else c(rest_names,sel_names) -> name_order
      }
      else if(how == "before"){
        if(position > 1)
          c(rest_names[1:(position-1)],sel_names,
            rest_names[position:length(rest_names)]) -> name_order
        else c(sel_names,rest_names) -> name_order
      }
      else stop("The `how` parameter could not be recognized.")
    }
  }
  setcolorder(dt,neworder = name_order)[]
}

# relocate_dt = function(.data,...,
#                        how= "first",
#                        where = NULL){
#   dt = as_dt(.data)
#
#   dt %>% select_dt(...) -> sel_dt
#   names(sel_dt) -> sel_names
#   names(dt) %>% setdiff(sel_names) -> rest_names
#   dt %>% select_dt(cols = rest_names) -> rest_dt
#   substitute(where) %>% deparse() -> where_n
#   if(str_detect(where_n,"^\"|^'")) where_n = where
#
#   if(how == "first") c(sel_dt,rest_dt) %>% as.data.table()
#   else if(how == "last") c(rest_dt,sel_dt) %>% as.data.table()
#   else if(where_n == "NULL") stop("The `where` parameter should be provided.")
#   else
#   {
#     if(where_n %like% "^is\\."){
#       eval(parse(text = str_glue("dt1 = select_dt(rest_dt,{where_n})")))
#       eval(parse(text = str_glue("dt2 = select_dt(rest_dt,-{where_n})")))
#       if(how == "after") cbind(dt1,sel_dt,dt2) %>% as.data.table()
#       else cbind(sel_dt,dt1,dt2) %>% as.data.table()
#     }
#     else{
#       chmatch(where_n,names(rest_dt)) -> position
#       if(how == "after"){
#         if(position < ncol(rest_dt))
#           as.data.table(cbind(rest_dt[,1:position],
#                               sel_dt,
#                               rest_dt[,(position+1):ncol(rest_dt)]))
#         else c(rest_dt,sel_dt) %>% as.data.table()
#       }
#       else if(how == "before"){
#         if(position > 1)
#           as.data.table(cbind(rest_dt[,1:(position-1)],
#                               sel_dt,
#                               rest_dt[,position:ncol(rest_dt)]))
#         else c(sel_dt,rest_dt) %>% as.data.table()
#       }
#       else stop("The `how` parameter could not be recognized.")
#     }
#   }
# }

globalVariables(c("dt1","dt2"))

#' @title Join tables
#' @name join
#' @description  The mutating joins add columns from `y` to `x`,
#' matching rows based on the keys:
#'
#' * `inner_join_dt()`: includes all rows in `x` and `y`.
#' * `left_join_dt()`: includes all rows in `x`.
#' * `right_join_dt()`: includes all rows in `y`.
#' * `full_join_dt()`: includes all rows in `x` or `y`.
#' @description
#' Filtering joins filter rows from `x` based on the presence or absence
#' of matches in `y`:
#'
#' * `semi_join_dt()` return all rows from `x` with a match in `y`.
#' * `anti_join_dt()` return all rows from `x` without a match in `y`.
#' @param x A data.table
#' @param y A data.table
#' @param by (Optional) A character vector of variables to join by.
#'
#'   If `NULL`, the default, `*_join_dt()` will perform a natural join, using all
#'   variables in common across `x` and `y`. A message lists the variables so that you
#'   can check they're correct; suppress the message by supplying `by` explicitly.
#'
#'   To join by different variables on `x` and `y`, use a named vector.
#'   For example, `by = c("a" = "b")` will match `x$a` to `y$b`.
#'
#'   To join by multiple variables, use a vector with length > 1.
#'   For example, `by = c("a", "b")` will match `x$a` to `y$a` and `x$b` to
#'   `y$b`. Use a named vector to match different variables in `x` and `y`.
#'   For example, `by = c("a" = "b", "c" = "d")` will match `x$a` to `y$b` and
#'   `x$c` to `y$d`.
#' @param on  (Optional)
#' Indicate which columns in x should be joined with which columns in y.
#' Examples included:
#'   1.\code{.by = c("a","b")} (this is a must for \code{set_full_join_dt});
#'   2.\code{.by = c(x1="y1", x2="y2")};
#'   3.\code{.by = c("x1==y1", "x2==y2")};
#'   4.\code{.by = c("a", V2="b")};
#'   5.\code{.by = .(a, b)};
#'   6.\code{.by = c("x>=a", "y<=b")} or \code{.by = .(x>=a, y<=b)}.
#' @param suffix If there are non-joined duplicate variables in x and y, these
#'   suffixes will be added to the output to disambiguate them. Should be a
#'   character vector of length 2.
#' @return A data.table
#' @examples
#'
#' workers = fread("
#'     name company
#'     Nick Acme
#'     John Ajax
#'     Daniela Ajax
#' ")
#'
#' positions = fread("
#'     name position
#'     John designer
#'     Daniela engineer
#'     Cathie manager
#' ")
#'
#' workers %>% inner_join_dt(positions)
#' workers %>% left_join_dt(positions)
#' workers %>% right_join_dt(positions)
#' workers %>% full_join_dt(positions)
#'
#' # filtering joins
#' workers %>% anti_join_dt(positions)
#' workers %>% semi_join_dt(positions)
#'
#' # To suppress the message, supply 'by' argument
#' workers %>% left_join_dt(positions, by = "name")
#'
#' # Use a named 'by' if the join variables have different names
#' positions2 = setNames(positions, c("worker", "position")) # rename first column in 'positions'
#' workers %>% inner_join_dt(positions2, by = c("name" = "worker"))
#'
#' # the syntax of 'on' could be a bit different
#' workers %>% inner_join_dt(positions2,on = "name==worker")
#'
#'

#' @rdname join
#' @export
inner_join_dt = function(x,y,by = NULL, on = NULL,suffix = c(".x",".y")){
  x = as_dt(x)
  y = as_dt(y)
  on_ = substitute(on) %>% deparse()
  by_ = substitute(by) %>% deparse()
  if(on_ != "NULL") x[y, nomatch = 0L, on = on]
  else if(by_ == "NULL"){
    by = intersect(names(x), names(y))
    by_name = str_c(by, collapse = ",")
    message(str_glue("Joining by: {by_name}\n\n"))
    merge.data.table(x,y,by = by,suffixes = suffix)
  }else if(is.null(names(by))) merge.data.table(x,y,by = by,suffixes = suffix)
  else merge.data.table(x,y,by.x = names(by),by.y = by,suffixes = suffix)
}

#' @rdname join
#' @export
left_join_dt = function(x,y,by = NULL, on = NULL,suffix = c(".x",".y")){
  x = as_dt(x)
  y = as_dt(y)
  on_ = substitute(on) %>% deparse()
  by_ = substitute(by) %>% deparse()
  if(on_ != "NULL") y[x, on = on]
  else if(by_ == "NULL"){
    by = intersect(names(x), names(y))
    by_name = str_c(by, collapse = ",")
    message(str_glue("Joining by: {by_name}\n\n"))
    merge.data.table(x,y,by = by,all.x = TRUE,suffixes = suffix)
  }else if(is.null(names(by))) merge.data.table(x,y,by = by,all.x = TRUE,suffixes = suffix)
  else merge.data.table(x,y,by.x = names(by),by.y = by,all.x = TRUE,suffixes = suffix)
}

#' @rdname join
#' @export
right_join_dt = function(x,y,by = NULL, on = NULL,suffix = c(".x",".y")){
  x = as_dt(x)
  y = as_dt(y)
  on_ = substitute(on) %>% deparse()
  by_ = substitute(by) %>% deparse()
  if(on_ != "NULL") x[y, on = on]
  else if(by_ == "NULL"){
    by = intersect(names(x), names(y))
    by_name = str_c(by, collapse = ",")
    message(str_glue("Joining by: {by_name}\n\n"))
    merge.data.table(x,y,by = by,all.y = TRUE,suffixes = suffix)
  }else if(is.null(names(by))) merge.data.table(x,y,by = by,all.y = TRUE,suffixes = suffix)
  else merge.data.table(x,y,by.x = names(by),by.y = by,all.y = TRUE,suffixes = suffix)
}

#' @rdname join
#' @export
full_join_dt = function(x,y,by = NULL, on = NULL,suffix = c(".x",".y")){
  x = as_dt(x)
  y = as_dt(y)
  on_ = substitute(on) %>% deparse()
  by_ = substitute(by) %>% deparse()
  if(on_ != "NULL") {
    if(by_!="null"){
      rbind(x[, .SD, .SDcols = by],
            y[, .SD, .SDcols = by]) %>%
        unique()-> unique_keys
      y[x[.(unique_keys), on = on], on = on]
    }else{
      rbind(x[, .SD, .SDcols = on],
            y[, .SD, .SDcols = on]) %>%
        unique()-> unique_keys
      y[x[.(unique_keys), on = on], on = on]
    }
  } else if(by_ == "NULL"){
    by = intersect(names(x), names(y))
    by_name = str_c(by, collapse = ",")
    message(str_glue("Joining by: {by_name}\n\n"))
    merge.data.table(x,y,by = by,all = TRUE,suffixes = suffix)
  }else if(is.null(names(by))) merge.data.table(x,y,by = by,all = TRUE,suffixes = suffix)
  else merge.data.table(x,y,by.x = names(by),by.y = by,all = TRUE,suffixes = suffix)
}

#' @rdname join
#' @export
anti_join_dt = function(x,y,by = NULL, on = NULL){
  x = as_dt(x)
  y = as_dt(y)
  on_ = substitute(on) %>% deparse()
  by_ = substitute(by) %>% deparse()
  if(on_ != "NULL") x[!y, on = on]
  else if(by_ == "NULL"){
    by = intersect(names(x), names(y))
    by_name = str_c(by, collapse = ",")
    message(str_glue("Joining by: {by_name}\n\n"))
    x[!y, on = by]
  }else x[!y, on = by]
}

#' @rdname join
#' @export
semi_join_dt = function(x,y,by = NULL, on = NULL){
  x = as_dt(x)
  y = as_dt(y)
  on_ = substitute(on) %>% deparse()
  by_ = substitute(by) %>% deparse()
  if(on_ != "NULL") {
    w = unique(x[y, on = on, nomatch = 0L, which = TRUE, allow.cartesian = TRUE])
    x[w]
  }
  else{
    if(by_ == "NULL"){
      by = intersect(names(x), names(y))
      by_name = str_c(by, collapse = ",")
      message(str_glue("Joining by: {by_name}\n\n"))
    }
    w = unique(x[y, on = by, nomatch = 0L, which = TRUE, allow.cartesian = TRUE])
    x[w]
  }
}

#' @title Rename column in data.frame
#' @description Rename one or more columns in the data.frame.
#' @param .data data.frame
#' @param ... statements of rename, e.g. `sl = Sepal.Length` means the column named
#' as "Sepal.Length" would be renamed to "sl"
#' @param .fn A function used to transform the selected columns.
#'  Should return a character vector the same length as the input.
#' @return data.table
#' @seealso \code{\link[dplyr]{rename}}
#' @examples
#' iris %>%
#'   rename_dt(sl = Sepal.Length,sw = Sepal.Width) %>%
#'   head()
#' iris %>% rename_with_dt(toupper)
#' iris %>% rename_with_dt(toupper,"^Pe")
#'
#' @rdname rename_dt
#' @export

rename_dt = function(.data,...){
  dt = as.data.table(.data)
  substitute(list(...)) %>%
    lapply(deparse) %>%
    .[-1] -> dot_string
  new_names = names(dot_string)
  old_names = as.character(dot_string)
  setnames(dt,old = old_names,new = new_names)[]
}

#' @rdname rename_dt
#' @export
rename_with_dt = function(.data,.fn,...){
  dt = as.data.table(.data)
  dt[0] %>% select_dt(...) %>% names() -> old_names
  .fn -> new_names
  setnames(dt,old = old_names,new = new_names)[]
}

#' @title Subset rows using their positions
#' @description
#' `slice_dt()` lets you index rows by their (integer) locations. It allows you
#' to select, remove, and duplicate rows. It is accompanied by a number of
#' helpers for common use cases:
#'
#' * `slice_head_dt()` and `slice_tail_dt()` select the first or last rows.
#' * `slice_sample_dt()` randomly selects rows.
#' * `slice_min_dt()` and `slice_max_dt()` select rows with highest or lowest values
#'   of a variable.
#' @param .data A data.table
#' @param ... Provide either positive values to keep, or negative values to drop.
#'   The values provided must be either all positive or all negative.
#' @param by Slice by which group(s)?
#' @param n When larger than or equal to 1, the number of rows.
#'   When between 0 and 1, the proportion of rows to select.
#'
#' @param order_by Variable or function of variables to order by.
#'
#' @param with_ties Should ties be kept together? The default, `TRUE`,
#'   may return more rows than you request. Use `FALSE` to ignore ties,
#'   and return the first `n` rows.
#'
#' @param replace Should sampling be performed with (`TRUE`) or without
#'   (`FALSE`, the default) replacement.
#'
#' @return A data.table
#' @seealso \code{\link[dplyr]{slice}}
#' @examples
#'
#' a = iris
#' slice_dt(a,1,2)
#' slice_dt(a,2:3)
#' slice_dt(a,141:.N)
#' slice_dt(a,1,.N)
#' slice_head_dt(a,5)
#' slice_head_dt(a,0.1)
#' slice_tail_dt(a,5)
#' slice_tail_dt(a,0.1)
#' slice_max_dt(a,Sepal.Length,10)
#' slice_max_dt(a,Sepal.Length,10,with_ties = FALSE)
#' slice_min_dt(a,Sepal.Length,10)
#' slice_min_dt(a,Sepal.Length,10,with_ties = FALSE)
#' slice_sample_dt(a,10)
#' slice_sample_dt(a,0.1)
#'
#'
#' # use by to slice by group
#'
#' ## following codes get the same results
#' slice_dt(a,1:3,by = "Species")
#' slice_dt(a,1:3,by = Species)
#' slice_dt(a,1:3,by = .(Species))
#'
#' slice_head_dt(a,2,by = Species)
#' slice_tail_dt(a,2,by = Species)
#'
#' slice_max_dt(a,Sepal.Length,3,by = Species)
#' slice_max_dt(a,Sepal.Length,3,by = Species,with_ties = FALSE)
#' slice_min_dt(a,Sepal.Length,3,by = Species)
#' slice_min_dt(a,Sepal.Length,3,by = Species,with_ties = FALSE)
#'
#' # in `slice_sample_dt`, "by" could only take character class
#' slice_sample_dt(a,.1,by = "Species")
#' slice_sample_dt(a,3,by = "Species")
#' slice_sample_dt(a,51,replace = TRUE,by = "Species")
#'

#' @rdname slice
#' @export
slice_dt = function(.data,...,by = NULL){
  dt = as_dt(.data)
  substitute(by) %>% deparse() -> by_char
  if(by_char == "NULL") dt[eval(substitute(c(...)))]
  else
    eval(substitute(
    dt[,.SD[c(...)],by]
  ))
}


# slice_dt = function(.data,...){
#   as_dt(.data)[eval(substitute(c(...)))]
# }

#' @rdname slice
#' @export
slice_head_dt = function(.data,n,by = NULL){
  .data = as_dt(.data)
  substitute(by) %>% deparse() -> by_char
  if(by_char == "NULL"){
    if(abs(n) >= 1L) head(.data,n)
    else head(.data,nrow(.data)*n)
  }else{
    if(abs(n) >= 1L)
      eval(substitute(
        .data[,head(.SD,n),by]
      ))
    else
      eval(substitute(
        .data[,head(.SD,nrow(.SD)*n),by]
      ))
  }
}


# slice_head_dt = function(.data,n){
#   .data = as_dt(.data)
#   if(abs(n) >= 1L) head(.data,n)
#   else head(.data,nrow(.data)*n)
# }

#' @rdname slice
#' @export
slice_tail_dt = function(.data,n,by = NULL){
  .data = as_dt(.data)
  substitute(by) %>% deparse() -> by_char
  if(by_char == "NULL"){
    if(abs(n) >= 1L) tail(.data,n)
    else tail(.data,nrow(.data)*n)
  }else{
    if(abs(n) >= 1L)
      eval(substitute(
        .data[,tail(.SD,n),by]
      ))
    else
      eval(substitute(
        .data[,tail(.SD,nrow(.SD)*n),by]
      ))
  }
}

# slice_tail_dt = function(.data,n){
#   .data = as_dt(.data)
#   if(abs(n) >= 1L) tail(.data,n)
#   else tail(.data,nrow(.data)*n)
# }

#' @rdname slice
#' @export
slice_max_dt = function(.data,order_by,n,by = NULL,with_ties = TRUE){
  dt = as_dt(.data)
  if(with_ties) ties_method = "min" else ties_method = "first"
  wt = substitute(order_by) %>% deparse()
  by = substitute(by) %>% deparse()
  if(n <= 0) stop("Invalid input, n should take a positive value.")

  # if(by == "NULL"){
  #   eval(parse(text = str_glue(
  #     " dt[frank(-{wt},ties.method = ties_method,na.last = 'keep')<=
  #   min(.N,ifelse(n<1,.N*n,n))][order(-{wt})]"
  #   )))
  # } else{
  #   eval(parse(text = str_glue(
  #     " dt[,.SD[frank(-{wt},ties.method = ties_method,na.last = 'keep')<=
  #   min(.N,ifelse(n<1,.N*n,n))],by = {by}]"
  #   )))
  # }

  if(by == "NULL"){
    eval(parse(text = str_glue(
      " dt[frankv({wt},ties.method = ties_method,order=-1,na.last = 'keep')<=
    min(.N,ifelse(n<1,.N*n,n))][order(-{wt})]"
    )))
  } else{
    eval(parse(text = str_glue(
      " dt[,.SD[frankv({wt},ties.method = ties_method,order=-1,na.last = 'keep')<=
    min(.N,ifelse(n<1,.N*n,n))],by = {by}]"
    )))
  }
}

#' @rdname slice
#' @export
slice_min_dt = function(.data,order_by,n,by = NULL,with_ties = TRUE){
  dt = as_dt(.data)
  if(with_ties) ties_method = "min" else ties_method = "first"
  wt = substitute(order_by) %>% deparse()
  by = substitute(by) %>% deparse()
  if(n <= 0) stop("Invalid input, n should take a positive value.")
  if(by == "NULL"){
    eval(parse(text = str_glue(
      " dt[frank({wt},ties.method = ties_method,na.last = 'keep')<=
    min(.N,ifelse(n<1,.N*n,n))][order({wt})]"
    )))
  } else{
    eval(parse(text = str_glue(
      " dt[,.SD[frank({wt},ties.method = ties_method,na.last = 'keep')<=
    min(.N,ifelse(n<1,.N*n,n))],by = {by}]"
    )))
  }
}

#' @rdname slice
#' @export
slice_sample_dt = function(.data, n, replace = FALSE, by = NULL){
  if(n >= 1L) sample_n_dt(.data,size = n,replace = replace,by = by)
  else if(n > 0) sample_frac_dt(.data,size = n,replace = replace,by = by)
  else stop("Invalid input, n should take a positive value.")
}

#' @title Pivot data from long to wide
#' @description Transform a data frame from long format to wide by increasing the number of columns and decreasing the number of rows.
#' @param .data A data.frame
#' @param ... Optional. The unchanged group in the transformation.
#' Could use integer vector, could receive what \code{select_dt} receives.
#' @param name Chracter.One column name of class to spread
#' @param value Chracter.One column name of value to spread.
#' If \code{NULL}, use all other variables.
#' @param fun Should the data be aggregated before casting?
#' Defaults to \code{NULL}, which uses \code{length} for aggregation.
#' If a function is provided, with aggregated by this function.
#' @param fill Value with which to fill missing cells. Default uses \code{NA}.
#' @details The parameter of `name` and `value` should always
#' be provided and should be explicit called (with the parameter names attached).
#' @return data.table
#' @seealso \code{\link[tidyfst]{longer_dt}},
#'  \code{\link[data.table]{dcast}},
#'  \code{\link[tidyr]{pivot_wider}}
#' @examples
#'  stocks = data.frame(
#'    time = as.Date('2009-01-01') + 0:9,
#'    X = rnorm(10, 0, 1),
#'    Y = rnorm(10, 0, 2),
#'    Z = rnorm(10, 0, 4)
#'  ) %>%
#'    longer_dt(time) -> longer_stocks
#'
#'  longer_stocks
#'
#'  longer_stocks %>%
#'    wider_dt("time",
#'             name = "name",
#'             value = "value")
#'
#'  longer_stocks %>%
#'    mutate_dt(one = 1) %>%
#'    wider_dt("time",
#'             name = "name",
#'             value = "one")
#'
#' ## using "fun" parameter for aggregation
#' DT <- data.table(v1 = rep(1:2, each = 6),
#'                  v2 = rep(rep(1:3, 2), each = 2),
#'                  v3 = rep(1:2, 6),
#'                  v4 = rnorm(6))
#' ## for each combination of (v1, v2), add up all values of v4
#' DT %>%
#'   wider_dt(v1,v2,
#'            value = "v4",
#'            name = ".",
#'            fun = sum)

#' @export

wider_dt = function(.data,
                    ...,
                    name,
                    value = NULL,
                    fun = NULL,
                    fill = NA){
  dt = as_dt(.data)
  group = dt[0] %>% select_mix(...) %>% names() %>% str_c(collapse = "+")
  if(group == "") group = "..."
  if(is.null(value)) value = "."

  substitute(fun) %>% deparse() -> fun
  if(fun == "list"){
    call_string = str_glue("dcast(dt,{group}~{name}, fun.aggregate = list,
                          value.var = value,fill = {fill}) %>%
                           unchop_dt(is.list)")
  }else{
    call_string = str_glue("dcast(dt,{group}~{name}, fun.aggregate = {fun},
                          value.var = value,fill = {fill})")
  }
  eval(parse(text = call_string))
}

#' @title Extract the nth value from a vector
#' @description \code{nth} get the value from a vector with its position,
#'  while \code{maxth} and \code{minth} get the nth highest or lowest value
#'  from the vector.
#' @param v A vector
#' @param n For\code{nth}, a single integer specifying the position. Default uses \code{1}.
#' Negative integers index from the end
#'  (i.e. -1L will return the last value in the vector).
#'  If a double is supplied, it will be silently truncated.
#'  For \code{maxth} and \code{minth}, a single integer indicating the nth
#'  highest or lowest value.
#' @return A single value.
#' @references https://stackoverflow.com/questions/2453326/fastest-way-to-find-second-third-highest-lowest-value-in-vector-or-column/66367996#66367996
#' @examples
#'
#' x = 1:10
#' nth(x, 1)
#' nth(x, 5)
#' nth(x, -2)
#'
#' y = c(10,3,4,5,2,1,6,9,7,8)
#' maxth(y,3)
#' minth(y,3)
#'

#' @rdname nth
#' @export

nth = function(v,n = 1){
  n <- trunc(n)
  ifelse(
    n >= 0,
    v[[n]],
    v[[length(v)+n+1]]
  )
}

#' @rdname nth
#' @export
maxth = function(v,n = 1){
  v = na.omit(v)
  no = length(v)
  if(n > no){
    warning("N greater than length(v). Setting n = length(v)")
    n = no
  }
  sort(v,partial = no - n + 1)[no - n + 1]
}

#' @rdname nth
#' @export
minth = function(v,n = 1){
  v = na.omit(v)
  no = length(v)
  if(n > no){
    warning("N greater than length(v). Setting n = length(v)")
    n = no
  }
  -sort(-v, partial=no - n + 1)[no - n + 1]
}

#' @title Round a number and make it show zeros
#' @description
#'  Rounds values in its first argument to the specified number of decimal places,
#'  returning character, ensuring first decimal digits are showed even when they are zeros.
#'
#' @param x A numeric vector.
#' @param digits Integer indicating the number of decimal places. Defaults to 0.
#'
#' @return A character vector.
#' @seealso \code{\link[base]{round}}
#' @references https://stackoverflow.com/questions/42105336/how-to-round-a-number-and-make-it-show-zeros
#' @examples
#' a = 14.0034
#' round0(a,2)
#'
#' b = 10
#' round0(b,1)
#'
#' @export

round0 = \(x,digits = 0){
  format(round(x,digits = digits),nsmall = digits)
}

---
title: "Example 4: Nest"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 4: Nest}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```
  This vignette displays how to use nesting in `tidyfst`. It has referred to `tidyr`s vignette in <https://tidyr.tidyverse.org/articles/nest.html>. Now fist, we nest the "mtcars" data.frame by "cyl" column.
```{r setup}
library(tidyfst)

# nest by "cyl" column
mtcars_nested <- mtcars %>% 
  nest_dt(cyl) # you can use "cyl" too, very flexible

# inspect the output data.table
mtcars_nested

```

  Now, we want to do a regression within the nested group "cyl". We'll use the famous `lapply` to complete this:

```{r}
mtcars_nested2 <- mtcars_nested %>% 
  mutate_dt(model = lapply(ndt,function(df) lm(mpg ~ wt, data = df)))

mtcars_nested2
```
  We could see that the model is stored in the column "model".
  Now, we try to get the fitted value in the model.
```{r}
mtcars_nested3 <- mtcars_nested2 %>% 
  mutate_dt(model_predict = lapply(model, predict))
mtcars_nested3$model_predict
```
  We could find that the "model_predict" is a list of numeric vectors. Let's try to unnest the target column "model_predict". 
```{r}
mtcars_nested3 %>% unnest_dt(model_predict)
```
  This process would remove all the other list column automatically. For instance, in our case, the column "ndt" is removed.

---
title: "tidyfst包实例分析"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{chinese_tutorial}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```


我的R语言小伙伴最近分享了自己使用R来做工业级数据清洗的[经验](https://mp.weixin.qq.com/s/NVlCPss32j6Ohdrc9Edx-A)，最近我自己在不断测试我的新包tidyfst，因此就拿这个data.table的案例来尝试一下。

## 测试数据构造
  本次测试，将不会加载data.table包，但是其实tidyfst里面无处不是data.table的元素，而且也导出了很多内置的data.table函数，比如as.data.table和data.table。所以这些代码在tidyfst中就可以自如地使用。
```{r}
library(tidyfst)
diamonds <- ggplot2::diamonds
n = 1e5  #如果想做工业级测试，可以继续增加数量
set.seed(2020)
dtranges <- seq.Date(from = as.Date("2011-01-01"),
                     to = as.Date("2020-01-01"),
                     by = 1)
n1 <- sample(nrow(diamonds), n, replace = TRUE)
dat1 <- as.data.table(diamonds[n1, ])
dat1[, "dt"] <- sample(dtranges, n, replace = TRUE)  # 增加dt列
n2 <- sample(nrow(dat1), nrow(dat1)/1000)
dat1[n2, "price"] <- NA # price列构造千分之一缺失值
dat2 <- data.table(dt = sample(dtranges, min(n/1000, length(dtranges))),
                   price1 = sample(1000, min(n/1000, length(dtranges)), replace = TRUE))

dat3 <- data.table(dt = sample(dtranges, min(n/1000, length(dtranges))),
                   price2 = sample(1000, min(n/1000, length(dtranges)), replace = TRUE))

print(dat1)
```

## 基础
### 小技巧
后面的分析，经常要根据日期进行计算。所以，先对日期进行排序，就能够提高运行速度。在tidyfst中，可以使用`arrange_dt`函数来对数据进行原位的各种操作，其中就包括排序。
```{r}
dat1 = arrange_dt(dat1,dt)
dat1
```
那么，现在dat1的数据就按照日期排好序了。

### 聚合
#### 1.求每种切割类型、每种颜色钻石的平均价格、中位数价格与最高价格
在tidyfst中，我设置了一个`sys_time_print`函数，可以方便地输出`system.time()`函数返回的结果。
```{r}
sys_time_print({
  r1_1 <- dat1 %>% 
    summarise_dt(
      by = .(cut,color),
      mean_price = mean(price, na.rm = TRUE),
      median_price = median(price, na.rm = TRUE),
      max_price = max(price, na.rm = TRUE)
    )
})
r1_1
```

tidyfst是永远不可能比data.table快的，但是如果你觉得上面的代码更容易掌握、更容易读懂，而在日常工作中多花零点几秒的运行时间没有太大问题（实际上节省了大家的交流时间，甚至就是节省将来自己再次读懂自己代码的时间），tidyfst就值得拥有。

#### 2.求每天最高出售价格对应的那笔订单

```{r}
sys_time_print({
  r1_2 <- dat1 %>% 
    arrange_dt(dt,-price) %>% 
    drop_na_dt(price) %>% 
    group_dt(
      by = dt,
      head(1)
    )
})
r1_2
```
### join
#### 1.dat1与dat2以dt列左连接
实质上，merge函数已经优化得很好。tidyfst设计`*_join`系列函数的时候，只是为了一种不一样的语法结构来帮助实现不同的连接，因为它确实更加直观一些。但是实质上它还是merge.data.table函数的包装版本。
```{r}
sys_time_print({
  r2_1 <- dat1 %>% 
    left_join_dt(dat2,by = "dt")
})
r2_1
```
#### 2.多重join
  
```{r}
sys_time_print({
  mymerge <- function(x, y) left_join_dt(x, y, by = "dt")
  r2_2 <- Reduce(mymerge, list(dat1, dat2, dat3))
})
r2_2
```
### 长宽表转换
#### 1.长表转宽表

```{r}
sys_time_print({
  mean1 <- function(x) mean(x, na.rm = TRUE)
  max1 <- function(x) max(x, na.rm = TRUE)
  r3_1 <-dat1 %>% 
    wider_dt(cut,
             value = c("depth", "price"),
             name = "color",
             fun = list(mean1,max1))
})
r3_1
```

#### 2.宽表转长表

```{r}
sys_time_print({
  r3_2 <-dat1 %>% 
    select_dt(cut,color,x,y,z) %>% 
    longer_dt(cut,color,
              name = "xyz",
              value = "xyzvalue")
})

r3_2 
```

## 高阶
### 向上/下填充空值
对于填充空值来说，可以这样操作：
```{r}
sys_time_print({
  dat1 %>% fill_na_dt(price) -> dat1
})
dat1

```


### 添加子维度聚合结果为新列

#### 1.以dat1为例，添加两列，一列为以cut、color聚合求price的均值，另一列是求标准差

```{r}

sys_time_print({
  mutate_dt(dat1,
           mean_price = mean(price, na.rm = TRUE),
           sd_price = sd(price, na.rm = TRUE),
           by = .(cut, color))
})

dat1
```

#### 2.以dat1为例，以dt分组添加一列序号id
```{r}

sys_time_print({
  dat1 %>% 
  group_dt(
    by = dt,
    mutate_dt(id = seq(.N))
  ) -> dat1
})
dat1
```

### 移动函数

```{r}

sys_time_print({
  dat1 %>% 
    group_dt(
      by = color,
      mutate_dt(
        MA10_price = frollmean(price, 10),
        MSD10_price = frollapply(price, 10, FUN = sd)
      )
    ) -> dat1
})

dat1
```


## 系统参数
```{r}
sessionInfo()
```

---
title: "Use data.table the tidy way: An ultimate tutorial of tidyfst"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{english_tutorial}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  #eval = FALSE,
  comment = "#>"
)
```

  I love the tidy syntax of dplyr and the ultimate speed of data.table. Why not take the both? That is why I have started the work of tidyfst, bridging the tidy syntax and computation performance via translation. This tool is especially friendly for dplyr users who want to learn some data.table, but data.table could also benefit from it (more or less).<br><br>
  A great comparison of data.table and dplyr was displayed at <https://atrebas.github.io/post/2019-03-03-datatable-dplyr/> (thanks to Atrebas). I love this tutorial very much because it dig rather deep into many features from both packages. Here I'll try to implement all operations from that tutorial, and the potential users could find why they would prefer tidyfst for some (if not most) tasks.<br><br>
  The below examples have all been checked with tidyfst. Now let's begin.

## Create example data
```{r setup}
library(tidyfst)

set.seed(1L)

## Create a data table
DF <- data.table(V1 = rep(c(1L, 2L), 5)[-10],
                V2 = 1:9,
                V3 = c(0.5, 1.0, 1.5),
                V4 = rep(LETTERS[1:3], 3))
copy(DF) -> DT

class(DF)
DF
```

# Basic operations
## Filter rows
```{r}
### Filter rows using indices
slice_dt(DF, 3:4)

### Discard rows using negative indices
slice_dt(DF, -(3:7))

### Filter rows using a logical expression
filter_dt(DF, V2 > 5)
filter_dt(DF, V4 %in% c("A", "C"))
filter_dt(DF, V4 %chin% c("A", "C")) # fast %in% for character

### Filter rows using multiple conditions
filter_dt(DF, V1 == 1 & V4 == "A")

### Filter unique rows
distinct_dt(DF) # unique(DF)
distinct_dt(DF, V1,V4)

### Discard rows with missing values
drop_na_dt(DF) # na.omit(DF)

### Other filters
sample_n_dt(DF, 3)      # n random rows
sample_frac_dt(DF, 0.5) # fraction of random rows
slice_max_dt(DF, V1,1)     # top n entries (includes equals)

filter_dt(DT,V4 %like% "^B")
filter_dt(DT,V2 %between% c(3, 5))
filter_dt(DT,between(V2, 3, 5, incbounds = FALSE))
filter_dt(DT,V2 %inrange% list(-1:1, 1:3))  # see also ?inrange
```

## Sort rows
```{r}
### Sort rows by column
arrange_dt(DF, V3)

### Sort rows in decreasing order
arrange_dt(DF, -V3)

### Sort rows based on several columns
arrange_dt(DF, V1, -V2)
```

## Select columns
```{r}
### Select one column using an index (not recommended)
pull_dt(DT,3) # returns a vector
select_dt(DT,3) # returns a data.table

### Select one column using column name
select_dt(DF, V2) # returns a data.table
pull_dt(DF, V2)   # returns a vector

### Select several columns
select_dt(DF, V2, V3, V4)
select_dt(DF, V2:V4) # select columns between V2 and V4

### Exclude columns
select_dt(DF, -V2, -V3)

### Select/Exclude columns using a character vector
cols <- c("V2", "V3")
select_dt(DF,cols = cols)
select_dt(DF,cols = cols,negate = TRUE)

### Other selections
select_dt(DF, cols = paste0("V", 1:2))
relocate_dt(DF, V4) # reorder columns
select_dt(DF, "V")
select_dt(DF, "3$")
select_dt(DF, ".2")
select_dt(DF, "V1")
select_dt(DF, -"^V2")
# remove variables using "-" prior to function
```

## Summarise data
```{r}
### Summarise one column
summarise_dt(DF, sum(V1)) # returns a data.table
summarise_dt(DF, sumV1 = sum(V1)) # returns a data.table

### Summarise several columns
summarise_dt(DF, sum(V1), sd(V3))

### Summarise several columns and assign column names
DF %>%
  summarise_dt(sumv1 = sum(V1),
              sdv3  = sd(V3))

### Summarise a subset of rows
DT[1:4, sum(V1)]
DF %>%
  slice_dt(1:4) %>%
  summarise_dt(sum(V1))

### Misc
summarise_dt(DF, nth(V3,1))
summarise_dt(DF, nth(V3,-1))
summarise_dt(DF, nth(V3, 5))
summarise_dt(DF, uniqueN(V4))
uniqueN(DF)
```


## group computation (by)
```{r}
### By group
# not recommended
DF %>%
  group_dt(
    by = V4,
    summarise_dt(sumV2 = sum(V2))
  )
# recommended
DF %>% 
  summarise_dt(sumV2 = sum(V2),by = V4)


### By several groups
DF %>% 
  summarise_dt(sumV2 = sum(V2),by = .(V1,V4))

### Calling function in by
DF %>% 
  summarise_dt(sumV2 = sum(V2),by = tolower(V4))


### Assigning column name in by
DF %>% 
  summarise_dt(sumV2 = sum(V2),by = .(abc = tolower(V4)))


### Using a condition in by
DF %>% 
  summarise_dt(sumV2 = sum(V2),by = V4 == "A")



### By on a subset of rows
DF %>% 
  slice_dt(1:5) %>% 
  summarise_dt(sumV1 = sum(V1),by = V4)

### Count number of observations for each group
count_dt(DF, V4)

### Add a column with number of observations for each group
add_count_dt(DF, V1)

### Retrieve the first/last/nth observation for each group
DF %>% summarise_dt(by = V4,nth(V2,1))
DF %>% summarise_dt(by = V4,nth(V2,-1))
DF %>% summarise_dt(by = V4,nth(V2,2))

```

# Going further

## Advanced columns manipulation

```{r}
### Summarise all the columns
summarise_vars(DT,.func = max)

### Summarise several columns
summarise_vars(DT,c("V1", "V2"),mean)

### Summarise several columns by group
DT %>% 
    summarise_vars(c("V1", "V2"),mean,by = V4)

## using patterns (regex)
DT %>% 
    summarise_vars("V1|V2",mean,by = V4)

## Summarise with more than one function by group
# when you can't find a way, you can always use `in_dt` to use data.table
DT %>% 
  in_dt(, by = V4, 
     c(lapply(.SD, sum),
       lapply(.SD, mean)))

### Summarise using a condition
summarise_vars(DF, is.numeric, mean)

### Modify all the columns
mutate_vars(DF, .func = rev)

### Modify several columns (dropping the others)
DF %>% 
  select_dt(cols = c("V1", "V2")) %>% 
  mutate_vars(.func = sqrt)
DF %>% 
  select_dt(-V4) %>% 
  mutate_vars(.func = exp)

### Modify several columns (keeping the others)
DF %>% 
  mutate_vars(c("V1", "V2"), sqrt)
DF %>% 
  mutate_vars(-"V4",  exp)

### Modify columns using a condition (dropping the others)
select_dt(DT,is.numeric)

### Modify columns using a condition (keeping the others)
mutate_vars(DT,is.numeric,as.integer)

### Use a complex expression
DF %>% 
  group_dt(
    by = V4,
    slice_dt(1:2) %>% 
      transmute_dt(V1 = V1,
            V2 = "X")
  )

### Use multiple expressions (with DT[,{j}])
DT %>% 
  in_dt(,{
      print(V1) #  comments here!
      print(summary(V1))
      x <- V1 + sum(V2)
     .(A = 1:.N, B = x) # last list returned as a data.table
     }
  )

```

## Advanced use of by

```{r}
### Select first/last/… row by group
DT %>% 
  group_dt(
    by = V4,
    head(1)
  )

DT %>% 
  group_dt(
    by = V4,
    tail(2)
  )

DT %>% 
  group_dt(
    by = V4,
    slice_dt(1,.N)
  )

### Select rows using a nested query
DF %>% 
  group_dt(
    by = V4,
    arrange_dt(V2) %>% 
      slice_dt(1)
  )

### Add a group counter column
DT %>% 
  mutate_dt(Grp = .GRP,by = .(V4, V1))

### Get row number of first (and last) observation by group
DT %>% summarise_dt(I = .I,by = V4)
DT %>% summarise_dt(I = .I[1],by = V4)
DT %>% summarise_dt(I = .I[c(1,.N)],by = V4)

### Handle list-columns by group

DT %>% 
  select_dt(V1,V4) %>% 
  chop_dt(V1) # return V1 as a list
DT %>% nest_dt(V4) # subsets of the data

### Grouping sets (multiple by at once)
# use data.table directly, tidyfst does not provide new methods for it yet
data.table::rollup(DT,
       .(SumV2 = sum(V2)),
       by = c("V1", "V4"))

data.table::rollup(DT,
       .(SumV2 = sum(V2), .N),
       by = c("V1", "V4"),
       id = TRUE)

data.table::cube(DT,
     .(SumV2 = sum(V2), .N),
     by = c("V1", "V4"),
     id = TRUE)

data.table::groupingsets(DT,
             .(SumV2 = sum(V2), .N),
             by   = c("V1", "V4"),
             sets = list("V1", c("V1", "V4")),
             id   = TRUE)

```

# Miscellaneous

## Read / Write data
tidyfst exports `data.table::fread` and `data.table::fwrite` directly.

```{r,eval=FALSE}
### Write data to a csv file
fwrite(DT, "DT.csv")

### Write data to a tab-delimited file
fwrite(DT, "DT.txt", sep = "\t")

### Write list-column data to a csv file
fwrite(setDT(list(0, list(1:5))), "DT2.csv")
#
### Read a csv / tab-delimited file
fread("DT.csv")
# fread("DT.csv", verbose = TRUE) # full details
fread("DT.txt", sep = "\t")

### Read a csv file selecting / droping columns
fread("DT.csv", select = c("V1", "V4"))
fread("DT.csv", drop = "V4")
# NA
### Read and rbind several files
rbindlist(lapply(c("DT.csv", "DT.csv"), fread))
# c("DT.csv", "DT.csv") %>% lapply(fread) %>% rbindlist

```

## Reshape data

```{r}
### Melt data (from wide to long)
fsetequal(DT,DF)

mDT = DT %>% longer_dt(V3,V4)
mDF = DF %>% longer_dt(-"V1|V2")

fsetequal(mDT,mDF)
mDT

### Cast data (from long to wide)
mDT %>% 
  wider_dt(V4,name = "name",value = "value")
# below is a special case and could only be done in tidyfst
mDT %>% 
  wider_dt(V4,name = "name",value = "value",fun = list)

mDT %>% 
  wider_dt(V4,name = "name",value = "value",fun = sum)

### Split
split(DT, by = "V4")

```

## Other
```{r}
### Lead/Lag
lag_dt(1:10,n = 1)
lag_dt(1:10,n = 1:2)
lead_dt(1:10,n = 1)

```

# Join/Bind data sets

## Join
```{r}
x <- data.table(Id  = c("A", "B", "C", "C"),
                X1  = c(1L, 3L, 5L, 7L),
                XY  = c("x2", "x4", "x6", "x8"),
                key = "Id")

y <- data.table(Id  = c("A", "B", "B", "D"),
                Y1  = c(1L, 3L, 5L, 7L),
                XY  = c("y1", "y3", "y5", "y7"),
                key = "Id")

### left join
left_join_dt(x, y, by = "Id")

### right join
right_join_dt(x, y, by = "Id")

### inner join
inner_join_dt(x, y, by = "Id")

### full join
full_join_dt(x, y, by = "Id")

### semi join
semi_join_dt(x, y, by = "Id")

### anti join
anti_join_dt(x, y, by = "Id")

```

## Bind
```{r}
x <- data.table(1:3)
y <- data.table(4:6)
z <- data.table(7:9, 0L)
### Bind rows
rbind(x, y)
rbind(x, z, fill = TRUE)

### Bind rows using a list
rbindlist(list(x, y), idcol = TRUE)

### Bind columns
cbind(x, y)

```

## Set operations
```{r}
x <- data.table(c(1, 2, 2, 3, 3))
y <- data.table(c(2, 2, 3, 4, 4))
### Intersection
fintersect(x, y)
fintersect(x, y, all = TRUE)

### Difference
fsetdiff(x, y)
fsetdiff(x, y, all = TRUE)

### Union
funion(x, y)
funion(x, y, all = TRUE)

### Equality
fsetequal(x, x[order(-V1),])
all.equal(x, x) # S3 method
setequal(x, x[order(-V1),])

```

# Summary
To break all these codes through, tidyfst has improved bit by bit. If you are using tidyfst frequently, you'll find that while it enjoys a tidy syntax, it is more like you are using data.table in another style. Compared with many other packages with similar goals, tidyfst sticks to many principles of data.table (and is more like data.table in many ways). However, the ultimate goal is still clear: providing users with state-of-the-art data manipulation tools with least pain. Therefore, keep it simple and make it fast. Enjoy tidyfst~

---
title: "Example 2: Join tables"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example 2: Join tables}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```
  This post has referred to a vignette from `dplyr`, you can find it in <https://dplyr.tidyverse.org/articles/two-table.html>. We'll try to display how to join data tables in this vignette. First, load the packages we need and get some data.
```{r setup}
library(tidyfst)
library(nycflights13)

flights2 <- flights %>% 
  select_dt(year,month,day, hour, origin, dest, tailnum, carrier)

```
  Do a left join with a simple:
```{r}
flights2 %>% 
  left_join_dt(airlines)
```
  
## Controlling how the tables are matched
  Join works the same as `dplyr`:
```{r}
flights2 %>% left_join_dt(weather)
flights2 %>% left_join_dt(planes, by = "tailnum")
flights2 %>% left_join_dt(airports, c("dest" = "faa"))
flights2 %>% left_join_dt(airports, c("origin" = "faa"))
```

## Types of join

```{r}
df1 <- data.table(x = c(1, 2), y = 2:1)
df2 <- data.table(x = c(1, 3), a = 10, b = "a")

df1 %>% inner_join_dt(df2) 
df1 %>% left_join_dt(df2)
df1 %>% right_join_dt(df2)
df1 %>% full_join_dt(df2)

```
  If all you have is a data.frame or tibble, you have no need to change the format. Feed the data directly:
```{r}
df1 <- data.frame(x = c(1, 1, 2), y = 1:3)
df2 <- data.frame(x = c(1, 1, 2), z = c("a", "b", "a"))

df1 %>% left_join_dt(df2)
```
  The "_dt" suffix should remind you that this is backed up by `data.table` and will always return a data.table in the end.

## Filtering joins
  Filtering joins have also been supported in `tidyfst`.
```{r}
flights %>% 
  anti_join_dt(planes, by = "tailnum") %>% 
  count_dt(tailnum, sort = TRUE)
```
  Other examples (`semi_join_dt()` and `anti_join_dt()` never duplicate; they only ever remove observations.):
```{r}
df1 <- data.frame(x = c(1, 1, 3, 4), y = 1:4)
df2 <- data.frame(x = c(1, 1, 2), z = c("a", "b", "a"))

# Four rows to start with:
df1 %>% nrow()

# And we get four rows after the join
df1 %>% inner_join_dt(df2, by = "x") %>% nrow()

# But only two rows actually match
df1 %>% semi_join_dt(df2, by = "x") %>% nrow()
```
  
## Set operations
  For set operations, wrap `data.table`'s function directly, but the functions will automatically turn any data.frame into data.table. Examples are listed as below:
  
```{r}
x = iris[c(2,3,3,4),]
x2 = iris[2:4,]
y = iris[c(3:5),]

intersect_dt(x, y)            # intersect
intersect_dt(x, y, all=TRUE)  # intersect all
setdiff_dt(x, y)              # except
setdiff_dt(x, y, all=TRUE)    # except all
union_dt(x, y)                # union
union_dt(x, y, all=TRUE)      # union all
setequal_dt(x, x2, all=FALSE) # setequal
setequal_dt(x, x2)     
```
  
For more details, just find the help from `data.table` using `?setops`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dt.R
\name{in_dt}
\alias{in_dt}
\alias{as_dt}
\title{Short cut to data.table}
\usage{
in_dt(.data, ...)

as_dt(.data)
}
\arguments{
\item{.data}{A data.frame}

\item{...}{Recieve \code{B} in data.table's \code{A[B]} syntax.}
}
\description{
To use facilities provided by \pkg{data.table}, but do not have to
load \pkg{data.table} package.
}
\details{
The \code{as_dt} could turn any data frame to data.table class. If the data is
not a data frame, return error.

The \code{in_dt} function creates a virtual environment in data.table, it could be
piped well because it still follows the principals of \pkg{tidyfst}, which are: (1) Never
use in place replacement and (2) Always recieves a data frame (data.frame/tibble/data.table)
and returns a data.table. Therefore, the in place functions like \code{:=} will still
return the results.
}
\examples{
iris \%>\% as_dt()
iris \%>\% in_dt(order(-Sepal.Length),.SD[.N],by=Species)
}
\seealso{
\code{\link[data.table]{data.table}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_by.R
\name{group_by_dt}
\alias{group_by_dt}
\alias{group_exe_dt}
\title{Group by variable(s) and implement operations}
\usage{
group_by_dt(.data, ..., cols = NULL)

group_exe_dt(.data, ...)
}
\arguments{
\item{.data}{A data frame}

\item{...}{Variables to group by for \code{group_by_dt},
namely the columns to sort by. Do not quote the column names.
Any data manipulation arguments that could be
implemented on a data.frame for \code{group_exe_dt}.
It can receive what \code{select_dt} receives.}

\item{cols}{A character vector of column names to group by.}
}
\value{
A data.table with keys
}
\description{
Carry out data manipulation within specified groups. Different from \code{group_dt},
the implementation is split into two operations, namely grouping and implementation.

Using \code{setkey} and \code{setkeyv} in \pkg{data.table}
to carry out \code{group_by}-like functionalities in \pkg{dplyr}. This is
not only convenient but also efficient in computation.
}
\details{
\code{group_by_dt} and \code{group_exe_dt} are a pair of functions
to be used in combination. It utilizes the feature of key setting in data.table,
which provides high performance for group operations, especially when you have
to operate by specific groups frequently.
}
\examples{

# aggregation after grouping using group_exe_dt
as.data.table(iris) -> a
a \%>\%
  group_by_dt(Species) \%>\%
  group_exe_dt(head(1))

a \%>\%
  group_by_dt(Species) \%>\%
  group_exe_dt(
    head(3) \%>\%
      summarise_dt(sum = sum(Sepal.Length))
  )

mtcars \%>\%
  group_by_dt("cyl|am") \%>\%
  group_exe_dt(
    summarise_dt(mpg_sum = sum(mpg))
  )
# equals to
mtcars \%>\%
  group_by_dt(cols = c("cyl","am")) \%>\%
  group_exe_dt(
    summarise_dt(mpg_sum = sum(mpg))
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg_load.R
\name{pkg_load}
\alias{pkg_load}
\alias{pkg_unload}
\title{Load or unload R package(s)}
\usage{
pkg_load(..., pkg_names = NULL)

pkg_unload(..., pkg_names = NULL)
}
\arguments{
\item{...}{Name(s) of package(s).}

\item{pkg_names}{(Optional)Character vector containing packages to load
or unload. Default uses \code{NULL}.}
}
\description{
This function is a wrapper for \code{\link[base]{require}} and
\code{\link[base]{detach}}.  \code{pkg_load} checks to see if a
package is installed, if not it attempts to install the package
from CRAN. \code{pkg_unload} can detach one or more loaded packages.
}
\examples{
\dontrun{
pkg_load(data.table)
pkg_unload(data.table)

pkg_load(stringr,fst)
pkg_unload(stringr,fst)

pkg_load(pkg_names = c("data.table","fst"))
p_unload(pkg_names = c("data.table","fst"))

pkg_load(data.table,stringr,fst)
pkg_unload("all") # shortcut to unload all loaded packages
}
}
\seealso{
\code{\link[base]{require}},
\code{\link[base]{detach}},
\code{\link[pacman]{p_load}},
\code{\link[pacman]{p_unload}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind_tf_idf_dt.R
\name{bind_tf_idf_dt}
\alias{bind_tf_idf_dt}
\title{Compute TF–IDF Using data.table with Optional Counting and Grouping}
\usage{
bind_tf_idf_dt(.data, group_col = NULL, doc_col, term_col, n_col = NULL)
}
\arguments{
\item{.data}{A data.frame or data.table of text data.}

\item{group_col}{Character name of grouping column, or `NULL` for no grouping.}

\item{doc_col}{Character name of document identifier column.}

\item{term_col}{Character name of term/word column.}

\item{n_col}{(Optional) Character name of pre-counted term-frequency column.
If `NULL` (default), counts are computed via `.N`.}
}
\value{
A data.table containing:
  - Original grouping, document, and term columns
  - `n`, `tf`, `idf`, and `tf_idf`
}
\description{
This function computes term frequency–inverse document frequency (tf–idf)
on a dataset with one row per term occurrence (or pre-counted).
It preserves original column names and returns new columns:
  - `n`: raw count (computed or user-supplied)
  - `tf`: term frequency per document
  - `idf`: inverse document frequency per group (or corpus)
  - `tf_idf`: tf × idf
If `group_col` is `NULL`, all documents are treated as a single group.
}
\examples{

# With groups
df <- data.frame(
  category = rep(c("A","B"), each = 6),
  doc_id   = rep(c("d1","d2","d3"), times = 4),
  word     = c("apple","banana","apple","banana","cherry","apple",
               "dog","cat","dog","mouse","cat","dog"),
  stringsAsFactors = FALSE
)
result <- bind_tf_idf_dt(df, "category", "doc_id", "word")
result

# Without groups
df \%>\%
  filter_dt(category == "A") \%>\%
  bind_tf_idf_dt(doc_col = "doc_id",term_col = "word")

# With counts provided
df \%>\%
  filter_dt(category == "A") \%>\%
  count_dt() \%>\%
  bind_tf_idf_dt(doc_col = "doc_id",term_col = "word",n_col = "n")
df \%>\%
  count_dt() \%>\%
  bind_tf_idf_dt(group_col = "category",
                 doc_col = "doc_id",
                 term_col = "word",n_col = "n")

}
\seealso{
\code{\link[tidytext]{bind_tf_idf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distinct_dt.R
\name{distinct_dt}
\alias{distinct_dt}
\title{Select distinct/unique rows in data.frame}
\usage{
distinct_dt(.data, ..., .keep_all = FALSE, fromLast = FALSE)
}
\arguments{
\item{.data}{data.frame}

\item{...}{Optional variables to use when determining uniqueness.
If there are multiple rows for a given combination of inputs,
only the first row will be preserved.
If omitted, will use all variables.}

\item{.keep_all}{If \code{TRUE}, keep all variables in data.frame. If a combination of ... is not distinct,
this keeps the first row of values.}

\item{fromLast}{Logical indicating if duplication should be
considered from the reverse side. Defaults to \code{FALSE}.}
}
\value{
data.table
}
\description{
Select only unique/distinct rows from a data frame.
}
\examples{
iris \%>\% distinct_dt()
iris \%>\% distinct_dt(Species)
iris \%>\% distinct_dt(Species,.keep_all = TRUE)
mtcars \%>\% distinct_dt(cyl,vs)
mtcars \%>\% distinct_dt(cyl,vs,.keep_all = TRUE)
mtcars \%>\% distinct_dt(cyl,vs,.keep_all = TRUE,fromLast = TRUE)


}
\seealso{
\code{\link[dplyr]{distinct}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impute_dt.R
\name{impute_dt}
\alias{impute_dt}
\title{Impute missing values with mean, median or mode}
\usage{
impute_dt(.data, ..., .func = "mode")
}
\arguments{
\item{.data}{A data.frame}

\item{...}{Columns to select}

\item{.func}{Character, "mode" (default), "mean" or "median".
Could also define it by oneself.}
}
\value{
A data.table
}
\description{
Impute the columns of data.frame with its mean, median or mode.
}
\examples{

Pclass <- c(3, 1, 3, 1, 3, 2, 2, 3, NA, NA)
Sex <- c('male', 'male', 'female', 'female', 'female',
         'female', NA, 'male', 'female', NA)
Age <- c(22, 38, 26, 35, NA,
         45, 25, 39, 28, 40)
SibSp <- c(0, 1, 3, 1, 2, 3, 2, 2, NA, 0)
Fare <- c(7.25, 71.3, 7.92, NA, 8.05, 8.46, 51.9, 60, 32, 15)
Embarked <- c('S', NA, 'S', 'Q', 'Q', 'S', 'C', 'S', 'C', 'S')
data <- data.frame('Pclass' = Pclass,
 'Sex' = Sex, 'Age' = Age, 'SibSp' = SibSp,
 'Fare' = Fare, 'Embarked' = Embarked)

data
data \%>\% impute_dt() # defalut uses "mode" as `.func`
data \%>\% impute_dt(is.numeric,.func = "mean")
data \%>\% impute_dt(is.numeric,.func = "median")

my_fun = function(x){
  x[is.na(x)] = (max(x,na.rm = TRUE) - min(x,na.rm = TRUE))/2
  x
}
data \%>\% impute_dt(is.numeric,.func = my_fun)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lag_lead.R
\name{lead_dt}
\alias{lead_dt}
\alias{lag_dt}
\title{Fast lead/lag for vectors}
\usage{
lead_dt(x, n = 1L, fill = NA)

lag_dt(x, n = 1L, fill = NA)
}
\arguments{
\item{x}{A vector}

\item{n}{a positive integer of length 1,
giving the number of positions to lead or lag by. Default uses 1}

\item{fill}{Value to use for padding when the window goes beyond the input length.
Default uses \code{NA}}
}
\value{
A vector
}
\description{
Find the "next" or "previous" values in a vector.
It has wrapped \pkg{data.table}'s \code{shift} function.
}
\examples{
lead_dt(1:5)
lag_dt(1:5)
lead_dt(1:5,2)
lead_dt(1:5,n = 2,fill = 0)
}
\seealso{
\code{\link[dplyr]{lead}},\code{\link[data.table]{shift}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutate_dt.R
\name{mutate_dt}
\alias{mutate_dt}
\alias{transmute_dt}
\title{Mutate columns in data.frame}
\usage{
mutate_dt(.data, ..., by)

transmute_dt(.data, ..., by)
}
\arguments{
\item{.data}{data.frame}

\item{...}{List of variables or name-value pairs of summary/modifications
functions.}

\item{by}{(Optional) Mutate by what group?}
}
\value{
data.table
}
\description{
Adds or updates columns in data.frame.
}
\examples{

iris \%>\% mutate_dt(one = 1,Sepal.Length = Sepal.Length + 1)
iris \%>\% transmute_dt(one = 1,Sepal.Length = Sepal.Length + 1)
# add group number with symbol `.GRP`
iris \%>\% mutate_dt(id = 1:.N,grp = .GRP,by = Species)

}
\seealso{
\code{\link[dplyr]{mutate}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidymat.R
\name{mat_df}
\alias{mat_df}
\alias{df_mat}
\title{Conversion between tidy table and named matrix}
\usage{
mat_df(m)

df_mat(df, row, col, value)
}
\arguments{
\item{m}{A matrix}

\item{df}{A data.frame with at least 3 columns, one for row name,
one for column name, and one for values. The names for column and
row should be unique.}

\item{row}{Unquoted expression of column name for row}

\item{col}{Unquoted expression of column name for column}

\item{value}{Unquoted expression of column name for values}
}
\value{
For \code{mat_df}, a data.frame.
For \code{df_mat}, a named matrix.
}
\description{
Convenient fucntions to implement conversion between
 tidy table and named matrix.
}
\examples{

mm = matrix(c(1:8,NA),ncol = 3,dimnames = list(letters[1:3],LETTERS[1:3]))
mm
tdf = mat_df(mm)
tdf
mat = df_mat(tdf,row,col,value)
setequal(mm,mat)

tdf \%>\%
  setNames(c("A","B","C")) \%>\%
  df_mat(A,B,C)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_prop.R
\name{percent}
\alias{percent}
\alias{add_prop}
\title{Add percentage to counts in data.frame}
\usage{
percent(x, digits = 1)

add_prop(.data, count_name = last(names(.data)), digits = 1)
}
\arguments{
\item{x}{A number (numeric).}

\item{digits}{How many digits to keep in the percentage. Default uses 1.}

\item{.data}{A data frame.}

\item{count_name}{Column name of counts (Character).
Default uses the last column of data.frame.}
}
\description{
Add percentage for counts in the data.frame, both numeric and
 character with `%` would be provided.
}
\examples{

 percent(0.9057)
 percent(0.9057,3)

 iris \%>\%
   count_dt(Species) \%>\%
   add_prop()

 iris \%>\%
   count_dt(Species) \%>\%
   add_prop(count_name = "n",digits = 2)

}
\references{
https://stackoverflow.com/questions/7145826/how-to-format-a-number-as-percentage-in-r
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complete.R
\name{complete_dt}
\alias{complete_dt}
\title{Complete a data frame with missing combinations of data}
\usage{
complete_dt(.data, ..., fill = NA)
}
\arguments{
\item{.data}{data.frame}

\item{...}{Specification of columns to expand.The selection of columns is
supported by the flexible \code{\link[tidyfst]{select_dt}}.
To find all unique combinations of provided columns, including those not found in the data,
supply each variable as a separate argument. But the two modes (select the
needed columns and fill outside values) could not be mixed,
find more details in examples.}

\item{fill}{Atomic value to fill into the missing cell, default uses \code{NA}.}
}
\value{
data.table
}
\description{
Turns implicit missing values into explicit missing values.
 All the combinations of column values (should be unique) will be constructed.
 Other columns will be filled with NAs or constant value.
}
\details{
When the provided columns with addtion data are of different length,
all the unique combinations would be returned. This operation should be used
only on unique entries, and it will always returned the unique entries.

If you supply fill parameter, these values will also replace existing explicit missing values in the data set.
}
\examples{
df <- data.table(
  group = c(1:2, 1),
  item_id = c(1:2, 2),
  item_name = c("a", "b", "b"),
  value1 = 1:3,
  value2 = 4:6
)

df \%>\% complete_dt(item_id,item_name)
df \%>\% complete_dt(item_id,item_name,fill = 0)
df \%>\% complete_dt("item")
df \%>\% complete_dt(item_id=1:3)
df \%>\% complete_dt(item_id=1:3,group=1:2)
df \%>\% complete_dt(item_id=1:3,group=1:3,item_name=c("a","b","c"))

}
\seealso{
\code{\link[tidyr]{complete}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slice.R
\name{slice_dt}
\alias{slice_dt}
\alias{slice_head_dt}
\alias{slice_tail_dt}
\alias{slice_max_dt}
\alias{slice_min_dt}
\alias{slice_sample_dt}
\title{Subset rows using their positions}
\usage{
slice_dt(.data, ..., by = NULL)

slice_head_dt(.data, n, by = NULL)

slice_tail_dt(.data, n, by = NULL)

slice_max_dt(.data, order_by, n, by = NULL, with_ties = TRUE)

slice_min_dt(.data, order_by, n, by = NULL, with_ties = TRUE)

slice_sample_dt(.data, n, replace = FALSE, by = NULL)
}
\arguments{
\item{.data}{A data.table}

\item{...}{Provide either positive values to keep, or negative values to drop.
The values provided must be either all positive or all negative.}

\item{by}{Slice by which group(s)?}

\item{n}{When larger than or equal to 1, the number of rows.
When between 0 and 1, the proportion of rows to select.}

\item{order_by}{Variable or function of variables to order by.}

\item{with_ties}{Should ties be kept together? The default, `TRUE`,
may return more rows than you request. Use `FALSE` to ignore ties,
and return the first `n` rows.}

\item{replace}{Should sampling be performed with (`TRUE`) or without
(`FALSE`, the default) replacement.}
}
\value{
A data.table
}
\description{
`slice_dt()` lets you index rows by their (integer) locations. It allows you
to select, remove, and duplicate rows. It is accompanied by a number of
helpers for common use cases:

* `slice_head_dt()` and `slice_tail_dt()` select the first or last rows.
* `slice_sample_dt()` randomly selects rows.
* `slice_min_dt()` and `slice_max_dt()` select rows with highest or lowest values
  of a variable.
}
\examples{

a = iris
slice_dt(a,1,2)
slice_dt(a,2:3)
slice_dt(a,141:.N)
slice_dt(a,1,.N)
slice_head_dt(a,5)
slice_head_dt(a,0.1)
slice_tail_dt(a,5)
slice_tail_dt(a,0.1)
slice_max_dt(a,Sepal.Length,10)
slice_max_dt(a,Sepal.Length,10,with_ties = FALSE)
slice_min_dt(a,Sepal.Length,10)
slice_min_dt(a,Sepal.Length,10,with_ties = FALSE)
slice_sample_dt(a,10)
slice_sample_dt(a,0.1)


# use by to slice by group

## following codes get the same results
slice_dt(a,1:3,by = "Species")
slice_dt(a,1:3,by = Species)
slice_dt(a,1:3,by = .(Species))

slice_head_dt(a,2,by = Species)
slice_tail_dt(a,2,by = Species)

slice_max_dt(a,Sepal.Length,3,by = Species)
slice_max_dt(a,Sepal.Length,3,by = Species,with_ties = FALSE)
slice_min_dt(a,Sepal.Length,3,by = Species)
slice_min_dt(a,Sepal.Length,3,by = Species,with_ties = FALSE)

# in `slice_sample_dt`, "by" could only take character class
slice_sample_dt(a,.1,by = "Species")
slice_sample_dt(a,3,by = "Species")
slice_sample_dt(a,51,replace = TRUE,by = "Species")

}
\seealso{
\code{\link[dplyr]{slice}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sql_join.R
\name{sql_join}
\alias{sql_join}
\alias{sql_join_dt}
\title{Case insensitive table joining like SQL}
\usage{
sql_join_dt(x, y, by = NULL, type = "inner", suffix = c(".x", ".y"))
}
\arguments{
\item{x}{A data.table}

\item{y}{A data.table}

\item{by}{(Optional) A character vector of variables to join by.

  If `NULL`, the default, `*_join_dt()` will perform a natural join, using all
  variables in common across `x` and `y`. A message lists the variables so that you
  can check they're correct; suppress the message by supplying `by` explicitly.

  To join by different variables on `x` and `y`, use a named vector.
  For example, `by = c("a" = "b")` will match `x$a` to `y$b`.

  To join by multiple variables, use a vector with length > 1.
  For example, `by = c("a", "b")` will match `x$a` to `y$a` and `x$b` to
  `y$b`. Use a named vector to match different variables in `x` and `y`.
  For example, `by = c("a" = "b", "c" = "d")` will match `x$a` to `y$b` and
  `x$c` to `y$d`.

  Notice that in `sql_join`, the joining variables would turn to upper case
  in the output table.}

\item{type}{Which type of join would you like to use?
Default uses "inner", other options include
"left", "right", "full", "anti", "semi".}

\item{suffix}{If there are non-joined duplicate variables in x and y, these
suffixes will be added to the output to disambiguate them. Should be a
character vector of length 2.}
}
\value{
A data.table
}
\description{
Work like the `*_join_dt` series functions, joining
tables with common or customized keys  in various ways. The only
difference is the joining is case insensitive like SQL.
}
\examples{
dt1 = data.table(x = c("A","b"),y = 1:2)
dt2 = data.table(x = c("a","B"),z = 4:5)
sql_join_dt(dt1,dt2)
}
\seealso{
\code{\link[tidyfst]{join}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_fst_chunked.R
\name{import_fst_chunked}
\alias{import_fst_chunked}
\alias{get_fst_chunk_size}
\title{Read a fst file by chunks}
\usage{
import_fst_chunked(
  path,
  chunk_size = 10000L,
  chunk_f = identity,
  combine_f = rbindlist
)

get_fst_chunk_size(path, nrows)
}
\arguments{
\item{path}{Path to fst file}

\item{chunk_size}{Integer. The number of rows to include in each chunk}

\item{chunk_f}{A function implemented on every chunk.}

\item{combine_f}{A function to aggregate all the elements from the list of results from chunks.}

\item{nrows}{Number of rows to test.}
}
\value{
For `import_fst_chunked`, default to the whole data.frame in data.table.
 Could be adjusted to any type. For `get_fst_chunk_size`, return the file size.
}
\description{
For `import_fst_chunked`, if a large fst file which could not be imported into the memory all at once,
 this function could read the fst file by chunks and preprocessed the chunk to ensure the
 results yielded by the chunks are small enough to be summarised in the end.
 For `get_fst_chunk_size`, this function can measure the memory used by a specified row number.
}
\examples{

\dontrun{
  # Generate some random data frame with 10 million rows and various column types
  nr_of_rows <- 1e7
  df <- data.frame(
    Logical = sample(c(TRUE, FALSE, NA), prob = c(0.85, 0.1, 0.05), nr_of_rows, replace = TRUE),
    Integer = sample(1L:100L, nr_of_rows, replace = TRUE),
    Real = sample(sample(1:10000, 20) / 100, nr_of_rows, replace = TRUE),
    Factor = as.factor(sample(labels(UScitiesD), nr_of_rows, replace = TRUE))
  )

  # Write the file to disk
  fst_file <- tempfile(fileext = ".fst")
  write_fst(df, fst_file)

  # Get the size of 10000 rows
  get_fst_chunk_size(fst_file,1e4)

  # File all rows that Integer == 7 by chunks
  import_fst_chunked(fst_file,chunk_f = \(x) x[Integer==7])

}

}
\seealso{
\code{\link[readr]{read_csv_chunked}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cummean.R
\name{cummean}
\alias{cummean}
\title{Cumulative mean}
\usage{
cummean(x)
}
\arguments{
\item{x}{a numeric or complex object,
or an object that can be coerced to one of these.}
}
\description{
Returns a vector whose elements are the cumulative mean of the elements of the argument.
}
\examples{
cummean(1:10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wider_dt.R
\name{wider_dt}
\alias{wider_dt}
\title{Pivot data from long to wide}
\usage{
wider_dt(.data, ..., name, value = NULL, fun = NULL, fill = NA)
}
\arguments{
\item{.data}{A data.frame}

\item{...}{Optional. The unchanged group in the transformation.
Could use integer vector, could receive what \code{select_dt} receives.}

\item{name}{Chracter.One column name of class to spread}

\item{value}{Chracter.One column name of value to spread.
If \code{NULL}, use all other variables.}

\item{fun}{Should the data be aggregated before casting?
Defaults to \code{NULL}, which uses \code{length} for aggregation.
If a function is provided, with aggregated by this function.}

\item{fill}{Value with which to fill missing cells. Default uses \code{NA}.}
}
\value{
data.table
}
\description{
Transform a data frame from long format to wide by increasing the number of columns and decreasing the number of rows.
}
\details{
The parameter of `name` and `value` should always
be provided and should be explicit called (with the parameter names attached).
}
\examples{
 stocks = data.frame(
   time = as.Date('2009-01-01') + 0:9,
   X = rnorm(10, 0, 1),
   Y = rnorm(10, 0, 2),
   Z = rnorm(10, 0, 4)
 ) \%>\%
   longer_dt(time) -> longer_stocks

 longer_stocks

 longer_stocks \%>\%
   wider_dt("time",
            name = "name",
            value = "value")

 longer_stocks \%>\%
   mutate_dt(one = 1) \%>\%
   wider_dt("time",
            name = "name",
            value = "one")

## using "fun" parameter for aggregation
DT <- data.table(v1 = rep(1:2, each = 6),
                 v2 = rep(rep(1:3, 2), each = 2),
                 v3 = rep(1:2, 6),
                 v4 = rnorm(6))
## for each combination of (v1, v2), add up all values of v4
DT \%>\%
  wider_dt(v1,v2,
           value = "v4",
           name = ".",
           fun = sum)
}
\seealso{
\code{\link[tidyfst]{longer_dt}},
 \code{\link[data.table]{dcast}},
 \code{\link[tidyr]{pivot_wider}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replace_dt.R
\name{replace_dt}
\alias{replace_dt}
\title{Fast value replacement in data frame}
\usage{
replace_dt(.data, ..., from = is.nan, to = NA)
}
\arguments{
\item{.data}{A data.frame}

\item{...}{Colunms to be replaced. If not specified, use all columns.}

\item{from}{A value, a vector of values or a function returns a logical value.
Defaults to \code{is.nan}.}

\item{to}{A value. Defaults to \code{NA}.}
}
\value{
A data.table.
}
\description{
While \code{replace_na_dt} could replace all NAs to another
value, \code{replace_dt} could replace any value(s) to another specific
value.
}
\examples{
iris \%>\% mutate_vars(is.factor,as.character) -> new_iris

new_iris \%>\%
  replace_dt(Species, from = "setosa",to = "SS")
new_iris \%>\%
  replace_dt(Species,from = c("setosa","virginica"),to = "sv")
new_iris \%>\%
  replace_dt(Petal.Width, from = .2,to = 2)
new_iris \%>\%
  replace_dt(from = .2,to = NA)
new_iris \%>\%
  replace_dt(is.numeric, from = function(x) x > 3, to = 9999 )
}
\seealso{
\code{\link[tidyfst]{replace_na_dt}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/round0.R
\name{round0}
\alias{round0}
\title{Round a number and make it show zeros}
\usage{
round0(x, digits = 0)
}
\arguments{
\item{x}{A numeric vector.}

\item{digits}{Integer indicating the number of decimal places. Defaults to 0.}
}
\value{
A character vector.
}
\description{
Rounds values in its first argument to the specified number of decimal places,
 returning character, ensuring first decimal digits are showed even when they are zeros.
}
\examples{
a = 14.0034
round0(a,2)

b = 10
round0(b,1)

}
\references{
https://stackoverflow.com/questions/42105336/how-to-round-a-number-and-make-it-show-zeros
}
\seealso{
\code{\link[base]{round}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_dt.R
\name{count_dt}
\alias{count_dt}
\alias{add_count_dt}
\title{Count observations by group}
\usage{
count_dt(.data, ..., sort = TRUE, .name = "n")

add_count_dt(.data, ..., .name = "n")
}
\arguments{
\item{.data}{data.table/data.frame data.frame will be automatically converted
to data.table.}

\item{...}{Variables to group by, could receive what `select_dt` receives.}

\item{sort}{logical. If TRUE result will be sorted in desending order by resulting variable.}

\item{.name}{character. Name of resulting variable. Default uses "n".}
}
\value{
data.table
}
\description{
Count the unique values of one or more variables.
}
\examples{
iris \%>\% count_dt(Species)
iris \%>\% count_dt(Species,.name = "count")
iris \%>\% add_count_dt(Species)
iris \%>\% add_count_dt(Species,.name = "N")

mtcars \%>\% count_dt(cyl,vs)
mtcars \%>\% count_dt("cyl|vs")
mtcars \%>\% count_dt(cyl,vs,.name = "N",sort = FALSE)
mtcars \%>\% add_count_dt(cyl,vs)
mtcars \%>\% add_count_dt("cyl|vs")

}
\seealso{
\code{\link[dplyr]{count}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nth.R
\name{nth}
\alias{nth}
\alias{maxth}
\alias{minth}
\title{Extract the nth value from a vector}
\usage{
nth(v, n = 1)

maxth(v, n = 1)

minth(v, n = 1)
}
\arguments{
\item{v}{A vector}

\item{n}{For\code{nth}, a single integer specifying the position. Default uses \code{1}.
Negative integers index from the end
 (i.e. -1L will return the last value in the vector).
 If a double is supplied, it will be silently truncated.
 For \code{maxth} and \code{minth}, a single integer indicating the nth
 highest or lowest value.}
}
\value{
A single value.
}
\description{
\code{nth} get the value from a vector with its position,
 while \code{maxth} and \code{minth} get the nth highest or lowest value
 from the vector.
}
\examples{

x = 1:10
nth(x, 1)
nth(x, 5)
nth(x, -2)

y = c(10,3,4,5,2,1,6,9,7,8)
maxth(y,3)
minth(y,3)

}
\references{
https://stackoverflow.com/questions/2453326/fastest-way-to-find-second-third-highest-lowest-value-in-vector-or-column/66367996#66367996
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setops.R
\name{intersect_dt}
\alias{intersect_dt}
\alias{union_dt}
\alias{setdiff_dt}
\alias{setequal_dt}
\title{Set operations for data frames}
\usage{
intersect_dt(x, y, all = FALSE)

union_dt(x, y, all = FALSE)

setdiff_dt(x, y, all = FALSE)

setequal_dt(x, y, all = TRUE)
}
\arguments{
\item{x}{A data.frame}

\item{y}{A data.frame}

\item{all}{Logical. When \code{FALSE} (default),
removes duplicate rows on the result.}
}
\value{
A data.table
}
\description{
Wrappers of set operations in \pkg{data.table}.
Only difference is it could be applied to non-data.table data frames by
recognizing and coercing them to data.table automatically.
}
\examples{

x = iris[c(2,3,3,4),]
x2 = iris[2:4,]
y = iris[c(3:5),]

intersect_dt(x, y)            # intersect
intersect_dt(x, y, all=TRUE)  # intersect all
setdiff_dt(x, y)              # except
setdiff_dt(x, y, all=TRUE)    # except all
union_dt(x, y)                # union
union_dt(x, y, all=TRUE)      # union all
setequal_dt(x, x2, all=FALSE) # setequal
setequal_dt(x, x2)            # setequal all

}
\seealso{
\code{\link[data.table]{setops}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print_options}
\alias{print_options}
\title{Set global printing method for data.table}
\usage{
print_options(
  topn = 5,
  nrows = 100,
  class = TRUE,
  row.names = TRUE,
  col.names = "auto",
  print.keys = TRUE,
  trunc.cols = FALSE
)
}
\arguments{
\item{topn}{The number of rows to be printed from the beginning and
end of tables with more than \code{nrow} rows.}

\item{nrows}{The number of rows which will be printed before truncation is enforced.}

\item{class}{If \code{TRUE}, the resulting output will include above each column its storage class (or a self-evident abbreviation thereof).}

\item{row.names}{If \code{TRUE}, row indices will be printed.}

\item{col.names}{One of three flavours for controlling the display of column names in output. \code{"auto"} includes column names above the data, as well as below the table if \code{nrow(x) > 20}. \code{"top"} excludes this lower register when applicable, and \code{"none"} suppresses column names altogether (as well as column classes if \code{class = TRUE}.}

\item{print.keys}{If \code{TRUE}, any \code{\link{key}} currently assigned to \code{x} will be printed prior to the preview of the data.}

\item{trunc.cols}{If \code{TRUE}, only the columns that can be printed in the console without wrapping the columns to new lines will be printed (similar to \code{tibbles}).}
}
\value{
None. This function is used for its side effect of changing options.
}
\description{
This function allow user to define how data.table is printed.
}
\details{
Notice that \pkg{tidyfst} has a slightly different printing default for data.table,
 which is it always prints the keys and variable class (not like \pkg{data.table}).
}
\examples{

iris \%>\% as.data.table()
print_options(topn = 3,trunc.cols = TRUE)
iris \%>\% as.data.table()

# set all settings to default in tidyfst
print_options()
iris \%>\% as.data.table()

}
\seealso{
\code{\link[data.table]{print.data.table}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fst.R
\name{fst}
\alias{fst}
\alias{parse_fst}
\alias{slice_fst}
\alias{select_fst}
\alias{filter_fst}
\alias{summary_fst}
\title{Parse,inspect and extract data.table from fst file}
\usage{
parse_fst(path)

slice_fst(ft, row_no)

select_fst(ft, ...)

filter_fst(ft, ...)

summary_fst(ft)
}
\arguments{
\item{path}{path to fst file}

\item{ft}{An object of class fst_table, returned by \code{parse_fst}}

\item{row_no}{An integer vector (Positive)}

\item{...}{The filter conditions}
}
\value{
\code{parse_fst} returns a fst_table class.

\code{select_fst} and \code{filter_fst} returns a data.table.
}
\description{
A tookit of APIs for reading fst file as data.table, could select by column, row and conditional filtering.
}
\details{
\code{summary_fst} could provide some basic information about
the fst table.
}
\examples{

\dontrun{
  fst::write_fst(iris,"iris_test.fst")
  # parse the file but not reading it
  parse_fst("iris_test.fst") -> ft
  ft

  class(ft)
  lapply(ft,class)
  names(ft)
  dim(ft)
  summary_fst(ft)

  # get the data by query
  ft \%>\% slice_fst(1:3)
  ft \%>\% slice_fst(c(1,3))

  ft \%>\% select_fst(Sepal.Length)
  ft \%>\% select_fst(Sepal.Length,Sepal.Width)
  ft \%>\% select_fst("Sepal.Length")
  ft \%>\% select_fst(1:3)
  ft \%>\% select_fst(1,3)
  ft \%>\% select_fst("Se")
  ft \%>\% select_fst("nothing")
  ft \%>\% select_fst("Se|Sp")
  ft \%>\% select_fst(cols = names(iris)[2:3])

  ft \%>\% filter_fst(Sepal.Width > 3)
  ft \%>\% filter_fst(Sepal.Length > 6 , Species == "virginica")
  ft \%>\% filter_fst(Sepal.Length > 6 & Species == "virginica" & Sepal.Width < 3)

  unlink("iris_test.fst")
}
}
\seealso{
\code{\link[fst]{fst}}, \code{\link[fst]{metadata_fst}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_dt.R
\name{select_dt}
\alias{select_dt}
\alias{select_mix}
\title{Select column from data.frame}
\usage{
select_dt(.data, ..., cols = NULL, negate = FALSE)

select_mix(.data, ..., rm.dup = TRUE)
}
\arguments{
\item{.data}{data.frame}

\item{...}{List of variables or name-value pairs of summary/modifications
functions. It can also recieve conditional function to select columns.
When starts with `-`(minus symbol) or `!`, return the negative columns.}

\item{cols}{(Optional)A numeric or character vector.}

\item{negate}{Applicable when regular expression and "cols" is used.
If \code{TRUE}, return the non-matched pattern. Default uses \code{FALSE}.}

\item{rm.dup}{Should duplicated columns be removed? Defaults to \code{TRUE}.}
}
\value{
data.table
}
\description{
Select specific column(s) via various ways. One can select columns by their column names, indexes or regular expression recognizing the column name(s).
}
\examples{
iris \%>\% select_dt(Species)
iris \%>\% select_dt(Sepal.Length,Sepal.Width)
iris \%>\% select_dt(Sepal.Length:Petal.Length)
iris \%>\% select_dt(-Sepal.Length)
iris \%>\% select_dt(-Sepal.Length,-Petal.Length)
iris \%>\% select_dt(-(Sepal.Length:Petal.Length))
iris \%>\% select_dt(c("Sepal.Length","Sepal.Width"))
iris \%>\% select_dt(-c("Sepal.Length","Sepal.Width"))
iris \%>\% select_dt(1)
iris \%>\% select_dt(-1)
iris \%>\% select_dt(1:3)
iris \%>\% select_dt(-(1:3))
iris \%>\% select_dt(1,3)
iris \%>\% select_dt("Pe")
iris \%>\% select_dt(-"Se")
iris \%>\% select_dt(!"Se")
iris \%>\% select_dt("Pe",negate = TRUE)
iris \%>\% select_dt("Pe|Sp")
iris \%>\% select_dt(cols = 2:3)
iris \%>\% select_dt(cols = 2:3,negate = TRUE)
iris \%>\% select_dt(cols = c("Sepal.Length","Sepal.Width"))
iris \%>\% select_dt(cols = names(iris)[2:3])

iris \%>\% select_dt(is.factor)
iris \%>\% select_dt(-is.factor)
iris \%>\% select_dt(!is.factor)

# select_mix could provide flexible mix selection
select_mix(iris, Species,"Sepal.Length")
select_mix(iris,1:2,is.factor)

select_mix(iris,Sepal.Length,is.numeric)
# set rm.dup to FALSE could save the duplicated column names
select_mix(iris,Sepal.Length,is.numeric,rm.dup = FALSE)

}
\seealso{
\code{\link[dplyr]{select}}, \code{\link[dplyr]{select_if}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/object_size.R
\name{object_size}
\alias{object_size}
\title{Nice printing of report the Space Allocated for an Object}
\usage{
object_size(object)
}
\arguments{
\item{object}{an R object.}
}
\value{
An object of class "object_size"
}
\description{
Provides an estimate of the memory that is being used to store an R object.
A wrapper of `object.size`, but use a nicer printing unit.
}
\examples{

iris \%>\% object_size()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8_encoding.R
\name{utf8_encoding}
\alias{utf8_encoding}
\title{Use UTF-8 for character encoding in a data frame}
\usage{
utf8_encoding(.data)
}
\arguments{
\item{.data}{A data.frame.}
}
\value{
A data.table with characters in UTF-8 encoding
}
\description{
\code{fread} from \pkg{data.table} could not recognize the encoding
and return the correct form, this could be unconvenient for text mining tasks. The
\code{utf8-encoding} could use "UTF-8" as the encoding to override the current
encoding of characters in a data frame.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_dt.R
\name{group_dt}
\alias{group_dt}
\alias{rowwise_dt}
\title{Data manipulation within groups}
\usage{
group_dt(.data, by = NULL, ...)

rowwise_dt(.data, ...)
}
\arguments{
\item{.data}{A data.frame}

\item{by}{Variables to group by,unquoted name of grouping variable of list of unquoted names of grouping variables.}

\item{...}{Any data manipulation arguments that could be implemented on a data.frame.}
}
\value{
data.table
}
\description{
Carry out data manipulation within specified groups.
}
\details{
If you want to use \code{summarise_dt} and \code{mutate_dt} in
\code{group_dt}, it is better to use the "by" parameter in those functions,
that would be much faster because you don't have to use \code{.SD} (which takes
extra time to copy).
}
\examples{
iris \%>\% group_dt(by = Species,slice_dt(1:2))
iris \%>\% group_dt(Species,filter_dt(Sepal.Length == max(Sepal.Length)))
iris \%>\% group_dt(Species,summarise_dt(new = max(Sepal.Length)))

# you can pipe in the `group_dt`
iris \%>\% group_dt(Species,
                  mutate_dt(max= max(Sepal.Length)) \%>\%
                    summarise_dt(sum=sum(Sepal.Length)))

# for users familiar with data.table, you can work on .SD directly
# following codes get the first and last row from each group
iris \%>\%
  group_dt(
    by = Species,
    rbind(.SD[1],.SD[.N])
  )

#' # for summarise_dt, you can use "by" to calculate within the group
mtcars \%>\%
  summarise_dt(
   disp = mean(disp),
   hp = mean(hp),
   by = cyl
)

  # but you could also, of course, use group_dt
 mtcars \%>\%
   group_dt(by =.(vs,am),
     summarise_dt(avg = mean(mpg)))

  # and list of variables could also be used
 mtcars \%>\%
   group_dt(by =list(vs,am),
            summarise_dt(avg = mean(mpg)))

# examples for `rowwise_dt`
df <- data.table(x = 1:2, y = 3:4, z = 4:5)

df \%>\% mutate_dt(m = mean(c(x, y, z)))

df \%>\% rowwise_dt(
  mutate_dt(m = mean(c(x, y, z)))
)
}
\references{
https://stackoverflow.com/questions/36802385/use-by-each-row-for-data-table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/longer_dt.R
\name{longer_dt}
\alias{longer_dt}
\title{Pivot data from wide to long}
\usage{
longer_dt(.data, ..., name = "name", value = "value", na.rm = FALSE)
}
\arguments{
\item{.data}{A data.frame}

\item{...}{Pattern for unchanged group or unquoted names. Pattern can accept
regular expression to match column names. It can recieve what \code{select_dt}
recieves.}

\item{name}{Name for the measured variable names column.
The default name is 'name'.}

\item{value}{Name for the molten data values column(s).
The default name is 'value'.}

\item{na.rm}{If \code{TRUE}, \code{NA} values will be removed from the molten data.}
}
\value{
A data.table
}
\description{
Turning a wide table to its longer form. It takes multiple columns and collapses into key-value pairs.
}
\examples{

## Example 1:
stocks = data.frame(
  time = as.Date('2009-01-01') + 0:9,
  X = rnorm(10, 0, 1),
  Y = rnorm(10, 0, 2),
  Z = rnorm(10, 0, 4)
)

stocks

stocks \%>\%
  longer_dt(time)

stocks \%>\%
  longer_dt("ti")

# Example 2:

\donttest{
  library(tidyr)

  billboard \%>\%
    longer_dt(
      -"wk",
      name = "week",
      value = "rank",
      na.rm = TRUE
    )

  # or use:
  billboard \%>\%
    longer_dt(
      artist,track,date.entered,
      name = "week",
      value = "rank",
      na.rm = TRUE
    )

  # or use:
  billboard \%>\%
    longer_dt(
      1:3,
      name = "week",
      value = "rank",
      na.rm = TRUE
    )
}
}
\seealso{
\code{\link[tidyfst]{wider_dt}},
  \code{\link[data.table]{melt}},
  \code{\link[tidyr]{pivot_longer}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arrange_dt.R
\name{arrange_dt}
\alias{arrange_dt}
\title{Arrange entries in data.frame}
\usage{
arrange_dt(.data, ...)
}
\arguments{
\item{.data}{data.frame}

\item{...}{Arrange by what group? Minus symbol means arrange by
descending order.}
}
\value{
data.table
}
\description{
Order the rows of a data frame rows by the values of selected columns.
}
\examples{

iris \%>\% arrange_dt(Sepal.Length)

# minus for decreasing order
iris \%>\% arrange_dt(-Sepal.Length)

# arrange by multiple variables
iris \%>\% arrange_dt(Sepal.Length,Petal.Length)

}
\seealso{
\code{\link[dplyr]{arrange}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutate_when.R
\name{mutate_when}
\alias{mutate_when}
\alias{mutate_vars}
\title{Conditional update of columns in data.table}
\usage{
mutate_when(.data, when, ..., by)

mutate_vars(.data, .cols = NULL, .func, ..., by)
}
\arguments{
\item{.data}{data.frame}

\item{when}{An object which can be coerced to logical mode}

\item{...}{Name-value pairs of expressions for \code{mutate_when}.
Additional parameters to be passed to parameter '.func' in \code{mutate_vars}.}

\item{by}{(Optional) Mutate by what group?}

\item{.cols}{Any types that can be accepted by \code{\link[tidyfst]{select_dt}}.}

\item{.func}{Function to be run within each column, should return a value or
vectors with same length.}
}
\value{
data.table
}
\description{
Update or add columns when the given condition is met.

\code{mutate_when} integrates \code{mutate} and \code{case_when}
in \pkg{dplyr} and make a new tidy verb for data.table. \code{mutate_vars} is
 a super function to do updates in specific columns according to conditions.
}
\examples{
iris[3:8,]
iris[3:8,] \%>\%
  mutate_when(Petal.Width == .2,
              one = 1,Sepal.Length=2)

iris \%>\% mutate_vars("Pe",scale)
iris \%>\% mutate_vars(is.numeric,scale)
iris \%>\% mutate_vars(-is.factor,scale)
iris \%>\% mutate_vars(1:2,scale)
iris \%>\% mutate_vars(.func = as.character)
}
\seealso{
\code{\link[tidyfst]{select_dt}}, \code{\link[dplyr]{case_when}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nest_dt.R
\name{nest_dt}
\alias{nest_dt}
\alias{unnest_dt}
\alias{squeeze_dt}
\alias{chop_dt}
\alias{unchop_dt}
\title{Nest and unnest}
\usage{
nest_dt(.data, ..., mcols = NULL, .name = "ndt")

unnest_dt(.data, ...)

squeeze_dt(.data, ..., .name = "ndt")

chop_dt(.data, ...)

unchop_dt(.data, ...)
}
\arguments{
\item{.data}{data.table, nested or unnested}

\item{...}{The variables for nest group(for \code{nest_dt}),
columns to be nested(for \code{squeeze_dt} and \code{chop_dt}),
or column(s) to be unnested(for \code{unnest_dt}).
Could recieve anything that \code{\link[tidyfst]{select_dt}} could receive.}

\item{mcols}{Name-variable pairs in the list, form like}

\item{.name}{Character. The nested column name. Defaults to "ndt".
\code{list(petal="^Pe",sepal="^Se")}, see example.}
}
\value{
data.table, nested or unnested
}
\description{
Create or melt list columns in data.frame.

Analogous function for \code{nest} and \code{unnest} in \pkg{tidyr}.
\code{unnest_dt} will automatically remove other list-columns except for the
target list-columns (which would be unnested later). Also, \code{squeeze_dt} is
designed to merge multiple columns into list column.
}
\details{
In the \code{nest_dt}, the data would be nested to a column named `ndt`,
 which is short for nested data.table.

The \code{squeeze_dt} would not remove the originial columns.

The \code{unchop_dt} is the reverse operation of \code{chop_dt}.

These functions are experiencing the experimental stage, especially
the \code{unnest_dt}. If they don't work on some circumtances, try \pkg{tidyr}
package.
}
\examples{

# examples for nest_dt
# nest by which columns?
 mtcars \%>\% nest_dt(cyl)
 mtcars \%>\% nest_dt("cyl")
 mtcars \%>\% nest_dt(cyl,vs)
 mtcars \%>\% nest_dt(vs:am)
 mtcars \%>\% nest_dt("cyl|vs")
 mtcars \%>\% nest_dt(c("cyl","vs"))

 # change the nested column name
 mtcars \%>\% nest_dt(cyl,.name = "data")

# nest two columns directly
iris \%>\% nest_dt(mcols = list(petal="^Pe",sepal="^Se"))

# nest more flexibly
iris \%>\% nest_dt(mcols = list(ndt1 = 1:3,
  ndt2 = "Pe",
  ndt3 = Sepal.Length:Sepal.Width))

# examples for unnest_dt
# unnest which column?
 mtcars \%>\% nest_dt("cyl|vs") \%>\%
   unnest_dt(ndt)
 mtcars \%>\% nest_dt("cyl|vs") \%>\%
   unnest_dt("ndt")

df <- data.table(
  a = list(c("a", "b"), "c"),
  b = list(c(TRUE,TRUE),FALSE),
  c = list(3,c(1,2)),
  d = c(11, 22)
)

df
df \%>\% unnest_dt(a)
df \%>\% unnest_dt(2)
df \%>\% unnest_dt("c")
df \%>\% unnest_dt(cols = names(df)[3])

# You can unnest multiple columns simultaneously
df \%>\% unnest_dt(1:3)
df \%>\% unnest_dt(a,b,c)
df \%>\% unnest_dt("a|b|c")

# examples for squeeze_dt
# nest which columns?
iris \%>\% squeeze_dt(1:2)
iris \%>\% squeeze_dt("Se")
iris \%>\% squeeze_dt(Sepal.Length:Petal.Width)
iris \%>\% squeeze_dt(1:2,.name = "data")

# examples for chop_dt
df <- data.table(x = c(1, 1, 1, 2, 2, 3), y = 1:6, z = 6:1)
df \%>\% chop_dt(y,z)
df \%>\% chop_dt(y,z) \%>\% unchop_dt(y,z)
}
\references{
https://www.r-bloggers.com/much-faster-unnesting-with-data-table/

https://stackoverflow.com/questions/25430986/create-nested-data-tables-by-collapsing-rows-into-new-data-tables
}
\seealso{
\code{\link[tidyr]{nest}}, \code{\link[tidyr]{chop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/separate.R
\name{separate_dt}
\alias{separate_dt}
\title{Separate a character column into two columns using
a regular expression separator}
\usage{
separate_dt(
  .data,
  separated_colname,
  into,
  sep = "[^[:alnum:]]+",
  remove = TRUE
)
}
\arguments{
\item{.data}{A data frame.}

\item{separated_colname}{Column to be separated, can be a character or alias.}

\item{into}{Character vector of length 2.}

\item{sep}{Separator between columns.}

\item{remove}{If \code{TRUE}, remove input column from output data frame.}
}
\description{
Given either regular expression,
\code{separate_dt()} turns a single character column into two columns.
}
\examples{
df <- data.frame(x = c(NA, "a.b", "a.d", "b.c"))
df \%>\% separate_dt(x, c("A", "B"))
# equals to
df \%>\% separate_dt("x", c("A", "B"))

# If you just want the second variable:
df \%>\% separate_dt(x,into = c(NA,"B"))
}
\seealso{
\code{\link[tidyr]{separate}}, \code{\link[tidyfst]{unite_dt}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/relocate_dt.R
\name{relocate_dt}
\alias{relocate_dt}
\title{Change column order}
\usage{
relocate_dt(.data, ..., how = "first", where = NULL)
}
\arguments{
\item{.data}{A data.frame}

\item{...}{Columns to move}

\item{how}{The mode of movement, including "first","last","after","before".
Default uses "first".}

\item{where}{Destination of columns selected by \code{...}.
Applicable for "after" and "before" mode.}
}
\value{
A data.table with rearranged columns.
}
\description{
Change the position of columns,
 using the same syntax as `select_dt()`. Check similar function
 as `relocate` in \pkg{dplyr}.
}
\examples{
df <- data.table(a = 1, b = 1, c = 1, d = "a", e = "a", f = "a")
df
df \%>\% relocate_dt(f)
df \%>\% relocate_dt(a,how = "last")

df \%>\% relocate_dt(is.character)
df \%>\% relocate_dt(is.numeric, how = "last")
df \%>\% relocate_dt("[aeiou]")

df \%>\% relocate_dt(a, how = "after",where = f)
df \%>\% relocate_dt(f, how = "before",where = a)
df \%>\% relocate_dt(f, how = "before",where = c)
df \%>\% relocate_dt(f, how = "after",where = c)

df2 <- data.table(a = 1, b = "a", c = 1, d = "a")
df2 \%>\% relocate_dt(is.numeric,
                    how = "after",
                    where = is.character)
df2 \%>\% relocate_dt(is.numeric,
                    how="before",
                    where = is.character)
}
\seealso{
\code{\link[dplyr]{relocate}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missing.R
\name{drop_na_dt}
\alias{drop_na_dt}
\alias{replace_na_dt}
\alias{delete_na_cols}
\alias{delete_na_rows}
\alias{fill_na_dt}
\alias{shift_fill}
\title{Dump, replace and fill missing values in data.frame}
\usage{
drop_na_dt(.data, ...)

replace_na_dt(.data, ..., to)

delete_na_cols(.data, prop = NULL, n = NULL)

delete_na_rows(.data, prop = NULL, n = NULL)

fill_na_dt(.data, ..., direction = "down")

shift_fill(x, direction = "down")
}
\arguments{
\item{.data}{data.frame}

\item{...}{Colunms to be replaced or filled. If not specified, use all columns.}

\item{to}{What value should NA replace by?}

\item{prop}{If proportion of NAs is larger than or equal to "prop", would be deleted.}

\item{n}{If number of NAs is larger than or equal to "n", would be deleted.}

\item{direction}{Direction in which to fill missing values.
Currently either "down" (the default) or "up".}

\item{x}{A vector with missing values to be filled.}
}
\value{
data.table
}
\description{
A set of tools to deal with missing values in data.frames. 
It can dump, replace, fill (with next or previous observation) or delete entries according to their missing values.
}
\details{
\code{drop_na_dt} drops the entries with NAs in specific columns.
\code{fill_na_dt} fill NAs with observations ahead ("down") or below ("up"),
which is also known as last observation carried forward (LOCF) and
next observation carried backward(NOCB).

\code{delete_na_cols} could drop the columns with NA proportion larger
than or equal to "prop" or NA number larger than or equal to "n",
\code{delete_na_rows} works alike but deals with rows.

\code{shift_fill} could fill a vector with missing values.
}
\examples{
df <- data.table(x = c(1, 2, NA), y = c("a", NA, "b"))
 df \%>\% drop_na_dt()
 df \%>\% drop_na_dt(x)
 df \%>\% drop_na_dt(y)
 df \%>\% drop_na_dt(x,y)

 df \%>\% replace_na_dt(to = 0)
 df \%>\% replace_na_dt(x,to = 0)
 df \%>\% replace_na_dt(y,to = 0)
 df \%>\% replace_na_dt(x,y,to = 0)

 df \%>\% fill_na_dt(x)
 df \%>\% fill_na_dt() # not specified, fill all columns
 df \%>\% fill_na_dt(y,direction = "up")

x = data.frame(x = c(1, 2, NA, 3), y = c(NA, NA, 4, 5),z = rep(NA,4))
x
x \%>\% delete_na_cols()
x \%>\% delete_na_cols(prop = 0.75)
x \%>\% delete_na_cols(prop = 0.5)
x \%>\% delete_na_cols(prop = 0.24)
x \%>\% delete_na_cols(n = 2)

x \%>\% delete_na_rows(prop = 0.6)
x \%>\% delete_na_rows(n = 2)

# shift_fill
y = c("a",NA,"b",NA,"c")

shift_fill(y) # equals to
shift_fill(y,"down")

shift_fill(y,"up")
}
\references{
https://stackoverflow.com/questions/23597140/how-to-find-the-percentage-of-nas-in-a-data-frame

https://stackoverflow.com/questions/2643939/remove-columns-from-dataframe-where-all-values-are-na

https://stackoverflow.com/questions/7235657/fastest-way-to-replace-nas-in-a-large-data-table
}
\seealso{
\code{\link[tidyr]{drop_na}},\code{\link[tidyr]{replace_na}},
\code{\link[tidyr]{fill}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind_rows_dt.R
\name{bind_rows_dt}
\alias{bind_rows_dt}
\title{Bind multiple data frames by row}
\usage{
bind_rows_dt(...)
}
\arguments{
\item{...}{Data frames to combine. Each argument can either be a data frame,
a list that could be a data frame, or a list of data frames.
Columns are matched by name, and any missing columns will be filled with `NA`.}
}
\value{
data.table
}
\description{
Bind any number of data frames by row, making a longer result.
Similar to `dplyr::bind_rows`, however, columns with same names but different data types
would be coerced to a single proper data type.
}
\examples{

bind_rows_dt(iris[1:3,],iris[6:8,])

# data frames with same name but different type
# numeric data would be coerced to character data in this case
df1 <- data.frame(x = 1:2, y = letters[1:2])
df2 <- data.frame(x = 4:5, y = 1:2)
bind_rows_dt(df1, df2)

}
\seealso{
\code{\link[dplyr]{bind_rows}},\code{\link[data.table]{rbindlist}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/uncount.R
\name{uncount_dt}
\alias{uncount_dt}
\title{"Uncount" a data frame}
\usage{
uncount_dt(.data, wt, .remove = TRUE)
}
\arguments{
\item{.data}{A data.frame}

\item{wt}{A vector of weights.}

\item{.remove}{Should the column for \code{weights} be removed?
Default uses \code{TRUE}.}
}
\description{
Duplicating rows according to a weighting variable.
 This is the opposite operation of `count_dt`.
 Analogous to `tidyr::uncount`.
}
\examples{

df <- data.table(x = c("a", "b"), n = c(1, 2))
uncount_dt(df, n)
uncount_dt(df,n,FALSE)
}
\seealso{
\code{\link[dplyr]{count}}, \code{\link[tidyr]{uncount}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample.R
\name{sample_dt}
\alias{sample_dt}
\alias{sample_n_dt}
\alias{sample_frac_dt}
\title{Sample rows randomly from a table}
\usage{
sample_dt(.data, n = NULL, prop = NULL, replace = FALSE, by = NULL)

sample_n_dt(.data, size, replace = FALSE, by = NULL)

sample_frac_dt(.data, size, replace = FALSE, by = NULL)
}
\arguments{
\item{.data}{A data.frame}

\item{n}{Number of rows to select}

\item{prop}{Fraction of rows to select}

\item{replace}{Sample with or without replacement? Default uses \code{FALSE}.}

\item{by}{(Optional) Character. Specify if you want to sample by group.}

\item{size}{For \code{sample_n_dt}, the number of rows to select.
For \code{sample_frac_dt}, the fraction of rows to select.}
}
\value{
data.table
}
\description{
Select a number or proportion of rows randomly from the data frame

\code{sample_dt} is a merged version of \code{sample_n_dt} and
\code{sample_frac_dt}, this could be convenient.
}
\examples{
sample_n_dt(mtcars, 10)
sample_n_dt(mtcars, 50, replace = TRUE)
sample_frac_dt(mtcars, 0.1)
sample_frac_dt(mtcars, 1.5, replace = TRUE)


sample_dt(mtcars,n=10)
sample_dt(mtcars,prop = 0.1)


# sample by group(s)
iris \%>\% sample_n_dt(2,by = "Species")
iris \%>\% sample_frac_dt(.1,by = "Species")

mtcars \%>\% sample_n_dt(1,by = c("cyl","vs"))
}
\seealso{
\code{\link[dplyr]{sample_n}},\code{\link[dplyr]{sample_frac}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/join.R
\name{join}
\alias{join}
\alias{inner_join_dt}
\alias{left_join_dt}
\alias{right_join_dt}
\alias{full_join_dt}
\alias{anti_join_dt}
\alias{semi_join_dt}
\title{Join tables}
\usage{
inner_join_dt(x, y, by = NULL, on = NULL, suffix = c(".x", ".y"))

left_join_dt(x, y, by = NULL, on = NULL, suffix = c(".x", ".y"))

right_join_dt(x, y, by = NULL, on = NULL, suffix = c(".x", ".y"))

full_join_dt(x, y, by = NULL, on = NULL, suffix = c(".x", ".y"))

anti_join_dt(x, y, by = NULL, on = NULL)

semi_join_dt(x, y, by = NULL, on = NULL)
}
\arguments{
\item{x}{A data.table}

\item{y}{A data.table}

\item{by}{(Optional) A character vector of variables to join by.

  If `NULL`, the default, `*_join_dt()` will perform a natural join, using all
  variables in common across `x` and `y`. A message lists the variables so that you
  can check they're correct; suppress the message by supplying `by` explicitly.

  To join by different variables on `x` and `y`, use a named vector.
  For example, `by = c("a" = "b")` will match `x$a` to `y$b`.

  To join by multiple variables, use a vector with length > 1.
  For example, `by = c("a", "b")` will match `x$a` to `y$a` and `x$b` to
  `y$b`. Use a named vector to match different variables in `x` and `y`.
  For example, `by = c("a" = "b", "c" = "d")` will match `x$a` to `y$b` and
  `x$c` to `y$d`.}

\item{on}{(Optional)
Indicate which columns in x should be joined with which columns in y.
Examples included:
  1.\code{.by = c("a","b")} (this is a must for \code{set_full_join_dt});
  2.\code{.by = c(x1="y1", x2="y2")};
  3.\code{.by = c("x1==y1", "x2==y2")};
  4.\code{.by = c("a", V2="b")};
  5.\code{.by = .(a, b)};
  6.\code{.by = c("x>=a", "y<=b")} or \code{.by = .(x>=a, y<=b)}.}

\item{suffix}{If there are non-joined duplicate variables in x and y, these
suffixes will be added to the output to disambiguate them. Should be a
character vector of length 2.}
}
\value{
A data.table
}
\description{
The mutating joins add columns from `y` to `x`,
matching rows based on the keys:

* `inner_join_dt()`: includes all rows in `x` and `y`.
* `left_join_dt()`: includes all rows in `x`.
* `right_join_dt()`: includes all rows in `y`.
* `full_join_dt()`: includes all rows in `x` or `y`.

Filtering joins filter rows from `x` based on the presence or absence
of matches in `y`:

* `semi_join_dt()` return all rows from `x` with a match in `y`.
* `anti_join_dt()` return all rows from `x` without a match in `y`.
}
\examples{

workers = fread("
    name company
    Nick Acme
    John Ajax
    Daniela Ajax
")

positions = fread("
    name position
    John designer
    Daniela engineer
    Cathie manager
")

workers \%>\% inner_join_dt(positions)
workers \%>\% left_join_dt(positions)
workers \%>\% right_join_dt(positions)
workers \%>\% full_join_dt(positions)

# filtering joins
workers \%>\% anti_join_dt(positions)
workers \%>\% semi_join_dt(positions)

# To suppress the message, supply 'by' argument
workers \%>\% left_join_dt(positions, by = "name")

# Use a named 'by' if the join variables have different names
positions2 = setNames(positions, c("worker", "position")) # rename first column in 'positions'
workers \%>\% inner_join_dt(positions2, by = c("name" = "worker"))

# the syntax of 'on' could be a bit different
workers \%>\% inner_join_dt(positions2,on = "name==worker")


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pull_dt.R
\name{pull_dt}
\alias{pull_dt}
\title{Pull out a single variable}
\usage{
pull_dt(.data, col)
}
\arguments{
\item{.data}{data.frame}

\item{col}{A name of column or index (should be positive).}
}
\value{
vector
}
\description{
Extract vector from data.frame, works likt `[[`. Analogous function for \code{pull} in \pkg{dplyr}
}
\examples{
mtcars \%>\% pull_dt(2)
mtcars \%>\% pull_dt(cyl)
mtcars \%>\% pull_dt("cyl")
}
\seealso{
\code{\link[dplyr]{pull}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sys_time_print.R
\name{sys_time_print}
\alias{sys_time_print}
\alias{pst}
\title{Convenient print of time taken}
\usage{
sys_time_print(expr)

pst(expr)
}
\arguments{
\item{expr}{Valid R expression to be timed.}
}
\value{
A character vector of the form HH:MM:SS,
or SS.MMMsec if under 60 seconds (invisibly for \code{show_time}). See examples.
}
\description{
Convenient printing of time elapsed. A wrapper of
\code{data.table::timetaken}, but showing the results more directly.
}
\examples{

sys_time_print(Sys.sleep(1))

a = iris
sys_time_print({
  res = iris \%>\%
    mutate_dt(one = 1)
})
res
}
\seealso{
\code{\link[data.table]{timetaken}}, \code{\link[base]{system.time}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_dt.R
\name{filter_dt}
\alias{filter_dt}
\title{Filter entries in data.frame}
\usage{
filter_dt(.data, ...)
}
\arguments{
\item{.data}{data.frame}

\item{...}{List of variables or name-value pairs of summary/modifications
functions.}
}
\value{
data.table
}
\description{
Choose rows where conditions are true.
}
\examples{
iris \%>\% filter_dt(Sepal.Length > 7)
iris \%>\% filter_dt(Sepal.Length == max(Sepal.Length))

# comma is not supported in tidyfst after v0.9.8
# which means you can't use:
\dontrun{
 iris \%>\% filter_dt(Sepal.Length > 7, Sepal.Width > 3)
}
# use following code instead
iris \%>\% filter_dt(Sepal.Length > 7 & Sepal.Width > 3)

}
\seealso{
\code{\link[dplyr]{filter}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transpose.R
\name{t_dt}
\alias{t_dt}
\title{Efficient transpose of data.frame}
\usage{
t_dt(.data)
}
\arguments{
\item{.data}{A data.frame/data.table/tibble}
}
\value{
A transposed data.frame
}
\description{
An efficient way to transpose data frames(data.frame/data.table/tibble).
}
\details{
This function would return the original data.frame structure,
keeping all the row names and column names. If the row names are not
available or, "V1,V2..." will be provided.
}
\examples{

t_dt(iris)
t_dt(mtcars)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pairwise.R
\name{pairwise_count_dt}
\alias{pairwise_count_dt}
\title{Count pairs of items within a group}
\usage{
pairwise_count_dt(
  .data,
  .group,
  .value,
  upper = FALSE,
  diag = FALSE,
  sort = TRUE
)
}
\arguments{
\item{.data}{A data.frame.}

\item{.group}{Column name of counting group.}

\item{.value}{Item to count pairs, will end up in \code{V1} and \code{V2} columns.}

\item{upper}{When \code{FALSE}(Default), duplicated combinations would be removed.}

\item{diag}{Whether to include diagonal (V1==V2) in output. Default uses \code{FALSE}.}

\item{sort}{Whether to sort rows by counts. Default uses \code{TRUE}.}
}
\value{
A data.table with 3 columns (named as "item1","item2" and "n"), containing combinations
 in "item1" and "item2", and counts in "n".
}
\description{
Count the number of times each pair
of items appear together within a group.
For example, this could count the number of times two words appear within documents.
This function has referred to \code{pairwise_count} in \strong{widyr} package,
but with very different defaults on several parameters.
}
\examples{

dat <- data.table(group = rep(1:5, each = 2),
              letter = c("a", "b",
                         "a", "c",
                         "a", "c",
                         "b", "e",
                         "b", "f"))
pairwise_count_dt(dat,group,letter)
pairwise_count_dt(dat,group,letter,sort = FALSE)
pairwise_count_dt(dat,group,letter,upper = TRUE)
pairwise_count_dt(dat,group,letter,diag = TRUE)
pairwise_count_dt(dat,group,letter,diag = TRUE,upper = TRUE)

# The column name could be specified using character.
pairwise_count_dt(dat,"group","letter")
}
\seealso{
\code{\link[widyr]{pairwise_count}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/col_max.R
\name{col_max}
\alias{col_max}
\alias{col_min}
\title{Get the column name of the max/min number each row}
\usage{
col_max(.data, ..., .name = "max_col")

col_min(.data, ..., .name = "min_col")
}
\arguments{
\item{.data}{A data.frame with numeric column(s)}

\item{...}{Variables for screening, could receive what `select_dt` receives.
When starts with '-'(minus symbol) or '!', return the negative columns.}

\item{.name}{The column name of the new added column}
}
\value{
A data.table
}
\description{
For a data.frame with numeric values, add a new column
specifying the column name of the first max/min value each row.
}
\examples{
set.seed(199057)
DT <- data.table(matrix(sample(10, 100, TRUE), ncol=10))
DT
col_max(DT)
col_max(DT,V1:V3)
col_max(DT,.name = "max_col_name")
col_min(DT)
col_min(DT,2:4)

col_max(iris)

}
\references{
https://stackoverflow.com/questions/17735859/for-each-row-return-the-column-name-of-the-largest-value
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarise_dt.R
\name{summarise_dt}
\alias{summarise_dt}
\alias{summarize_dt}
\alias{summarise_when}
\alias{summarize_when}
\alias{summarise_vars}
\alias{summarize_vars}
\title{Summarise columns to single values}
\usage{
summarise_dt(.data, ..., by = NULL)

summarize_dt(.data, ..., by = NULL)

summarise_when(.data, when, ..., by = NULL)

summarize_when(.data, when, ..., by = NULL)

summarise_vars(.data, .cols = NULL, .func, ..., by)

summarize_vars(.data, .cols = NULL, .func, ..., by)
}
\arguments{
\item{.data}{data.frame}

\item{...}{List of variables or name-value pairs of summary/modifications
functions for \code{summarise_dt}.Additional parameters to be passed to
 parameter '.func' in \code{summarise_vars}.}

\item{by}{unquoted name of grouping variable of list of unquoted names of
grouping variables. For details see \link[data.table]{data.table}}

\item{when}{An object which can be coerced to logical mode}

\item{.cols}{Columns to be summarised.}

\item{.func}{Function to be run within each column, should return a value or vectors with same length.}
}
\value{
data.table
}
\description{
Summarise group of values into one value for each group. If there is only one group, then only one value would be returned.
 The summarise function should always return a single value.
}
\details{
\code{summarise_vars} could complete summarise on specific columns.
}
\examples{
iris \%>\% summarise_dt(avg = mean(Sepal.Length))
iris \%>\% summarise_dt(avg = mean(Sepal.Length),by = Species)
mtcars \%>\% summarise_dt(avg = mean(hp),by = .(cyl,vs))

# the data.table way
mtcars \%>\% summarise_dt(cyl_n = .N, by = .(cyl, vs)) # `.` is short for list

iris \%>\% summarise_vars(is.numeric,min)
iris \%>\% summarise_vars(-is.factor,min)
iris \%>\% summarise_vars(1:4,min)

iris \%>\% summarise_vars(is.numeric,min,by ="Species")
mtcars \%>\% summarise_vars(is.numeric,mean,by = c("vs", "am"))

# use multiple functions on multiple columns
iris \%>\%
  summarise_vars(is.numeric,.func = list(mean,sd,median))
iris \%>\%
  summarise_vars(is.numeric,.func = list(mean,sd,median),by = Species)

}
\seealso{
\code{\link[dplyr]{summarise}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rownames.R
\name{rn_col}
\alias{rn_col}
\alias{col_rn}
\title{Tools for working with row names}
\usage{
rn_col(.data, var = "rowname")

col_rn(.data, var = "rowname")
}
\arguments{
\item{.data}{A data.frame.}

\item{var}{Name of column to use for rownames.}
}
\value{
\code{rn_col} returns a data.table,
\code{col_rn} returns a data frame.
}
\description{
The enhanced data.frame, including tibble and data.table, do not
support row names. To link to some base r facilities, there should be functions
to save information in row names. These functions are analogous to
\code{rownames_to_column} and \code{column_to_rownames} in \pkg{tibble}.
}
\examples{

 mtcars \%>\% rn_col()
 mtcars \%>\% rn_col("rn")

 mtcars \%>\% rn_col() -> new_mtcars

 new_mtcars \%>\% col_rn() -> old_mtcars
 old_mtcars
 setequal(mtcars,old_mtcars)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rename_dt.R
\name{rename_dt}
\alias{rename_dt}
\alias{rename_with_dt}
\title{Rename column in data.frame}
\usage{
rename_dt(.data, ...)

rename_with_dt(.data, .fn, ...)
}
\arguments{
\item{.data}{data.frame}

\item{...}{statements of rename, e.g. `sl = Sepal.Length` means the column named
as "Sepal.Length" would be renamed to "sl"}

\item{.fn}{A function used to transform the selected columns.
Should return a character vector the same length as the input.}
}
\value{
data.table
}
\description{
Rename one or more columns in the data.frame.
}
\examples{
iris \%>\%
  rename_dt(sl = Sepal.Length,sw = Sepal.Width) \%>\%
  head()
iris \%>\% rename_with_dt(toupper)
iris \%>\% rename_with_dt(toupper,"^Pe")

}
\seealso{
\code{\link[dplyr]{rename}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rec.R
\name{rec}
\alias{rec}
\alias{rec_num}
\alias{rec_char}
\title{Recode number or strings}
\usage{
rec_num(x, rec, keep = TRUE)

rec_char(x, rec, keep = TRUE)
}
\arguments{
\item{x}{A numeric or character vector.}

\item{rec}{String with recode pairs of old and new values.
Find the usage in examples.}

\item{keep}{Logical. Decide whether to keep the original values if not recoded.
Defaults to \code{TRUE}.}
}
\value{
A vector.
}
\description{
Recode discrete variables, including numerice and character
variable.
}
\examples{

x = 1:10
x
rec_num(x, rec = "1=10; 4=2")
rec_num(x, rec = "1:3=1; 4:6=2")
rec_num(x, rec = "1:3=1; 4:6=2",keep = FALSE)

y = letters[1:5]
y
rec_char(y,rec = "a=A;b=B")
rec_char(y,rec = "a,b=A;c,d=B")
rec_char(y,rec = "a,b=A;c,d=B",keep = FALSE)

}
\seealso{
\code{\link[sjmisc]{rec}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fst_io.R
\name{export_fst}
\alias{export_fst}
\alias{import_fst}
\title{Read and write fst files}
\usage{
export_fst(x, path, compress = 100, uniform_encoding = TRUE)

import_fst(
  path,
  columns = NULL,
  from = 1,
  to = NULL,
  as.data.table = TRUE,
  old_format = FALSE
)
}
\arguments{
\item{x}{a data frame to write to disk}

\item{path}{path to fst file}

\item{compress}{value in the range 0 to 100, indicating the amount of compression to use.
Lower values mean larger file sizes. The default compression is set to 50.}

\item{uniform_encoding}{If `TRUE`, all character vectors will be assumed to have elements with equal encoding.
The encoding (latin1, UTF8 or native) of the first non-NA element will used as encoding for the whole column.
This will be a correct assumption for most use cases.
If `uniform.encoding` is set to `FALSE`, no such assumption will be made and all elements will be converted
to the same encoding. The latter is a relatively expensive operation and will reduce write performance for
character columns.}

\item{columns}{Column names to read. The default is to read all columns.}

\item{from}{Read data starting from this row number.}

\item{to}{Read data up until this row number. The default is to read to the last row of the stored dataset.}

\item{as.data.table}{If TRUE, the result will be returned as a \code{data.table} object. Any keys set on
dataset \code{x} before writing will be retained. This allows for storage of sorted datasets. This option
requires \code{data.table} package to be installed.}

\item{old_format}{must be FALSE, the old fst file format is deprecated and can only be read and
converted with fst package versions 0.8.0 to 0.8.10.}
}
\value{
`import_fst` returns a data.table with the selected columns and rows. `export_fst`
writes `x` to a `fst` file and invisibly returns `x` (so you can use this function in a pipeline).
}
\description{
Wrapper for \code{\link[fst]{read_fst}} and \code{\link[fst]{write_fst}}
from \pkg{fst}, but use a different default. For data import, always return a data.table.
For data export, always compress the data to the smallest size.
}
\examples{
\dontrun{
export_fst(iris,"iris_fst_test.fst")
iris_dt = import_fst("iris_fst_test.fst")
iris_dt
unlink("iris_fst_test.fst")
}
}
\seealso{
\code{\link[fst]{read_fst}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global_setting.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{\%>\%}
\alias{\%like\%}
\alias{\%chin\%}
\alias{data.table}
\alias{as.data.table}
\alias{fread}
\alias{fwrite}
\alias{fintersect}
\alias{fsetdiff}
\alias{funion}
\alias{fsetequal}
\alias{frollapply}
\alias{fcoalesce}
\alias{uniqueN}
\alias{rbindlist}
\alias{tables}
\alias{like}
\alias{copy}
\alias{key}
\alias{CJ}
\alias{rleid}
\alias{rleidv}
\alias{fcase}
\alias{between}
\alias{set}
\alias{setDT}
\alias{setDF}
\alias{setnames}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{data.table}{\code{\link[data.table:chmatch]{\%chin\%}}, \code{\link[data.table:like]{\%like\%}}, \code{\link[data.table]{as.data.table}}, \code{\link[data.table]{between}}, \code{\link[data.table:J]{CJ}}, \code{\link[data.table]{copy}}, \code{\link[data.table]{data.table}}, \code{\link[data.table]{fcase}}, \code{\link[data.table:coalesce]{fcoalesce}}, \code{\link[data.table:setops]{fintersect}}, \code{\link[data.table]{fread}}, \code{\link[data.table]{frollapply}}, \code{\link[data.table:setops]{fsetdiff}}, \code{\link[data.table:setops]{fsetequal}}, \code{\link[data.table:setops]{funion}}, \code{\link[data.table]{fwrite}}, \code{\link[data.table:setkey]{key}}, \code{\link[data.table]{like}}, \code{\link[data.table]{rbindlist}}, \code{\link[data.table]{rleid}}, \code{\link[data.table:rleid]{rleidv}}, \code{\link[data.table:assign]{set}}, \code{\link[data.table]{setDF}}, \code{\link[data.table]{setDT}}, \code{\link[data.table:setattr]{setnames}}, \code{\link[data.table]{tables}}, \code{\link[data.table:duplicated]{uniqueN}}}

  \item{stringr}{\code{\link[stringr:pipe]{\%>\%}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_fst.R
\name{as_fst}
\alias{as_fst}
\title{Save a data.frame as a fst table}
\usage{
as_fst(.data)
}
\arguments{
\item{.data}{A data.frame}
}
\value{
An object of class \code{fst_table}
}
\description{
This function first export the data.frame to a temporal file,
and then parse it back as a fst table (class name is "fst_table").
}
\examples{

\dontrun{
  iris \%>\%
    as_fst() -> iris_fst
  iris_fst
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dummy_dt.R
\name{dummy_dt}
\alias{dummy_dt}
\title{Fast creation of dummy variables}
\usage{
dummy_dt(.data, ..., longname = TRUE)
}
\arguments{
\item{.data}{data.frame}

\item{...}{Columns you want to create dummy variables from.
Very flexible, find in the examples.}

\item{longname}{logical. Should the output column labeled with the
original column name? Default uses \code{TRUE}.}
}
\value{
data.table
}
\description{
Quickly create dummy (binary) columns from character and factor type columns in the inputted data (and numeric columns if specified.)
This function is useful for statistical analysis when you want binary columns rather than character columns.
}
\details{
If no columns provided, will return the original data frame.
 When NA exist in the input column, they would also be considered. If
 the input character column contains both NA and string "NA", they would be
 merged.

This function is inspired by \pkg{fastDummies} package, but provides
simple and precise usage, whereas \code{fastDummies::dummy_cols} provides more
features for statistical usage.
}
\examples{
iris \%>\% dummy_dt(Species)
iris \%>\% dummy_dt(Species,longname = FALSE)

mtcars \%>\% head() \%>\% dummy_dt(vs,am)
mtcars \%>\% head() \%>\% dummy_dt("cyl|gear")

# when there are NAs in the column
df <- data.table(x = c("a", "b", NA, NA),y = 1:4)
df \%>\%
  dummy_dt(x)

# when NA  and "NA" both exist, they would be merged
df <- data.table(x = c("a", "b", NA, "NA"),y = 1:4)
df \%>\%
  dummy_dt(x)

}
\references{
https://stackoverflow.com/questions/18881073/creating-dummy-variables-in-r-data-table
}
\seealso{
\code{\link[fastDummies]{dummy_cols}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unite.R
\name{unite_dt}
\alias{unite_dt}
\title{Unite multiple columns into one by pasting strings together}
\usage{
unite_dt(
  .data,
  united_colname,
  ...,
  sep = "_",
  remove = FALSE,
  na2char = FALSE
)
}
\arguments{
\item{.data}{A data frame.}

\item{united_colname}{The name of the new column, string only.}

\item{...}{A selection of columns. If want to select all columns,
pass "" to the parameter. See example.}

\item{sep}{Separator to use between values.}

\item{remove}{If \code{TRUE}, remove input columns from output data frame.}

\item{na2char}{If \code{FALSE}, missing values would be merged into \code{NA},
otherwise \code{NA} is treated as character "NA". This is different from
\pkg{tidyr}.}
}
\description{
Convenience function to paste together multiple columns into one.
}
\examples{
df <- expand.grid(x = c("a", NA), y = c("b", NA))
df

# Treat missing value as NA, default
df \%>\% unite_dt("z", x:y, remove = FALSE)
# Treat missing value as character "NA"
df \%>\% unite_dt("z", x:y, na2char = TRUE, remove = FALSE)
df \%>\%
  unite_dt("xy", x:y)

# Select all columns
iris \%>\% unite_dt("merged_name",".")
}
\seealso{
\code{\link[tidyr]{unite}},\code{\link[tidyfst]{separate_dt}}
}
